
//------------------------------------------------------------------------
// Plane Graphic Calculator
//
// This software has been donated to the public.
//
// Version 1.1 -- June 2001
//------------------------------------------------------------------------


//------------------------------------------------------------------------
// Build environment:
//
// Language:          Java 1.0
// Verified on:       JDK 1.1.7
//
// Build steps:
//                    javac PGC.java
//                    jar cf PGC.jar *.class
//------------------------------------------------------------------------

import java.awt.*;
import java.lang.Math;
import java.util.*;


//========================================================================
//========================================================================
// PGC Applet
//========================================================================
//========================================================================

public class PGC extends java.applet.Applet
	{
	private String[]                main_args;
	private PGC_Frame               main_frame = null;
	private Container               main_parent = null;
	private Dimension               main_size = null;
	private Point                   main_location = null;
	private Panel                   pnl_root_launch;
	private Panel                   pnl_root_calculator;
	private PGC_GraphPanel          pnl_graph;
	private Panel                   pnl_ctl;
	private Panel                   pnl_ctl_buttons;
	private Panel                   pnl_ctl_buttons_function;
	private Panel                   pnl_ctl_buttons_zoom;
	private Panel                   pnl_ctl_buttons_func;
	private Panel                   pnl_ctl_coordinates;
	private Panel                   pnl_ctl_coordinates_description;
	private PGC_ChoiceOp            choice_op;
	private PGC_ButtonFixed         btn_zoomout;
	private PGC_ButtonFixed         btn_zoomin;
	private Checkbox                chk_line;
	private Checkbox                chk_spin;
	private Checkbox                chk_plot;
	private PGC_ButtonFixed         btn_details;
	private PGC_ButtonFixed         btn_step;
	private PGC_ButtonFixed         btn_plot_clear;
	private PGC_ButtonFixed         btn_about;
	private PGC_ButtonFixed         btn_attach_detach;
	private Button                  btn_launch;
	private Label                   lbl_zoom;
	private PGC_LabelFixed          lbl_description;
	private PGC_LabelFixed          lbl_coordinate_x;
	private PGC_LabelFixed          lbl_coordinate_y;
	private PGC_LabelFixed          lbl_f_expr;
	private PGC_LabelFixed          lbl_g_expr;
	private PGC_TextFieldFixed      txt_f_expr;
	private PGC_TextFieldFixed      txt_g_expr;
	private PGC_TextFrame           about_frame = null;
	private Point                   about_location;
	private PGC_DetailsFrame        details_frame = null;
	private Point                   details_location;
	private static PGC_TextFrame    trace_frame = null;
	private PGC_Eval                eval = new PGC_Eval(this);

	//----------------------------------------------------------------

	public  boolean                 stand_alone                 = false;
	public  boolean                 launch_mode                 = false;
	public  static boolean          option_enable_trace         = true;
	public  String                  option_launch_button        = "";
	public  int                     option_launch_size_x        = 560;
	public  int                     option_launch_size_y        = 440;
	public  String                  option_background_image     = null;
	public  boolean                 option_control_panel        = true;
	public  boolean                 option_graph_border         = true;
	public  boolean                 option_builtin_functions    = true;
	public  boolean                 option_builtin_examples     = true;
	public  Complex                 option_center               = Complex.Number(0);
	public  int                     option_zoom                 = 80;
	public  boolean                 option_line                 = true;
	public  boolean                 option_plot                 = false;
	public  boolean                 option_spin                 = false;
	public  int                     option_active               = 0;
	public  int                     option_op                   = OP_ADD;
	public  PGC_Variable[]          variable                    = new PGC_Variable[VAR_COUNT];
	public  double                  t_minimum                   = -10.0;
	public  double                  t_maximum                   =  10.0;
	public  int                     t_count                     = 200;

	//----------------------------------------------------------------
	// PGC: MainFrame
	//----------------------------------------------------------------

	public void MainFrameLaunch()
		{
		//---------------------------------------------
		// Creates the PGC_Frame from the "Launch"
		// button.
		//---------------------------------------------
		if (main_frame == null)
			{
			main_frame = new PGC_Frame(this, "Plane Graphic Calculator");

			LayoutCalculator(); // defines pnl_root_calculator
			SetOptions();

			main_frame.setLayout(new BorderLayout());
			main_frame.add("Center", pnl_root_calculator);

			main_frame.pack();

			Dimension dim;
			if (main_size == null)
				{
				dim = new Dimension(option_launch_size_x, option_launch_size_y);
				dim.height += 50; // adjustment for title bar and frame height (a reasonable average needed)
				dim.width  += 6;  // adjustment for frame width (a reasonable average needed)

				main_size = dim;
				}
			else
				{
				dim = new Dimension(main_size); // remembered from previous attach
				}
			main_frame.resize(dim);

			if (main_location == null)
				{
				main_location = new Point(20, 20);
				}
			main_frame.move(main_location.x, main_location.y);

			main_frame.validate();
			if (btn_attach_detach != null) btn_attach_detach.setLabel("Close");
			main_frame.show();
			}
		else
			{
			main_frame.toFront();
			main_frame.show();
			main_frame.requestFocus();
			}
		}
	public void MainFrameUnlaunch()
		{
		//---------------------------------------------
		// Closes the PGC_Frame that was created
		// from the "Launch" button.
		//---------------------------------------------
		if (main_frame != null)
			{
			main_size = new Dimension(main_frame.size());      // remember for next detach

			Point loc = main_frame.location();
			main_location = new Point(loc.x, loc.y);  // remember for next detach

			main_frame.remove(this);
			main_frame.dispose();
			main_frame = null;
			}
		}
	public void MainFrameDetach()
		{
		//---------------------------------------------
		// Creates the PGC_Frame in stand_alone mode,
		// or detaches it from the web-page when the
		// "Detach" button is used.
		//---------------------------------------------
		main_parent = getParent(); // remember for next attach

		main_frame = new PGC_Frame(this, "Plane Graphic Calculator");

		Dimension dim;
		if (main_size == null)
			{
			if (main_parent == null)
				main_size = new Dimension(size()); // stand_alone -- no parent
			else
				main_size = new Dimension(main_parent.size()); // size from parent web-page

			dim = new Dimension(main_size);
			dim.height += 50; // adjustment for title bar and frame height (a reasonable average needed)
			dim.width  += 6;  // adjustment for frame width (a reasonable average needed)
			}
		else
			{
			dim = new Dimension(main_size); // remembered from previous attach
			}

		main_frame.resize(dim);

		if (main_location == null)
			{
			//-------------------------------------------------------
			// This is the first detach.  Show its location offset
			// slightly so it's obvious that it's a detached window.
			//-------------------------------------------------------
			if (main_parent == null)
				{
				main_location = new Point(20, 20);
				}
			else
				{
				Point loc = main_parent.location();
				main_location = new Point(loc.x + 20, loc.y + 20);
				}
			}
		main_frame.move(main_location.x, main_location.y);

		main_frame.add("Center", this);
		main_frame.validate();
		if (btn_attach_detach != null) btn_attach_detach.setLabel("Attach");
		main_frame.show();
		}
	public void MainFrameAttach()
		{
		//---------------------------------------------
		// Re-attaches the PGC_Frame to the web-page
		// when the "Attach" button is used.
		// Quits in stand_alone mode.
		//---------------------------------------------
		if (stand_alone)
			System.exit(0); // nothing to attach to

		main_size = new Dimension(getParent().size());      // remember for next detach

		Point loc = getParent().location();
		main_location = new Point(loc.x, loc.y);  // remember for next detach

		if (main_frame != null)
			{
			main_frame.remove(this);
			main_frame.dispose();
			main_frame = null;
			}
		if (main_parent != null)
			{
			main_parent.add("Center", this);
			main_parent.validate();
			if (btn_attach_detach != null) btn_attach_detach.setLabel("Detach");
			main_parent.show();
			}
		}
	public void MainFrameDestroy()
		{
		if (main_frame != null)
			{
			main_frame.remove(this);
			main_frame.dispose();
			main_frame = null;
			}
		FrameDestroy(about_frame);
		FrameDestroy(details_frame);
		}
	public void MainFrameClose()
		{
		if (launch_mode)
			{
			MainFrameUnlaunch();
			MainFrameDestroy();
			}
		else // attach-detach mode
			{
			MainFrameAttach();
			}
		}

	//----------------------------------------------------------------
	// PGC: Standard colors and layout
	//----------------------------------------------------------------

	public static Color color_panel                      = new Color(204,204,204);
	public static Color color_background                 = new Color(255,255,255);
	public static Color color_axis                       = new Color(  0,  0,  0);
	public static Color color_grid                       = new Color(236,236,236);
	public static Color color_origin_connector_line      = new Color(204,204,204);
	public static Color color_input_point                = new Color(  0,  0,  0);
	public static Color color_input_label                = new Color(102,102,102);
	public static Color color_func_f_point               = new Color(  0,  0,255);
	public static Color color_func_f_label               = new Color(  0,  0,255);
	public static Color color_func_f_plot                = new Color(153,153,255);
	public static Color color_func_g_point               = new Color(204,  0,  0);
	public static Color color_func_g_label               = new Color(204,  0,  0);
	public static Color color_func_g_plot                = new Color(255,153,153);
	public static Color color_func_h_point               = new Color(  0,153,  0);
	public static Color color_func_h_label               = new Color(  0,153,  0);
	public static Color color_func_h_plot                = new Color(  0,204,  0);
	public static Color color_func_k_point               = new Color(204,  0,204);
	public static Color color_func_k_label               = new Color(204,  0,204);
	public static Color color_func_k_plot                = new Color(204,102,204);
	public static Color color_func_m_point               = new Color(  0,153,153);
	public static Color color_func_m_label               = new Color(  0,153,153);
	public static Color color_func_m_plot                = new Color(102,204,204);
	public static Color color_func_n_point               = new Color(102,102,102);
	public static Color color_func_n_label               = new Color(102,102,102);
	public static Color color_func_n_plot                = new Color(153,153,153);
	public static Color color_lightray                   = new Color(255,204,  0);
	public static Color color_projection                 = new Color(255,102,153);
	public static Color color_projection_label           = new Color(255,  0,  0);
	public static Color color_3d_dark                    = new Color(153,153,153);
	public static Color color_3d_light                   = new Color(236,236,236);

	public static Dimension component_tiny_button_size   = new Dimension(24, -1);
	public static Dimension component_normal_button_size = new Dimension(64, -1);
	public static Dimension component_number_text_size   = new Dimension(164, -1);
	public static Dimension component_wide_text_size     = new Dimension(360, -1);
	public static Dimension component_full_text_size     = new Dimension(388, -1);

	public static GridBagConstraints StdConstraintsFillMax()
		{
		//-------------------------------------------------------
		// Standard constraints for a component that
		// fills the maximum space possible.
		//-------------------------------------------------------
		GridBagConstraints result = new GridBagConstraints();
		result.gridwidth  = GridBagConstraints.REMAINDER;
		result.weightx    = 1.0;
		result.weighty    = 1.0;
		result.fill       = GridBagConstraints.BOTH;
		return (result);
		}
	public static GridBagConstraints StdConstraintsFillMax(Insets insets)
		{
		GridBagConstraints result = StdConstraintsFillMax();
		result.insets = insets;
		return (result);
		}
	public static GridBagConstraints StdConstraints()
		{
		GridBagConstraints result = new GridBagConstraints();
		result.anchor = GridBagConstraints.WEST;
		return (result);
		}
	public static GridBagConstraints StdConstraints(Insets insets)
		{
		GridBagConstraints result = StdConstraints();
		result.insets = insets;
		return (result);
		}
	public static GridBagConstraints StdConstraintsNewLine()
		{
		GridBagConstraints result = new GridBagConstraints();
		result.gridwidth = GridBagConstraints.REMAINDER;
		result.anchor    = GridBagConstraints.WEST;
		return (result);
		}
	public static GridBagConstraints StdConstraintsNewLine(Insets insets)
		{
		GridBagConstraints result = StdConstraintsNewLine();
		result.insets = insets;
		return (result);
		}
	public static Panel StdPanel()
		{
		Panel result = new Panel();
		result.setBackground(PGC.color_panel);
		result.setLayout(new GridBagLayout());
		return (result);
		}

	//----------------------------------------------------------------
	// PGC: LayoutCalculator
	//----------------------------------------------------------------

	private void LayoutCalculator()
		{
		Component component;
		GridBagConstraints constraints;

		component = pnl_root_calculator = PGC.StdPanel();

		//----------------------------------------------
		// pnl_graph (graph surface -- the upper part)
		//----------------------------------------------

		pnl_root_calculator.add(component = pnl_graph = new PGC_GraphPanel(this));
		((GridBagLayout)(pnl_root_calculator.getLayout())).setConstraints(component, PGC.StdConstraintsFillMax());

		//----------------------------------------------
		// pnl_ctl (control panel -- the lower part)
		//----------------------------------------------

		if (option_control_panel)
			{
			pnl_root_calculator.add(component = pnl_ctl = PGC.StdPanel());
			((GridBagLayout)(pnl_root_calculator.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine(new Insets(5,10,5,0)));

			pnl_ctl.add(component = pnl_ctl_buttons = PGC.StdPanel());
			((GridBagLayout)(pnl_ctl.getLayout())).setConstraints(component, PGC.StdConstraints());

			pnl_ctl_buttons.add(component = pnl_ctl_buttons_function = PGC.StdPanel());
			((GridBagLayout)(pnl_ctl_buttons.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine(new Insets(3,0,0,0)));
			pnl_ctl_buttons_function.add(component = new Label("Function: "));
			pnl_ctl_buttons_function.add(component = choice_op = new PGC_ChoiceOp(this));
			choice_op.setBackground(Color.white);
			for (int op = 1; op < OP_COUNT; op++)
				{
				if ((op < OP_USER) && !option_builtin_functions) continue;
				if ((op > OP_USER) && !option_builtin_examples)  continue;
				choice_op.addItem(OpChoiceLabel(op));
				}
			pnl_ctl_buttons_function.add(component = new PGC_LabelFixed(new Dimension(10,-1)));
			pnl_ctl_buttons_function.add(component = btn_details = new PGC_ButtonFixed("Details", PGC.component_normal_button_size));
			pnl_ctl_buttons_function.add(component = btn_step = new PGC_ButtonFixed("Step", PGC.component_normal_button_size));
			pnl_ctl_buttons_function.add(component = btn_plot_clear = new PGC_ButtonFixed("Clear", PGC.component_normal_button_size));
			pnl_ctl_buttons_function.add(component = btn_about = new PGC_ButtonFixed("About", PGC.component_normal_button_size));
			if (!stand_alone)
				{
				pnl_ctl_buttons_function.add(component = btn_attach_detach = new PGC_ButtonFixed("Detach", PGC.component_normal_button_size));
				}

			pnl_ctl_buttons.add(component = pnl_ctl_buttons_zoom = PGC.StdPanel());
			((GridBagLayout)(pnl_ctl_buttons.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine(new Insets(3,0,0,0)));
			pnl_ctl_buttons_zoom.add(component = lbl_zoom = new Label("Zoom: 9999 X "));
			pnl_ctl_buttons_zoom.add(component = btn_zoomout = new PGC_ButtonFixed("-", PGC.component_tiny_button_size));
			pnl_ctl_buttons_zoom.add(component = btn_zoomin = new PGC_ButtonFixed("+", PGC.component_tiny_button_size));
			pnl_ctl_buttons_zoom.add(component = new PGC_LabelFixed(new Dimension(20,-1)));
			pnl_ctl_buttons_zoom.add(component = lbl_coordinate_x = new PGC_LabelFixed("x =", new Dimension(82, -1)));
			pnl_ctl_buttons_zoom.add(component = new PGC_LabelFixed(new Dimension(4,-1)));
			pnl_ctl_buttons_zoom.add(component = lbl_coordinate_y = new PGC_LabelFixed("y =", new Dimension(82, -1)));
			pnl_ctl_buttons_zoom.add(component = new PGC_LabelFixed(new Dimension(20,-1)));
			pnl_ctl_buttons_zoom.add(component = chk_plot = new Checkbox("Plot  ", null, false));
			pnl_ctl_buttons_zoom.add(component = chk_spin = new Checkbox("Spin  ", null, false));
			pnl_ctl_buttons_zoom.add(component = chk_line = new Checkbox("Line  ", null, false));

			pnl_ctl_buttons.add(component = pnl_ctl_buttons_func = PGC.StdPanel());
			((GridBagLayout)(pnl_ctl_buttons.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine(new Insets(3,0,0,0)));

			pnl_ctl_buttons_func.add(component = lbl_f_expr = new PGC_LabelFixed("f = ", new Dimension(22, -1)));
			((GridBagLayout)(pnl_ctl_buttons_func.getLayout())).setConstraints(component, PGC.StdConstraints());
			pnl_ctl_buttons_func.add(component = txt_f_expr = new PGC_TextFieldFixed(this, new Dimension(220, -1)));
			pnl_ctl_buttons_func.add(component =              new PGC_LabelFixed("", new Dimension(12, -1)));
			pnl_ctl_buttons_func.add(component = lbl_g_expr = new PGC_LabelFixed("g = ", new Dimension(26, -1)));
			((GridBagLayout)(pnl_ctl_buttons_func.getLayout())).setConstraints(component, PGC.StdConstraints());
			pnl_ctl_buttons_func.add(component = txt_g_expr = new PGC_TextFieldFixed(this, new Dimension(220, -1)));
			}

		pnl_graph.init();
		pnl_graph.show();
		}

	private void LayoutLaunchButton()
		{
		Component component;
		GridBagConstraints constraints;
		component = pnl_root_launch = PGC.StdPanel();
		pnl_root_launch.add(component = btn_launch = new Button(option_launch_button));
		((GridBagLayout)(pnl_root_launch.getLayout())).setConstraints(component, PGC.StdConstraintsFillMax());
		}

	//----------------------------------------------------------------
	// PGC: Variables
	//----------------------------------------------------------------

	public static final int VAR_A             =  0; // input
	public static final int VAR_B             =  1; // input
	public static final int VAR_C             =  2; // input
	public static final int VAR_D             =  3; // input
	public static final int VAR_F             =  4; // function
	public static final int VAR_G             =  5; // function
	public static final int VAR_H             =  6; // function
	public static final int VAR_K             =  7; // function
	public static final int VAR_M             =  8; // function
	public static final int VAR_N             =  9; // function
	public static final int VAR_PROJECTION_X  = 10; // projection
	public static final int VAR_PROJECTION_Y  = 11; // projection
	public static final int VAR_AUTOPLOT      = 12;
	public static final int VAR_AUTOINCREMENT = 13;
	public static final int VAR_COUNT         = 14;

	private void DefineVariables()
		{
		int index;

		//-------------------------------
		// Input variables
		//-------------------------------
		index = VAR_A;
		variable[index] = new PGC_Variable();
		variable[index].index              = index;
		variable[index].name               = "a";
		variable[index].is_input           = true;
		variable[index].coordinates        = Complex.Number(1.0, 1.0);

		index = VAR_B;
		variable[index] = new PGC_Variable();
		variable[index].index              = index;
		variable[index].name               = "b";
		variable[index].is_input           = true;
		variable[index].coordinates        = Complex.Number(1.125, -0.25);

		index = VAR_C;
		variable[index] = new PGC_Variable();
		variable[index].index              = index;
		variable[index].name               = "c";
		variable[index].is_input           = true;
		variable[index].coordinates        = Complex.Number(-0.5, -1.0);

		index = VAR_D;
		variable[index] = new PGC_Variable();
		variable[index].index              = index;
		variable[index].name               = "d";
		variable[index].is_input           = true;
		variable[index].coordinates        = Complex.Number(-0.25, 0.5);

		//-------------------------------
		// Function variables
		//-------------------------------
		index = VAR_F;
		variable[index] = new PGC_Variable();
		variable[index].index              = index;
		variable[index].name               = "f";
		variable[index].is_function        = true;
		variable[index].expression         = "";

		index = VAR_G;
		variable[index] = new PGC_Variable();
		variable[index].index              = index;
		variable[index].name               = "g";
		variable[index].is_function        = true;
		variable[index].expression         = "";

		index = VAR_H;
		variable[index] = new PGC_Variable();
		variable[index].index              = index;
		variable[index].name               = "h";
		variable[index].is_function        = true;
		variable[index].expression         = "";

		index = VAR_K;
		variable[index] = new PGC_Variable();
		variable[index].index              = index;
		variable[index].name               = "k";
		variable[index].is_function        = true;
		variable[index].expression         = "";

		index = VAR_M;
		variable[index] = new PGC_Variable();
		variable[index].index              = index;
		variable[index].name               = "m";
		variable[index].is_function        = true;
		variable[index].expression         = "";

		index = VAR_N;
		variable[index] = new PGC_Variable();
		variable[index].index              = index;
		variable[index].name               = "n";
		variable[index].is_function        = true;
		variable[index].expression         = "";

		//-------------------------------
		// Projection variables
		//-------------------------------

		index = VAR_PROJECTION_X;
		variable[index] = new PGC_Variable();
		variable[index].index              = index;
		variable[index].name               = "x";

		index = VAR_PROJECTION_Y;
		variable[index] = new PGC_Variable();
		variable[index].index              = index;
		variable[index].name               = "y";

		//-------------------------------
		// The autoplot variable
		//-------------------------------
		index = VAR_AUTOPLOT;
		variable[index] = new PGC_Variable();
		variable[index].index              = index;
		variable[index].name               = ""; // "x" or "t", but it's never labeled
		variable[index].is_autoplot_var    = true;

		//-------------------------------
		// The autoincrement variable
		//-------------------------------
		index = VAR_AUTOINCREMENT;
		variable[index] = new PGC_Variable();
		variable[index].index              = index;
		variable[index].name               = ""; // "s", but it's never labeled
		variable[index].is_autoincr_var    = true;
		variable[index].coordinates        = Complex.Number(0.0, 0.0);
		}

	//----------------------------------------------------------------
	// PGC: Operators
	//----------------------------------------------------------------

	public static final int OP_UNDEFINED  =  0;
	public static final int OP_ADD        =  1;
	public static final int OP_SUB        =  2;
	public static final int OP_MUL        =  3;
	public static final int OP_DIV        =  4;
	public static final int OP_POW        =  5;
	public static final int OP_DOT        =  6;
	public static final int OP_CROSS      =  7;
	public static final int OP_MAT        =  8;
	public static final int OP_SQRT       =  9;
	public static final int OP_CUBERT     = 10;
	public static final int OP_EXP        = 11;
	public static final int OP_USER       = 12;
	public static final int OP_EXAMPLE_1  = 13;
	public static final int OP_EXAMPLE_2  = 14;
	public static final int OP_EXAMPLE_3  = 15;
	public static final int OP_EXAMPLE_4  = 16;
	public static final int OP_EXAMPLE_5  = 17;
	public static final int OP_EXAMPLE_6  = 18;
	public static final int OP_EXAMPLE_7  = 19;
	public static final int OP_EXAMPLE_8  = 20;
	public static final int OP_COUNT      = 21;
	public static final int OP_EXAMPLE_9  = 21; // unused
	public static final int OP_EXAMPLE_10 = 22; // unused

	public static String OpDescription(int op)
		{
		String result;
		switch (op)
			{
			case OP_ADD        : result = "Addition";                  break;
			case OP_SUB        : result = "Subtraction";               break;
			case OP_MUL        : result = "Complex multiplication";    break;
			case OP_DIV        : result = "Complex division";          break;
			case OP_POW        : result = "Complex power";             break;
			case OP_DOT        : result = "Vector dot product";        break;
			case OP_CROSS      : result = "Vector cross product";      break;
			case OP_MAT        : result = "Matrix multiplication";     break;
			case OP_SQRT       : result = "Complex square roots";      break;
			case OP_CUBERT     : result = "Complex cube roots";        break;
			case OP_EXP        : result = "Complex exponential";       break;
			case OP_USER       : result = "User-defined functions";    break;
			case OP_EXAMPLE_1  : result = "Example 1";                 break;
			case OP_EXAMPLE_2  : result = "Example 2";                 break;
			case OP_EXAMPLE_3  : result = "Example 3";                 break;
			case OP_EXAMPLE_4  : result = "Example 4";                 break;
			case OP_EXAMPLE_5  : result = "Example 5";                 break;
			case OP_EXAMPLE_6  : result = "Example 6";                 break;
			case OP_EXAMPLE_7  : result = "Example 7";                 break;
			case OP_EXAMPLE_8  : result = "Example 8";                 break;
			case OP_EXAMPLE_9  : result = "Example 9";                 break;
			case OP_EXAMPLE_10 : result = "Example 10";                break;
			default            : result = "";                          break;
			}
		return (result);
		}
	public static String OpLabel(int op)
		{
		String result;
		switch (op)
			{
			case OP_ADD        : result = "f = a + b";        break;
			case OP_SUB        : result = "f = a - b";        break;
			case OP_MUL        : result = "f = a * b";        break;
			case OP_DIV        : result = "f = a / b";        break;
			case OP_POW        : result = "f = a ^ b";        break;
			case OP_DOT        : result = "f = a  b";        break;
			case OP_CROSS      : result = "f = a  b";        break;
			case OP_MAT        : result = "f = mat(a,b)  c"; break;
			case OP_SQRT       : result = "f = Va";          break;
			case OP_CUBERT     : result = "f = 3Va";         break;
			case OP_EXP        : result = "f = exp a";        break;
			case OP_USER       : result = "user-defined";     break;
			case OP_EXAMPLE_1  : result = "example 1";        break;
			case OP_EXAMPLE_2  : result = "example 2";        break;
			case OP_EXAMPLE_3  : result = "example 3";        break;
			case OP_EXAMPLE_4  : result = "example 4";        break;
			case OP_EXAMPLE_5  : result = "example 5";        break;
			case OP_EXAMPLE_6  : result = "example 6";        break;
			case OP_EXAMPLE_7  : result = "example 7";        break;
			case OP_EXAMPLE_8  : result = "example 8";        break;
			case OP_EXAMPLE_9  : result = "example 9";        break;
			case OP_EXAMPLE_10 : result = "example 10";       break;
			default            : result = "";                 break;
			}
		return (result);
		}
	public static String OpExpression(int op)
		{
		String result;
		switch (op)
			{
			case OP_ADD        : result = "a + b";                break;
			case OP_SUB        : result = "a - b";                break;
			case OP_MUL        : result = "a * b";                break;
			case OP_DIV        : result = "a / b";                break;
			case OP_POW        : result = "a ^ b";                break;
			case OP_DOT        : result = "a dot b";              break;
			case OP_CROSS      : result = "a cross b";            break;
			case OP_MAT        : result = "(a dot c, b dot c)";   break;
			case OP_SQRT       : result = "sqrt a";               break;
			case OP_CUBERT     : result = "a ^ (1/3)";            break;
			case OP_EXP        : result = "exp a";                break;
			default            : result = "";                     break;
			}
		return (result);
		}
	public static String OpChoiceLabel(int op)
		{
		// The choice label looks better with a few spaces around it.

		return ("  " + OpLabel(op) + "  ");
		}
	public static int OpEvalChoiceLabel(String label)
		{
		int result = OP_UNDEFINED;
		for (int op = 0; op < OP_COUNT; op++)
			{
			if (label.equals(OpChoiceLabel(op))) { result = op; break; }
			}
		return (result);
		}
	public static int OpArgCount(int op)
		{
		int result;
		switch (op)
			{
			case OP_ADD     : result = 2;   break;
			case OP_SUB     : result = 2;   break;
			case OP_MUL     : result = 2;   break;
			case OP_DIV     : result = 2;   break;
			case OP_POW     : result = 2;   break;
			case OP_DOT     : result = 2;   break;
			case OP_CROSS   : result = 2;   break;
			case OP_MAT     : result = 3;   break;
			case OP_SQRT    : result = 1;   break;
			case OP_CUBERT  : result = 1;   break;
			case OP_EXP     : result = 1;   break;
			case OP_USER    : result = 0;   break;
			default         : result = 0;   break;
			}
		return (result);
		}

	//----------------------------------------------------------------
	// PGC: Control functions
	//----------------------------------------------------------------

	public static void Trace(Object obj)
		{
		//-----------------------------------------------
		// Write diagnostic information to a separate
		// trace window.
		//
		// The window does not appear until the first
		// trace is written to it.
		//-----------------------------------------------
		if (option_enable_trace)
			{
			if (trace_frame == null)
				{
				trace_frame = new PGC_TextFrame(null, "Trace", 20, 80);
				trace_frame.init();
				trace_frame.show();
				}
			trace_frame.Write(obj);
			}
		}
	public static void AssertionFailure()
		{
		throw new Error("assertion failure");
		}
	public boolean GetLine()
		{
		return (pnl_graph.GetLine());
		}
	public void SetLine(boolean value)
		{
		if (pnl_ctl != null)
			{
			if (chk_line.getState() == !value)
				chk_line.setState(value);
			}
		pnl_graph.SetLine(value);
		}
	public void CheckLine(Boolean value)
		{
		if (value != null)
			SetLine(value.booleanValue());
		else
			SetLine(!GetLine()); // invert it
		}
	public boolean GetSpin()
		{
		return (pnl_graph.GetSpin());
		}
	public void SetSpin(boolean value)
		{
		if (value && !GetEnableSpin())
			{
			// Not enabled -- must set it to false.
			value = false;
			}
		if (pnl_ctl != null)
			{
			if (chk_spin.getState() == !value)
				chk_spin.setState(value);
			}
		pnl_graph.SetSpin(value);
		}
	public void CheckSpin(Boolean value)
		{
		if (value != null)
			SetSpin(value.booleanValue());
		else
			SetSpin(!GetSpin()); // invert it
		}
	public void EnableSpin(boolean value)
		{
		if (pnl_ctl != null)
			{
			chk_spin.enable(value);
			}
		pnl_graph.EnableSpin(value);
		if (!value) SetSpin(false);
		}
	public boolean GetEnableSpin()
		{
		return (pnl_graph.GetEnableSpin());
		}
	public boolean GetPlot()
		{
		return (pnl_graph.GetPlot());
		}
	public void SetPlot(boolean value)
		{
		if (pnl_ctl != null)
			{
			if (chk_plot.getState() == !value)
				chk_plot.setState(value);
			}
		pnl_graph.SetPlot(value);
		}
	public void CheckPlot(Boolean value)
		{
		if (value != null)
			SetPlot(value.booleanValue());
		else
			SetPlot(!GetPlot()); // invert it
		}
	public void Clear()
		{
		SetPlot(false);
		SetAutoincrementVariable(Complex.Number(0.0));
		pnl_graph.PlotClear();
		UpdateDisplay();
		}
	public void AttachDetach()
		{
		if (btn_attach_detach != null)
			{
			if (main_frame == null) { MainFrameDetach(); }
			else                    { MainFrameClose(); }
			}
		}
	public void GraphFocus()
		{
		pnl_graph.requestFocus();
		}
	public void About()
		{
		if (about_frame == null)
			{
			about_frame = new PGC_TextFrame(this, "Plane Graphic Calculator", 10, 50);
			about_frame.init();
			if (about_location != null)
				about_frame.move(about_location.x, about_location.y);
			else
				about_frame.move(50, 50);
			AboutText();
			about_frame.show();
			}
		about_frame.toFront();
		about_frame.show();
		about_frame.requestFocus();
		}
	private void AboutText()
		{
		String text = "";
		text += "\n";
		text += " Plane Graphic Calculator\n";
		text += " Version 1.1\n";
		text += " June 2001\n";
		text += "\n";
		text += " This software is donated to the public,\n";
		text += " and can be used with no restrictions.\n";
		text += "\n";
		about_frame.Write(text);
		about_frame.ShowTop();
		}
	public void Details()
		{
		if (details_frame == null)
			{
			details_frame = new PGC_DetailsFrame(this, "Details");
			details_frame.init();
			UpdateDisplay();
			if (details_location != null)
				details_frame.move(details_location.x, details_location.y);
			else
				details_frame.move(50, 50);
			details_frame.show();
			}
		else
			UpdateDisplay();

		details_frame.toFront();
		details_frame.show();
		details_frame.requestFocus();
		}
	public void FrameDestroy(Frame frame)
		{
		if (frame != null)
			{
			frame.dispose();

			if (frame == about_frame)
				{
				about_frame = null;
				about_location = frame.location(); // remember for next activation
				}
			else if (frame == details_frame)
				{
				details_frame = null;
				details_location = frame.location(); // remember for next activation
				}
			}
		}
	public PGC_Eval GetEval()
		{
		return (eval);
		}
	public void UpdateDisplay()
		{
		int op = GetOp();

		//--------------------------------------------------
		// Look at the user-defined function f to
		// see if it contains anything.  If so, then
		// set the op to OP_USER.
		//--------------------------------------------------
		if (true)
			{
			boolean user_defined = false;
			PGC_Eval eval = GetEval();
			eval.Evaluate(variable[PGC.VAR_F].expression);
			if (!eval.empty)
				{
				user_defined = true;
				}
			if (user_defined && (op > OP_UNDEFINED) && (op < OP_USER))
				{
				op = OP_USER;
				pnl_graph.SetOp(op);
				}
			}

		//--------------------------------------------------
		// Set the op.
		//--------------------------------------------------
		if (pnl_ctl != null)
			{
			if (choice_op.getSelectedItem() != OpChoiceLabel(op))
				choice_op.select(OpChoiceLabel(op));
			}

		if (details_frame != null)
			details_frame.lbl_error_message.setText("");

		//-------------------------------------------
		// Determine which variables are displayed.
		//-------------------------------------------

		for (int index = 0; index < variable.length; index++)
			variable[index].is_displayed = false;

		int arg_count = OpArgCount(op);

		if (arg_count == 1)
			{
			variable[PGC.VAR_A].is_displayed = true;
			variable[PGC.VAR_F].is_displayed = true;
			}
		else if (arg_count == 2)
			{
			variable[PGC.VAR_A].is_displayed = true;
			variable[PGC.VAR_B].is_displayed = true;
			variable[PGC.VAR_F].is_displayed = true;
			}
		else if (arg_count == 3)
			{
			variable[PGC.VAR_A].is_displayed = true;
			variable[PGC.VAR_B].is_displayed = true;
			variable[PGC.VAR_C].is_displayed = true;
			variable[PGC.VAR_F].is_displayed = true;
			}

		if (true)
			{
			//--------------------------------------------------
			// Look at the user-defined functions to see which
			// input variables should be displayed.
			//--------------------------------------------------
			PGC_Eval eval = GetEval();
			String   error_message = "";

			for (int index = 0; index < variable.length; index++)
				{
				if (variable[index].is_function)
					{
					if (index == PGC.VAR_F)
						{
						if ((op > OP_UNDEFINED) && (op < OP_USER))
							continue;  // a built-in function is already using "f"
						}
					eval.Evaluate(variable[index].expression);

					variable[index].is_displayed = (!eval.empty && !eval.uses_autoplot);

					for (int j = 0; j < variable.length; j++)
						{
						if (variable[j].is_input)
							variable[j].is_displayed = variable[j].is_displayed || eval.uses_input[j];
						}
					if ((eval.message.length() > 0) && (error_message.length() == 0))
						error_message = "ERROR in " + variable[index].name + ": " + eval.message;
					}
				}
			if (details_frame != null)
				{
				if (error_message.length() > 0)
					details_frame.lbl_error_message.setText(error_message);
				}

			for (int j = 0; j < variable.length; j++)
				{
				if ((variable[j].is_input) && (variable[j].is_displayed))
					arg_count++;
				}
			}

		//----------------------------------------
		// Enable "Spin" if there are arguments:
		//----------------------------------------
		EnableSpin(arg_count > 0);

		//----------------------------------------
		// Point coordinates in the details frame:
		//----------------------------------------
		if (details_frame != null)
			{
			boolean t;

			t = variable[PGC.VAR_A].is_displayed;
			details_frame.lbl_a.setText(t ? "a" : "");
			details_frame.lbl_a_x.setText(t ? " x= " : "");
			details_frame.txt_a_x.setEditable(t);
			details_frame.txt_a_x.enable(t);
			details_frame.txt_a_x.select(0,0);
			details_frame.lbl_a_y.setText(t ? " y= " : "");
			details_frame.txt_a_y.setEditable(t);
			details_frame.txt_a_y.enable(t);
			details_frame.txt_a_y.select(0,0);

			t = variable[PGC.VAR_B].is_displayed;
			details_frame.lbl_b.setText(t ? "b" : "");
			details_frame.lbl_b_x.setText(t ? " x= " : "");
			details_frame.txt_b_x.setEditable(t);
			details_frame.txt_b_x.enable(t);
			details_frame.txt_b_x.select(0,0);
			details_frame.lbl_b_y.setText(t ? " y= " : "");
			details_frame.txt_b_y.setEditable(t);
			details_frame.txt_b_y.enable(t);
			details_frame.txt_b_y.select(0,0);

			t = variable[PGC.VAR_C].is_displayed;
			details_frame.lbl_c.setText(t ? "c" : "");
			details_frame.lbl_c_x.setText(t ? " x= " : "");
			details_frame.txt_c_x.setEditable(t);
			details_frame.txt_c_x.enable(t);
			details_frame.txt_c_x.select(0,0);
			details_frame.lbl_c_y.setText(t ? " y= " : "");
			details_frame.txt_c_y.setEditable(t);
			details_frame.txt_c_y.enable(t);
			details_frame.txt_c_y.select(0,0);

			t = variable[PGC.VAR_D].is_displayed;
			details_frame.lbl_d.setText(t ? "d" : "");
			details_frame.lbl_d_x.setText(t ? " x= " : "");
			details_frame.txt_d_x.setEditable(t);
			details_frame.txt_d_x.enable(t);
			details_frame.txt_d_x.select(0,0);
			details_frame.lbl_d_y.setText(t ? " y= " : "");
			details_frame.txt_d_y.setEditable(t);
			details_frame.txt_d_y.enable(t);
			details_frame.txt_d_y.select(0,0);

			t = variable[PGC.VAR_F].is_displayed;
			details_frame.lbl_f.setText(t ? "f" : "");
			details_frame.lbl_f_x.setText(t ? " x= " : "");
			details_frame.txt_f_x.setEditable(false);
			details_frame.txt_f_x.enable(t);
			details_frame.txt_f_x.select(0,0);
			details_frame.lbl_f_y.setText(t ? " y= " : "");
			details_frame.txt_f_y.setEditable(false);
			details_frame.txt_f_y.enable(t);
			details_frame.txt_f_y.select(0,0);

			t = variable[PGC.VAR_G].is_displayed;
			details_frame.lbl_g.setText(t ? "g" : "");
			details_frame.lbl_g_x.setText(t ? " x= " : "");
			details_frame.txt_g_x.setEditable(false);
			details_frame.txt_g_x.enable(t);
			details_frame.txt_g_x.select(0,0);
			details_frame.lbl_g_y.setText(t ? " y= " : "");
			details_frame.txt_g_y.setEditable(false);
			details_frame.txt_g_y.enable(t);
			details_frame.txt_g_y.select(0,0);

			t = variable[PGC.VAR_H].is_displayed;
			details_frame.lbl_h.setText(t ? "h" : "");
			details_frame.lbl_h_x.setText(t ? " x= " : "");
			details_frame.txt_h_x.setEditable(false);
			details_frame.txt_h_x.enable(t);
			details_frame.txt_h_x.select(0,0);
			details_frame.lbl_h_y.setText(t ? " y= " : "");
			details_frame.txt_h_y.setEditable(false);
			details_frame.txt_h_y.enable(t);
			details_frame.txt_h_y.select(0,0);

			t = variable[PGC.VAR_K].is_displayed;
			details_frame.lbl_k.setText(t ? "k" : "");
			details_frame.lbl_k_x.setText(t ? " x= " : "");
			details_frame.txt_k_x.setEditable(false);
			details_frame.txt_k_x.enable(t);
			details_frame.txt_k_x.select(0,0);
			details_frame.lbl_k_y.setText(t ? " y= " : "");
			details_frame.txt_k_y.setEditable(false);
			details_frame.txt_k_y.enable(t);
			details_frame.txt_k_y.select(0,0);

			t = variable[PGC.VAR_M].is_displayed;
			details_frame.lbl_m.setText(t ? "m" : "");
			details_frame.lbl_m_x.setText(t ? " x= " : "");
			details_frame.txt_m_x.setEditable(false);
			details_frame.txt_m_x.enable(t);
			details_frame.txt_m_x.select(0,0);
			details_frame.lbl_m_y.setText(t ? " y= " : "");
			details_frame.txt_m_y.setEditable(false);
			details_frame.txt_m_y.enable(t);
			details_frame.txt_m_y.select(0,0);

			t = variable[PGC.VAR_N].is_displayed;
			details_frame.lbl_n.setText(t ? "n" : "");
			details_frame.lbl_n_x.setText(t ? " x= " : "");
			details_frame.txt_n_x.setEditable(false);
			details_frame.txt_n_x.enable(t);
			details_frame.txt_n_x.select(0,0);
			details_frame.lbl_n_y.setText(t ? " y= " : "");
			details_frame.txt_n_y.setEditable(false);
			details_frame.txt_n_y.enable(t);
			details_frame.txt_n_y.select(0,0);
			}

		//--------------------------------
		// Expression -- applet frame:
		//--------------------------------
		if (pnl_ctl != null)
			{
			this         .txt_f_expr.select(0,0);
			this         .txt_g_expr.select(0,0);

			this         .txt_f_expr.setText(variable[PGC.VAR_F].expression);
			this         .txt_g_expr.setText(variable[PGC.VAR_G].expression);

			Complex s = variable[VAR_AUTOINCREMENT].coordinates;
			if (s != null)
				{
				if (s.re < 0.5)
					this.btn_step.setLabel("Step");
				else
					this.btn_step.setLabel("Step " + (int)(s.re + 0.5));
				}
			}

		//--------------------------------
		// Expression -- details frame:
		//--------------------------------
		if (details_frame != null)
			{
			String s;

			details_frame.txt_h_expr.select(0,0);
			details_frame.txt_k_expr.select(0,0);
			details_frame.txt_m_expr.select(0,0);
			details_frame.txt_n_expr.select(0,0);

			details_frame.txt_h_expr.setText(variable[PGC.VAR_H].expression);
			details_frame.txt_k_expr.setText(variable[PGC.VAR_K].expression);
			details_frame.txt_m_expr.setText(variable[PGC.VAR_M].expression);
			details_frame.txt_n_expr.setText(variable[PGC.VAR_N].expression);

			details_frame.txt_t_min.select(0,0);
			details_frame.txt_t_max.select(0,0);
			details_frame.txt_t_rep.select(0,0);

			details_frame.txt_t_min.setText(String.valueOf(this.t_minimum));
			details_frame.txt_t_max.setText(String.valueOf(this.t_maximum));
			details_frame.txt_t_rep.setText(String.valueOf((int)this.t_count));
			}

		pnl_graph.UpdateDisplay(); // update the graph
		}
	public int GetOp()
		{
		if (pnl_graph != null)
			return (pnl_graph.GetOp());
		else
			return (OP_UNDEFINED);
		}
	public void SetOp(int op)
		{
		if ((op > OP_UNDEFINED) && (op < OP_USER))
			{
			//--------------------------------------------------
			// If it's not OP_USER, then clear the
			// defintion for function "f"
			//--------------------------------------------------
			SetVariableExpression(PGC.VAR_F, "");
			}
		if ((op >= OP_EXAMPLE_1) && (op <= OP_EXAMPLE_10))
			{
			for (int index = PGC.VAR_F; index <= PGC.VAR_N; index++)
				{
				SetVariableExpression(index, "");
				}
			SetSpin(false);
			SetPlot(false);
			Clear();
			switch (op)
				{
				case OP_EXAMPLE_1:
					{
					SetVariableExpression(PGC.VAR_F, "(a+b)/2");
					SetLine(false);
					break;
					}
				case OP_EXAMPLE_2:
					{
					SetVariableExpression(PGC.VAR_F, "(cy, cx)");
					SetVariableExpression(PGC.VAR_G, "(-fx, fy)");
					SetLine(true);
					break;
					}
				case OP_EXAMPLE_3:
					{
					SetVariableExpression(PGC.VAR_F, "2x - 1");
					SetVariableExpression(PGC.VAR_G, "2x^2 + 4x + 1");
					break;
					}
				case OP_EXAMPLE_4:
					{
					SetVariableExpression(PGC.VAR_F, "sin(x/ax * pi/2) * ay");
					SetLine(false);
					break;
					}
				case OP_EXAMPLE_5:
					{
					SetVariableExpression(PGC.VAR_F, "(cos t, sin t)");
					SetVariableExpression(PGC.VAR_G, "c + t d");
					SetLine(true);
					break;
					}
				case OP_EXAMPLE_6:
					{
					SetVariableExpression(PGC.VAR_F, "(ax cos t + i ay sin t) * b + c");
					SetLine(true);
					break;
					}
				case OP_EXAMPLE_7:
					{
					SetVariableExpression(PGC.VAR_F, "i^(4s/5)   {click the Step button}");
					SetVariableExpression(PGC.VAR_G, "{click the Step button 5 times}");
					SetLine(false);
					SetPlot(true);
					break;
					}
				case OP_EXAMPLE_8:
					{
					SetVariableExpression(PGC.VAR_F, "i^(8t(ax+(s/2000),ay/100))");
					SetLine(false);
					break;
					}
				case OP_EXAMPLE_9:
					{
					// unused
					break;
					}
				case OP_EXAMPLE_10:
					{
					// unused
					break;
					}
				}
			}
		pnl_graph.SetOp(op);
		UpdateDisplay();
		}
	public void SetZoomDelta(int delta)
		{
		pnl_graph.SetZoomDelta(delta);
		if (pnl_ctl != null)
			{
			String image = "";
			int zoom = GetZoom();
			if (zoom < 10000)
				image = "" + zoom;
			else if (zoom < 10000000)
				image = "" + zoom/1000 + "K";
			else
				image = "" + zoom/1000000 + "M";
			String text = "Zoom: " + image + " X  ";
			lbl_zoom.setText(text);
			}
		}
	public void SetZoom(int scale)
		{
		pnl_graph.SetZoom(scale);
		SetZoomDelta(0);
		}
	public int GetZoom()
		{
		return (pnl_graph.GetZoom());
		}
	public boolean LooksReal(Complex c)
		{
		//-----------------------------------------------
		// Return true if c "looks real enough" to be
		// plotted on an x-axis pixel.
		//-----------------------------------------------
		return (Math.abs(c.im) < (0.5 / (double)GetZoom()));
		}
	public boolean LooksEqual(Complex c, Complex d)
		{
		//-----------------------------------------------
		// Return true if c and d plot on the same
		// pixel.
		//-----------------------------------------------
		double re = Math.abs(c.re - d.re);
		double im = Math.abs(c.im - d.im);
		double err = (0.5 / (double)GetZoom());
		return ((re < err) && (im < err));
		}
	public void SetCenterDelta(int delta_x, int delta_y)
		{
		pnl_graph.SetCenterDelta(delta_x, delta_y);
		}
	public void SetCenter(Complex c)
		{
		pnl_graph.SetCenter(c);
		}
	public void SetActivePoint(int var)
		{
		pnl_graph.SetActivePoint(var);
		}
	public void DisplayCoordinates(Complex c)
		{
		if (pnl_ctl != null)
			{
			if (c == null)
				{
				lbl_coordinate_x.setText("x = ");
				lbl_coordinate_y.setText("y = ");
				}
			else
				{
				int zoom = GetZoom();
				int round_scale;
				double log_10 = Math.log(10.0);
				double log_zoom = Math.log(zoom) / log_10;
				double d_round_scale = Math.exp((Math.floor(log_zoom) + 2.0) * log_10);
				d_round_scale = Math.floor(d_round_scale + 0.5);
				if (d_round_scale > 1000000000.0)
					round_scale = 1000000000;
				else
					round_scale = (int)d_round_scale;

				double re_sign = (c.re > 0.0) ? 1.0 : (c.re < 0.0) ? -1.0 : 0.0;
				double im_sign = (c.im > 0.0) ? 1.0 : (c.im < 0.0) ? -1.0 : 0.0;
				double re = re_sign * (Math.floor(Math.abs(c.re) * round_scale + 0.5) / round_scale);
				double im = im_sign * (Math.floor(Math.abs(c.im) * round_scale + 0.5) / round_scale);
				lbl_coordinate_x.setText("x = " + String.valueOf(re));
				lbl_coordinate_y.setText("y = " + String.valueOf(im));
				}
			}
		}
	public Complex GetVariableCoordinates(int index)
		{
		Complex result = null;
		if (variable[index].is_displayed)
			result = variable[index].coordinates;
		return (result);
		}
	public void SetVariableCoordinates(int index, Complex z)
		{
		if (variable[index].is_displayed)
			variable[index].coordinates = z;
		}
	public String GetVariableExpression(int index)
		{
		String result = variable[index].expression;

		if ((GetOp() > OP_UNDEFINED) && (GetOp() < OP_USER) && (index == PGC.VAR_F))
			{
			result = OpExpression(GetOp()); // built-in
			}
		return (result);
		}
	public void GetVariableEvaluation(int index, PGC_Eval e)
		{
		Node node = variable[index].node;
		if (node != null)
			{
			e.Evaluate(node);
			}
		else
			{
			// Inefficient -- hopefully this isn't used much:
			String expression = GetVariableExpression(index);
			e.Evaluate(expression);
			}
		}
	public void SetVariableExpression(int index, String expression)
		{
		variable[index].expression = expression;
		eval.Evaluate(expression);
		variable[index].node = eval.node;
		}
	public Complex GetAutoplotVariable()
		{
		return (variable[PGC.VAR_AUTOPLOT].coordinates);
		}
	public void SetAutoplotVariable(Complex c)
		{
		variable[PGC.VAR_AUTOPLOT].coordinates = c;
		}
	public Complex GetAutoincrementVariable()
		{
		return (variable[PGC.VAR_AUTOINCREMENT].coordinates);
		}
	public void SetAutoincrementVariable(Complex c)
		{
		variable[PGC.VAR_AUTOINCREMENT].coordinates = c;
		}
	public void Step(int count)
		{
		if (pnl_graph.Painting())
			{
			// Ignore while painting.
			}
		else
			{
			for (int k = 0; k < count; k++)
				{
				// Set a = f (for iterative functions)
				if (variable[PGC.VAR_F].is_displayed)
					SetVariableCoordinates(PGC.VAR_A, variable[PGC.VAR_F].coordinates);

				// Increment the autoincrement variable
				Complex s = GetAutoincrementVariable();
				SetAutoincrementVariable(Complex.Number(s.re+1));

				pnl_graph.PaintNow(); // forcibly update the graph
				}
			UpdateDisplay();
			}
		}
	public void TextFieldsInput(PGC_TextFieldFixed field)
		{
		if (details_frame != null)
			{
			Complex a = Complex.Number();
			Complex b = Complex.Number();
			Complex c = Complex.Number();
			Complex d = Complex.Number();

			eval.Evaluate(details_frame.txt_a_x.getText()); if (eval.value != null) a.re = eval.value.re;
			eval.Evaluate(details_frame.txt_a_y.getText()); if (eval.value != null) a.im = eval.value.re;
			eval.Evaluate(details_frame.txt_b_x.getText()); if (eval.value != null) b.re = eval.value.re;
			eval.Evaluate(details_frame.txt_b_y.getText()); if (eval.value != null) b.im = eval.value.re;
			eval.Evaluate(details_frame.txt_c_x.getText()); if (eval.value != null) c.re = eval.value.re;
			eval.Evaluate(details_frame.txt_c_y.getText()); if (eval.value != null) c.im = eval.value.re;
			eval.Evaluate(details_frame.txt_d_x.getText()); if (eval.value != null) d.re = eval.value.re;
			eval.Evaluate(details_frame.txt_d_y.getText()); if (eval.value != null) d.im = eval.value.re;

			SetVariableCoordinates(PGC.VAR_A, a);
			SetVariableCoordinates(PGC.VAR_B, b);
			SetVariableCoordinates(PGC.VAR_C, c);
			SetVariableCoordinates(PGC.VAR_D, d);

			SetVariableExpression(PGC.VAR_F, this         .txt_f_expr.getText());
			SetVariableExpression(PGC.VAR_G, this         .txt_g_expr.getText());
			SetVariableExpression(PGC.VAR_H, details_frame.txt_h_expr.getText());
			SetVariableExpression(PGC.VAR_K, details_frame.txt_k_expr.getText());
			SetVariableExpression(PGC.VAR_M, details_frame.txt_m_expr.getText());
			SetVariableExpression(PGC.VAR_N, details_frame.txt_n_expr.getText());

			eval.Evaluate(details_frame.txt_t_min.getText()); if (eval.value != null) this.t_minimum     = eval.value.re;
			eval.Evaluate(details_frame.txt_t_max.getText()); if (eval.value != null) this.t_maximum     = eval.value.re;
			eval.Evaluate(details_frame.txt_t_rep.getText()); if (eval.value != null) this.t_count       = (int)eval.value.re;
			}

		//------------------
		// applet frame:
		//------------------
		if (pnl_ctl != null)
			{
			SetVariableExpression(PGC.VAR_F, this         .txt_f_expr.getText());
			SetVariableExpression(PGC.VAR_G, this         .txt_g_expr.getText());
			}

		UpdateDisplay();
		SetZoomDelta(0);
		}
	public void TextFieldsOutput(Complex a, Complex b, Complex c, Complex d, Complex f, Complex g, Complex h, Complex k, Complex m, Complex n)
		{
		if (details_frame != null)
			{
			PGC_TextFieldFixed tx;
			PGC_TextFieldFixed ty;

			tx = details_frame.txt_a_x; ty = details_frame.txt_a_y; if (a == null) { tx.setText(""); ty.setText(""); } else { tx.setText(String.valueOf(a.re)); ty.setText(String.valueOf(a.im)); }
			tx = details_frame.txt_b_x; ty = details_frame.txt_b_y; if (b == null) { tx.setText(""); ty.setText(""); } else { tx.setText(String.valueOf(b.re)); ty.setText(String.valueOf(b.im)); }
			tx = details_frame.txt_c_x; ty = details_frame.txt_c_y; if (c == null) { tx.setText(""); ty.setText(""); } else { tx.setText(String.valueOf(c.re)); ty.setText(String.valueOf(c.im)); }
			tx = details_frame.txt_d_x; ty = details_frame.txt_d_y; if (d == null) { tx.setText(""); ty.setText(""); } else { tx.setText(String.valueOf(d.re)); ty.setText(String.valueOf(d.im)); }
			tx = details_frame.txt_f_x; ty = details_frame.txt_f_y; if (f == null) { tx.setText(""); ty.setText(""); } else { tx.setText(String.valueOf(f.re)); ty.setText(String.valueOf(f.im)); }
			tx = details_frame.txt_g_x; ty = details_frame.txt_g_y; if (g == null) { tx.setText(""); ty.setText(""); } else { tx.setText(String.valueOf(g.re)); ty.setText(String.valueOf(g.im)); }
			tx = details_frame.txt_h_x; ty = details_frame.txt_h_y; if (h == null) { tx.setText(""); ty.setText(""); } else { tx.setText(String.valueOf(h.re)); ty.setText(String.valueOf(h.im)); }
			tx = details_frame.txt_k_x; ty = details_frame.txt_k_y; if (k == null) { tx.setText(""); ty.setText(""); } else { tx.setText(String.valueOf(k.re)); ty.setText(String.valueOf(k.im)); }
			tx = details_frame.txt_m_x; ty = details_frame.txt_m_y; if (m == null) { tx.setText(""); ty.setText(""); } else { tx.setText(String.valueOf(m.re)); ty.setText(String.valueOf(m.im)); }
			tx = details_frame.txt_n_x; ty = details_frame.txt_n_y; if (n == null) { tx.setText(""); ty.setText(""); } else { tx.setText(String.valueOf(n.re)); ty.setText(String.valueOf(n.im)); }
			}
		}
	public Image LoadBackgroundImage()
		{
		Image result = null;
		if ((option_background_image != null) && (option_background_image.length() > 0))
			result = getImage(getCodeBase(), option_background_image);
		return (result);
		}
	private void SetOptions()
		{
		//-----------------------------------
		// Initial control settings
		//-----------------------------------
		SetOp(option_op);
		SetCenter(option_center);
		SetZoom(option_zoom);
		SetLine(option_line);
		SetPlot(option_plot);
		SetSpin(option_spin);
		SetActivePoint(option_active);
		}

	//----------------------------------------------------------------
	// PGC: External entry points
	//----------------------------------------------------------------

	public static void main(String args[])
		{
		//--------------------------------
		// For stand-alone applet only.
		//--------------------------------
		PGC pgc = new PGC();

		if (true)
			{
			pgc.main_args = args;
			pgc.stand_alone = true;
			pgc.init();

			Dimension dim = new Dimension(pgc.option_launch_size_x, pgc.option_launch_size_y);
			pgc.resize(dim);

			pgc.start();
			pgc.MainFrameDetach();
			}

		//// if (false)
		////	{
		////	// Stand-alone testing for Eval...
		////	PGC_Eval e = new PGC_Eval(pgc);
		////	e.Evaluate(args[0]);
		////	if (e.value == null)
		////		System.out.println("value is null");
		////	else
		////		{
		////		System.out.print(e.value.re);
		////		System.out.print(" , ");
		////		System.out.println(e.value.im);
		////		}
		////	System.out.println(e.message);
		////	}
		}
	public void init()
		{
		//-------------------------------------------------------------------
		// Destroy all other frames (this is an issue for browser refresh).
		// (However, don't destroy the trace_frame, since it's helpful
		// to see a continuous trace window.)
		//-------------------------------------------------------------------
		MainFrameDestroy();

		//-----------------------
		// Parameters
		//-----------------------
		String param;
		PGC_Eval eval = GetEval();

		param = getParameter("enable_trace");
		if (param != null) option_enable_trace = (PGC_Utility.IValue(param) != 0);

		param = getParameter("launch_button");
		if (param != null) option_launch_button = new String(param);
		launch_mode = ((option_launch_button != null) && (option_launch_button.length() > 0));

		param = getParameter("launch_size");
		if (param != null) { eval.Evaluate(param); if (eval.value != null) { option_launch_size_x = (int)eval.value.re; option_launch_size_y = (int)eval.value.im; } }

		param = getParameter("background_image");
		if (param != null) option_background_image = new String(param);

		param = getParameter("control_panel");
		if (param != null) option_control_panel = (PGC_Utility.IValue(param) != 0);

		param = getParameter("graph_border");
		if (param != null) option_graph_border = (PGC_Utility.IValue(param) != 0);

		param = getParameter("builtin_examples");
		if (param != null) option_builtin_examples = (PGC_Utility.IValue(param) != 0);

		param = getParameter("builtin_functions");
		if (param != null) option_builtin_functions = (PGC_Utility.IValue(param) != 0);
		if (!option_builtin_functions) option_op = OP_USER;

		param = getParameter("center");
		if (param != null) { eval.Evaluate(param); if (eval.value != null) option_center = eval.value; }

		param = getParameter("zoom");
		if (param != null) { eval.Evaluate(param); if (eval.value != null) option_zoom = (int)eval.value.re; }

		param = getParameter("line");
		if (param != null) option_line = (PGC_Utility.IValue(param) != 0);

		param = getParameter("plot");
		if (param != null) option_plot = (PGC_Utility.IValue(param) != 0);

		param = getParameter("spin");
		if (param != null) option_spin = (PGC_Utility.IValue(param) != 0);

		param = getParameter("active");
		if (param != null)
			{
			if (param.equalsIgnoreCase("a")) option_active = PGC.VAR_A;
			if (param.equalsIgnoreCase("b")) option_active = PGC.VAR_B;
			if (param.equalsIgnoreCase("c")) option_active = PGC.VAR_C;
			if (param.equalsIgnoreCase("d")) option_active = PGC.VAR_D;
			}

		param = getParameter("t_minimum");
		if (param != null) { eval.Evaluate(param); if (eval.value != null) t_minimum = eval.value.re; }

		param = getParameter("t_maximum");
		if (param != null) { eval.Evaluate(param); if (eval.value != null) t_maximum = eval.value.re; }

		param = getParameter("t_count");
		if (param != null) { eval.Evaluate(param); if (eval.value != null) t_count = (int)eval.value.re; }

		//-----------------------
		// Core initialization
		//-----------------------
		DefineVariables();

		//---------------------------------------------------
		// Initialize variables from program parameters
		//---------------------------------------------------
			{
			for (int index = 0; index < variable.length; index++)
				{
				if (variable[index].is_input)
					{
					param = getParameter(variable[index].name);
					if (param != null)
						{
						String expression = param;
						eval.Evaluate(expression);
						if (eval.value != null)
							variable[index].coordinates = eval.value;
						}
					}
				else if (variable[index].is_function)
					{
					param = getParameter(variable[index].name);
					if (param != null)
						{
						SetVariableExpression(index, param);
						if ((index == PGC.VAR_F) && (param.length() > 0))
							option_op = OP_USER;
						}
					}
				}
			}
		if (launch_mode)
			{
			//----------------------------------
			// Display only a launch button
			//----------------------------------
			LayoutLaunchButton(); // defines pnl_root_launch
			setLayout(new BorderLayout());
			add("Center", pnl_root_launch);
			}
		else
			{
			//----------------------------------
			// Display the entire calculator
			//----------------------------------
			LayoutCalculator(); // defines pnl_root_calculator
			setLayout(new BorderLayout());
			add("Center", pnl_root_calculator);
			SetOptions();
			}
		}
	public void start()
		{
		}
	public void stop()
		{
		MainFrameDestroy();
		}
	public boolean action(Event evt, Object arg)
		{
		if (evt.id == evt.ACTION_EVENT)
			{
			if (evt.target == btn_details)         { Details();                   return (true); }
			if (evt.target == btn_step)            { Step(1);                     return (true); }
			if (evt.target == btn_plot_clear)      { Clear();                 return (true); }
			if (evt.target == btn_about)           { About();                     return (true); }
			if (evt.target == btn_attach_detach)   { AttachDetach();              return (true); }
			if (evt.target == btn_launch)          { MainFrameLaunch();           return (true); }
			if (evt.target == btn_zoomout)         { SetZoomDelta(-1);            return (true); }
			if (evt.target == btn_zoomin)          { SetZoomDelta(1);             return (true); }
			if (evt.target == chk_line)            { CheckLine((Boolean)evt.arg); return (true); }
			if (evt.target == chk_spin)            { CheckSpin((Boolean)evt.arg); return (true); }
			if (evt.target == chk_plot)            { CheckPlot((Boolean)evt.arg); return (true); }
			}
		return (super.action(evt, arg));
		}
	public boolean mouseDown(Event evt, int x, int y)
		{
		boolean result = super.mouseDown(evt, x, y);
		if (evt.target.getClass().toString().equals("class java.awt.Panel"))
			GraphFocus();
		return (result);
		}
	public boolean keyDown(Event evt, int key)
		{
		boolean result;
		if ((key == evt.ENTER) || (key == evt.ESCAPE))
			{GraphFocus(); result = true;}
		else if (evt.target.getClass().toString().equals("class PGC_ButtonFixed"))
			result = pnl_graph.keyDown(evt, key);
		else if (evt.target.getClass().toString().equals("class java.awt.Checkbox"))
			result = pnl_graph.keyDown(evt, key);
		else if (evt.target.getClass().toString().equals("class java.awt.Panel"))
			result = pnl_graph.keyDown(evt, key);
		else
			result = super.keyDown(evt, key);
		return (result);
		}
	public String getParameter(String name)
		{
		String result;
		if (stand_alone)
			{
			// FUTURE: Get parameters from main_args.
			result = null;
			}
		else
			result = super.getParameter(name);
		return (result);
		}
	}



//========================================================================
//========================================================================
// PGC graph panel
//========================================================================
//========================================================================

class PGC_GraphPanel extends Panel implements Runnable
	{
	private PGC                  owner;
	private Image                offscreen_img;
	private Graphics             offscreen_graphics;
	private Dimension            offscreen_size;
	private Image                background_img;
	private boolean              animation_enabled    = false;
	private Thread               animation_thread;
	private boolean              painting             = false;
	private Point                graphsize            = new Point(0, 0);
	private Point                base                 = new Point(0, 0);
	private Complex              center               = Complex.Number(0);
	private int                  op                   = PGC.OP_UNDEFINED;
	private int                  zoom                 = 1;
	private int                  tracking_var         = -1;     // if >= 0:    mouse button is down and is moving an input variable
	private Complex              tracking_scroll      = null;   // if != null: mouse button is down and is scrolling the graph
	private int                  current_var          = 0;
	private boolean              origin_connector     = false;
	private boolean              plotting             = false;
	private int                  font_char_width      = 0;
	private int                  font_char_height     = 0;
	private double               font_mid_distance    = 0.0;

	public PGC_GraphPanel(PGC own)
		{
		owner = own;

		for (int index = 0; index < owner.variable.length; index++)
			{
			if (owner.variable[index].is_function)
				{
				owner.variable[index].plotting_vec = new Vector();
				}
			}
		background_img = owner.LoadBackgroundImage();
		}

	//----------------------------------------------------------------
	// PGC graph panel: Control functions
	//----------------------------------------------------------------

	public boolean GetLine()
		{
		return (origin_connector);
		}
	public void SetLine(boolean value)
		{
		origin_connector = value;
		repaint();
		}
	public boolean GetSpin()
		{
		return (animation_thread != null);
		}
	public void SetSpin(boolean value)
		{
		if (value)
			{
			if (animation_thread == null)
				{
				animation_thread = new Thread(this);
				animation_thread.start();
				}
			}
		else
			{
			if (animation_thread != null)
				{
				animation_thread.stop();
				animation_thread = null;
				repaint();
				}
			}
		}
	public void EnableSpin(boolean value)
		{
		animation_enabled = value;
		}
	public boolean GetEnableSpin()
		{
		return (animation_enabled);
		}
	public boolean GetPlot()
		{
		return (plotting);
		}
	public void SetPlot(boolean value)
		{
		if (!value)
			PlotBreak();
		plotting = value;
		repaint();
		}
	public void PlotBreak()
		{
		if (plotting)
			{
			PlotAdd(PGC.VAR_F, null); // the next plot point will be disconnected
			PlotAdd(PGC.VAR_G, null); // the next plot point will be disconnected
			PlotAdd(PGC.VAR_H, null); // the next plot point will be disconnected
			PlotAdd(PGC.VAR_K, null); // the next plot point will be disconnected
			PlotAdd(PGC.VAR_M, null); // the next plot point will be disconnected
			PlotAdd(PGC.VAR_N, null); // the next plot point will be disconnected
			}
		}
	public void PlotAdd(int index, Complex c)
		{
		owner.variable[index].plotting_vec.addElement(c);
		}
	public void PlotClear()
		{
		for (int index = 0; index < owner.variable.length; index++)
			{
			if (owner.variable[index].is_function)
				{
				owner.variable[index].plotting_vec = null;
				owner.variable[index].plotting_vec = new Vector();
				}
			}
		repaint();
		}
	private int FindDisplayableInputVariable(int index)
		{
		boolean found = false;

		for (int j = 0; j < owner.variable.length; j++)
			{
			if (index >= owner.variable.length)
				index = 0;
			if ((owner.variable[index].is_input) && (owner.variable[index].is_displayed))
				{
				found = true;
				break;
				}
			index = index + 1;
			}
		if (!found)
			index = 0;
		return (index);
		}
	public void UpdateDisplay()
		{
		current_var = FindDisplayableInputVariable(current_var);
		SetZoomDelta(0); // Force a paint
		}
	public void PaintNow()
		{
		paint(getGraphics());
		}
	public int GetOp()
		{
		return (op);
		}
	public void SetOp(int new_op)
		{
		op = new_op;
		current_var = FindDisplayableInputVariable(current_var);
		PlotBreak();
		repaint();
		}
	public int GetZoom()
		{
		return (zoom);
		}
	public void SetZoomDelta(int delta)
		{
		//---------------------------------------------
		// delta>1 is zoom-in  (increase zoom)
		// delta<1 is zoom-out (decrease zoom)
		// delta=0 repaints
		//
		// The resulting zoom must be a positive
		// integer.
		//---------------------------------------------
		double dz = (double)zoom;

		if      (dz <=         10) {dz +=             delta;}
		else if (dz <=         40) {dz +=         2 * delta;}
		else if (dz <=        100) {dz +=         4 * delta;}
		else if (dz <=        400) {dz +=        10 * delta;}
		else if (dz <=       1000) {dz +=        40 * delta;}
		else if (dz <=       4000) {dz +=       100 * delta;}
		else if (dz <=      10000) {dz +=       400 * delta;}
		else if (dz <=      40000) {dz +=      1000 * delta;}
		else if (dz <=     100000) {dz +=      4000 * delta;}
		else if (dz <=     400000) {dz +=     10000 * delta;}
		else if (dz <=    1000000) {dz +=     40000 * delta;}
		else if (dz <=    4000000) {dz +=    100000 * delta;}
		else if (dz <=   10000000) {dz +=    400000 * delta;}
		else if (dz <=   40000000) {dz +=   1000000 * delta;}
		else if (dz <=  100000000) {dz +=   4000000 * delta;}
		else if (dz <=  400000000) {dz +=  10000000 * delta;}
		else if (dz <= 1000000000) {dz +=  40000000 * delta;}
		else                       {dz += 100000000 * delta;}

		int z = 1;
		if (dz < 1.0)
			z = 1;
		else if (dz > 2000000000.0)
			z = 2000000000;
		else
			z = (int)(dz + 0.5);

		if      (z <           1) {z = 1; }
		else if (z <          10) { }
		else if (z <          40) {z = (z +        1) /         2 *         2;}
		else if (z <         100) {z = (z +        3) /         4 *         4;}
		else if (z <         400) {z = (z +        9) /        10 *        10;}
		else if (z <        1000) {z = (z +       39) /        40 *        40;}
		else if (z <        4000) {z = (z +       99) /       100 *       100;}
		else if (z <       10000) {z = (z +      399) /       400 *       400;}
		else if (z <       40000) {z = (z +      999) /      1000 *      1000;}
		else if (z <      100000) {z = (z +     3999) /      4000 *      4000;}
		else if (z <      400000) {z = (z +     9999) /     10000 *     10000;}
		else if (z <     1000000) {z = (z +    39999) /     40000 *     40000;}
		else if (z <     4000000) {z = (z +    99999) /    100000 *    100000;}
		else if (z <    10000000) {z = (z +   399999) /    400000 *    400000;}
		else if (z <    40000000) {z = (z +   999999) /   1000000 *   1000000;}
		else if (z <   100000000) {z = (z +  3999999) /   4000000 *   4000000;}
		else if (z <   400000000) {z = (z +  9999999) /  10000000 *  10000000;}
		else if (z <  1000000000) {z = (z + 39999999) /  40000000 *  40000000;}
		else                      {z = (z + 99999999) / 100000000 * 100000000;}

		if (z < 1) {z = 1;} // in case of overflow to negative

		zoom = z;
		repaint();
		}
	public void SetZoom(int scale)
		{
		zoom = scale;
		SetZoomDelta(0);
		}
	public void SetCenterDelta(int delta_x, int delta_y)
		{
		//---------------------------------------------
		// Move center by (delta_x, delta_y) pixels.
		//---------------------------------------------
		Point center_point = new Point(base.x + delta_x, base.y + delta_y);
		center = GetComplex(center_point);
		repaint();
		}
	public void SetCenter(Complex c)
		{
		if (c == null)
			c = Complex.Number(0.0, 0.0);
		center = c;
		repaint();
		}
	public void SetActivePoint(int var)
		{
		current_var = FindDisplayableInputVariable(var);
		}
	public void NextActivePoint()
		{
		current_var = FindDisplayableInputVariable(current_var + 1);
		}
	public boolean Painting()
		{
		return (painting);
		}

	//----------------------------------------------------------------
	// PGC graph panel: Support functions
	//----------------------------------------------------------------

	private final int PIXMAX = 0x7FF0;
	private final int INTMAX = 0x7FFFFFFE;

	private int GetXOrigin()
		{
		double x = ((double)base.x - Math.floor((center.re * (double)zoom) + 0.5));

		if (x < (double)-INTMAX) x = (double)-INTMAX;
		if (x > (double) INTMAX) x = (double) INTMAX;

		return ((int)x);
		}
	private int GetYOrigin()
		{
		double y = ((double)base.y + Math.floor((center.im * (double)zoom) + 0.5));

		if (y < (double)-INTMAX) y = (double)-INTMAX;
		if (y > (double) INTMAX) y = (double) INTMAX;

		return ((int)y);
		}
	private void ComplexToPoint(Complex c, Point p)
		{
		double x_origin = (double)base.x - Math.floor((center.re * (double)zoom) + 0.5);
		double y_origin = (double)base.y + Math.floor((center.im * (double)zoom) + 0.5);
		double c_x = Math.floor((c.re * (double)zoom) + 0.5);
		double c_y = Math.floor((c.im * (double)zoom) + 0.5);
		double x = x_origin + c_x;
		double y = y_origin - c_y;

		//--------------------------------------------
		// If the point is outside the allowed
		// area (-INTMAX to INTMAX), then substitute
		// a point that's on the line defined by
		// vector c.  This way, the origin_connector
		// line will point in the correct direction
		// for a point like (1e99,2e99).
		//--------------------------------------------

		if (x < (double)-INTMAX) { x = (double)-INTMAX; if (c_x != 0.0) y = y_origin - (c_y * ((double)-INTMAX - x_origin) / c_x); }
		if (x > (double) INTMAX) { x = (double) INTMAX; if (c_x != 0.0) y = y_origin - (c_y * ((double) INTMAX - x_origin) / c_x); }
		if (y < (double)-INTMAX) { y = (double)-INTMAX; if (c_y != 0.0) x = x_origin + (c_x * (y_origin - (double)-INTMAX) / c_y); }
		if (y > (double) INTMAX) { y = (double) INTMAX; if (c_y != 0.0) x = x_origin + (c_x * (y_origin - (double) INTMAX) / c_y); }

		p.x = (int)x;
		p.y = (int)y;
		}
	private void PointToComplex(Point p, Complex c)
		{
		double x_origin = (double)base.x - Math.floor((center.re * (double)zoom) + 0.5);
		double y_origin = (double)base.y + Math.floor((center.im * (double)zoom) + 0.5);
		c.re = (double)(p.x - x_origin) / (double)zoom;
		c.im = (double)(y_origin - p.y) / (double)zoom;
		}
	private Point GetPoint(Complex c)
		{
		Point p = new Point(0,0);
		ComplexToPoint(c, p);
		return (p);
		}
	private Complex GetComplex(Point p)
		{
		Complex c = new Complex();
		PointToComplex(p, c);
		return (c);
		}
	private boolean PointVisible(Point p)
		{
		boolean result = (p.x >= 0) && (p.y >= 0) && (p.x < graphsize.x) && (p.y < graphsize.y);
		return (result);
		}
	private int PixelLimit(int px, int py, int qx, int qy, int new_px)
		{
		//----------------------------------------------------
		// px will be changed to new_px, so return a
		// value for py that preserves the slope of the line
		// from p to q.
		//----------------------------------------------------
		double y = py;
		double cx = (double)qx - (double)px;
		double cy = (double)qy - (double)py;
		if (cx != 0.0) y = (double)qy - (cy * ((double)qx - (double)new_px) / cx);
		int result = ((int)Math.floor(y+0.5));
		return (result);
		}
	private void DrawLine(int px, int py, int qx, int qy)
		{
		//--------------------------------------------
		// If a point is outside the allowed
		// area (-PIXMAX to PIXMAX), then substitute
		// a point that will draw the same line.
		//--------------------------------------------
		if (px < -PIXMAX) { py = PixelLimit(px, py, qx, qy, -PIXMAX); px = -PIXMAX; }
		if (px >  PIXMAX) { py = PixelLimit(px, py, qx, qy,  PIXMAX); px =  PIXMAX; }
		if (py < -PIXMAX) { px = PixelLimit(py, px, qy, qx, -PIXMAX); py = -PIXMAX; }
		if (py >  PIXMAX) { px = PixelLimit(py, px, qy, qx,  PIXMAX); py =  PIXMAX; }
		if (qx < -PIXMAX) { qy = PixelLimit(qx, qy, px, py, -PIXMAX); qx = -PIXMAX; }
		if (qx >  PIXMAX) { qy = PixelLimit(qx, qy, px, py,  PIXMAX); qx =  PIXMAX; }
		if (qy < -PIXMAX) { qx = PixelLimit(qy, qx, py, px, -PIXMAX); qy = -PIXMAX; }
		if (qy >  PIXMAX) { qx = PixelLimit(qy, qx, py, px,  PIXMAX); qy =  PIXMAX; }

		offscreen_graphics.drawLine(px, py, qx, qy);
		}
	private void DrawRect(int x, int y, int width, int height)
		{
		if ((x >= -PIXMAX) && (x <= PIXMAX) && (y >= -PIXMAX) && (y <= PIXMAX))
			offscreen_graphics.drawRect(x, y, width, height);
		}
	private void FillRect(int x, int y, int width, int height)
		{
		if ((x >= -PIXMAX) && (x <= PIXMAX) && (y >= -PIXMAX) && (y <= PIXMAX))
			offscreen_graphics.fillRect(x, y, width, height);
		}
	private void DrawString(String text, int x, int y)
		{
		if ((x >= -PIXMAX) && (x <= PIXMAX) && (y >= -PIXMAX) && (y <= PIXMAX))
			offscreen_graphics.drawString(text, x, y);
		}
	private void DrawImage(Image img, int x, int y, java.awt.image.ImageObserver observer)
		{
		if ((x >= -PIXMAX) && (x <= PIXMAX) && (y >= -PIXMAX) && (y <= PIXMAX))
			offscreen_graphics.drawImage(img, x, y, observer);
		}
	private int GetVariableCoordinatesIndex(Complex c)
		{
		//---------------------------------------------------------------
		// Find the input variable closest to the specified complex
		//---------------------------------------------------------------
		int     closest_index = -1;
		double  closest_dist  = 0.0;

		for (int index = 0; index < owner.variable.length; index++)
			{
			if ((owner.variable[index].is_input) && (owner.variable[index].is_displayed))
				{
				Complex diff = Complex.Sub(owner.variable[index].coordinates, c);
				double dist = (diff.re * diff.re) + (diff.im * diff.im);
				double pixel_dist_sq = (dist * (double)(zoom * zoom));
				if (pixel_dist_sq > 400.0) continue; // Don't select the dot unless the mouse is within a 20-pixel radius.
				if ((closest_index == -1) || (dist < closest_dist))
					{
					closest_index = index;
					closest_dist = dist;
					}
				}
			}

		return (closest_index);
		}
	private void GetFontInfo()
		{
		if (font_char_height == 0)
			{
			char ch = 'e'; // a typical character
			FontMetrics font_metrics = offscreen_graphics.getFontMetrics();
			font_char_height = font_metrics.getAscent();
			font_char_width = font_metrics.charWidth(ch);
			font_char_height -= (font_char_height >> 2); // adjustment for average height of "a" and "b"
			font_mid_distance = Math.sqrt((font_char_height * font_char_height) + (font_char_width * font_char_width)) / 2; // distance from base to mid-point of letter
			}
		}
	private void DrawPointLabel(PGC_Variable variable, Color color)
		{
		//---------------------------------
		// Draw text to label the point.
		//---------------------------------
		Complex c = variable.coordinates;

		if (c != null)
			{
			String text = variable.name;

			char ch = ' ';
			if (text.length() >= 1)
				ch = text.charAt(0);

			GetFontInfo(); // initialize the font_ variables

			int standoff = 6; // pixel distance between point and label
			double c_length = Complex.Abs(c);
			Complex offset = Complex.Number();
			if (c_length == 0.0)
				{
				offset.re = (1.0 / zoom) * (font_mid_distance + standoff);
				offset.im = 0;
				}
			else
				{
				offset.re = (c.re / c_length) * (1.0 / zoom) * (font_mid_distance + standoff);
				offset.im = (c.im / c_length) * (1.0 / zoom) * (font_mid_distance + standoff);
				}

			//------------------------------------------------------
			// Look at previously-drawn variables and see if this
			// would draw on top of them.  If so, rotate the label
			// offset.
			//
			// FUTURE: Needs more work!  For now, I'm just rotating
			// the label if two variables share the same pixel.
			// This rotation is intended to serve as a visual cue
			// for people who want to line up two points on top
			// of each other.  Upon experimenting, I found that
			// doing sophisticated label-rotation resulted in
			// a lot of extra motion that was quite distracting.
			//------------------------------------------------------
			int collision_count = 0;
			for (int index = 0; index < owner.variable.length; index++)
				{
				if (index != variable.index)
					{
					if (owner.variable[index].is_displayed && (owner.variable[index].coordinates != null))
						{
						Complex diff = Complex.Sub(owner.variable[index].coordinates, c);
						double dist = (diff.re * diff.re) + (diff.im * diff.im);
						double pixel_dist_sq = (dist * (double)(zoom * zoom));
						if (pixel_dist_sq < 1.0)
							{
							//--------------------------------
							// Check to see if the two points
							// display on the same pixel.
							//--------------------------------
							Point p0 = GetPoint(c);
							Point p1 = GetPoint(owner.variable[index].coordinates);
							if ((p0.x == p1.x) && (p0.y == p1.y))
								{
								//--------------------------------
								// You have a "collision" only
								// if the other guy has a lower
								// index than you do.
								//--------------------------------
								if (index < variable.index)
									collision_count++;
								}
							}
						}
					}
				}

			Complex rotation = Complex.Number(1.0);
			     if (collision_count == 1) rotation = Complex.Number( 0.0, -1.0);
			else if (collision_count == 2) rotation = Complex.Number(-1.0,  0.0);
			else if (collision_count == 3) rotation = Complex.Number( 0.0,  1.0);

			offset = Complex.Mul(offset, rotation);
			Complex location = Complex.Add(c, offset);

			Point p = GetPoint(location);

			p.x -= (font_char_width/2);
			p.y += (font_char_height/2);

			offscreen_graphics.setColor(color);
			DrawString(text, p.x, p.y);
			}
		}
	private void AnimationStep()
		{
		//----------------------------------------------
		// Rotate the current point counter-clockwise
		// by a little bit.
		//----------------------------------------------
		final double angle = (2.0 * 3.14159265358979324) / 200.0;
		Complex c = owner.variable[current_var].coordinates;
		Complex d = Complex.Number(Math.cos(angle), Math.sin(angle));
		owner.variable[current_var].coordinates = Complex.Mul(c, d);
		}
	private void UserDefined()
		{
		PGC_Eval eval = owner.GetEval();

		for (int index = 0; index < owner.variable.length; index++)
			{
			if (index == PGC.VAR_F)
				{
				if ((op > PGC.OP_UNDEFINED) && (op < PGC.OP_USER))
					continue;  // a built-in function is already using "f"
				}
			if ((owner.variable[index].is_function) && (owner.variable[index].is_displayed))
				{
				eval.Evaluate(owner.variable[index].expression);
				owner.variable[index].coordinates = eval.value;

				// FUTURE: if null and !eval.empty then display error message
				}
			}
		}

	//----------------------------------------------------------------
	// PGC graph panel: External entry points
	//----------------------------------------------------------------

	public void init()
		{
		}
	public void run()
		{
		setBackground(PGC.color_background);
		while (true)
			{
			//-------------------
			// Animation loop
			//-------------------
			if (!painting)
				{
				AnimationStep();
				repaint();
				}
			try { Thread.sleep(30); }
			catch (InterruptedException e) { }
			}
		}
	public void update(Graphics graphics)
		{
		paint(graphics);
		}
	public boolean mouseMove(Event evt, int x, int y)
		{
		Complex c = GetComplex(new Point(x,y));
		owner.DisplayCoordinates(c);
		return (true);
		}
	public boolean mouseDown(Event evt, int x, int y)
		{
		requestFocus();

		owner.SetSpin(false);

		if (tracking_var == -1)
			{
			if (evt.controlDown() || evt.metaDown())
				{
				Complex c = GetComplex(new Point(x,y));
				tracking_scroll = c;
				}
			else
				{
				Complex c = GetComplex(new Point(x,y));
				tracking_var = GetVariableCoordinatesIndex(c);
				if (tracking_var >= 0)
					current_var = tracking_var;
				repaint();
				}
			}
		return (true);
		}
	public boolean mouseDrag(Event evt, int x, int y)
		{
		owner.SetSpin(false);

		Complex c = GetComplex(new Point(x,y));
		owner.DisplayCoordinates(c);

		if (tracking_var >= 0)
			{
			if (evt.shiftDown())
				{
				// snap to grid line
				Complex ci = Complex.Floor(Complex.Number(c.re + 0.5, c.im + 0.5));
				double dx = Math.abs(c.re - ci.re);
				double dy = Math.abs(c.im - ci.im);
				if ((dx <= (9.0/zoom)) && (dy <= (9.0/zoom)))
					{
					c.re = ci.re;
					c.im = ci.im;
					}
				else if (dx > dy)
					c.im = ci.im;
				else
					c.re = ci.re;
				}
			owner.variable[tracking_var].coordinates = c;
			repaint();
			}
		else if (tracking_scroll != null)
			{
			Point scroll_point = GetPoint(tracking_scroll);
			Point center_point = new Point(scroll_point.x - (x-base.x), scroll_point.y - (y-base.y));
			owner.SetCenter(GetComplex(center_point));
			}
		return (true);
		}
	public boolean mouseUp(Event evt, int x, int y)
		{
		owner.SetSpin(false);

		if (tracking_var >= 0)
			{
			tracking_var = -1;
			repaint();
			}
		if (tracking_scroll != null)
			{
			tracking_scroll = null;
			}
		return (true);
		}
	public boolean mouseExit(Event evt, int x, int y)
		{
		owner.DisplayCoordinates(null);
		return (true);
		}
	public boolean keyDown(Event evt, int key)
		{
		boolean result = false;
		boolean arrow  = false;
		boolean next_active_point = false;
		int     dx = 0;
		int     dy = 0;
		int     speed = 1;

		if (evt.shiftDown()) speed = 5;
		boolean ctrl = evt.controlDown();

		if      (key == evt.UP)     { result = true; arrow = true; dx=0;      dy=-speed;}
		else if (key == evt.DOWN)   { result = true; arrow = true; dx=0;      dy=speed; }
		else if (key == evt.LEFT)   { result = true; arrow = true; dx=-speed; dy=0;     }
		else if (key == evt.RIGHT)  { result = true; arrow = true; dx=speed;  dy=0;     }
		else if (key == '+')        { result = true; owner.SetZoomDelta(1); }
		else if (key == '=')        { result = true; owner.SetZoomDelta(1); }
		else if (key == '-')        { result = true; owner.SetZoomDelta(-1); }
		else if (key == 'l')        { result = true; owner.CheckLine(null); }
		else if (key == 'L')        { result = true; owner.CheckLine(null); }
		else if (key == 's')        { result = true; owner.CheckSpin(null); }
		else if (key == 'S')        { result = true; owner.CheckSpin(null); }
		else if (key == 'p')        { result = true; owner.CheckPlot(null); }
		else if (key == 'P')        { result = true; owner.CheckPlot(null); }
		else if (key == 'c')        { result = true; owner.Clear(); }
		else if (key == 'C')        { result = true; owner.Clear(); }
		else if (key == ' ')        { result = true; owner.Step(1); }
		else if (key == '>')        { result = true; owner.Step(100); }
		else if (key == 'z')        { result = true; owner.SetCenter(owner.option_center); owner.SetZoom(owner.option_zoom); }
		else if (key == 'Z')        { result = true; owner.SetCenter(owner.option_center); owner.SetZoom(owner.option_zoom); }
		else if (key == 'n')        { result = true; next_active_point = true; }
		else if (key == 'N')        { result = true; next_active_point = true; }

		if (arrow)
			{
			if (ctrl)
				{
				// Scroll the graph
				owner.SetCenterDelta(dx*20, dy*20);
				}
			else
				{
				// Move the active variable
				Point p = GetPoint(owner.variable[current_var].coordinates);
				p.x += dx;
				p.y += dy;
				owner.variable[current_var].coordinates = GetComplex(p);
				repaint();
				}
			}
		if (next_active_point)
			{
			NextActivePoint();
			}
		return (result);
		}
	public void paint(Graphics graphics)
		{
		Dimension dim = size();

		graphsize.x = dim.width;
		graphsize.y = dim.height;

		if ((graphsize.x < 1) || (graphsize.y < 1)) return;

		painting = true;

		base.x = graphsize.x / 2;
		base.y = graphsize.y / 2;

		// For double buffering:
		if ((offscreen_img == null) || (graphsize.x != offscreen_size.width) || (graphsize.y != offscreen_size.height))
			{
			offscreen_img = createImage(graphsize.x, graphsize.y);
			offscreen_size = new Dimension(dim);
			offscreen_graphics = offscreen_img.getGraphics();
			offscreen_graphics.setFont(getFont());
			}

		//----------------------
		// Apply functions
		//----------------------
			{
			owner.variable[PGC.VAR_F].coordinates = null;
			owner.variable[PGC.VAR_G].coordinates = null;
			owner.variable[PGC.VAR_H].coordinates = null;
			owner.variable[PGC.VAR_K].coordinates = null;
			owner.variable[PGC.VAR_M].coordinates = null;
			owner.variable[PGC.VAR_N].coordinates = null;

			     if (op == PGC.OP_ADD)    { owner.variable[PGC.VAR_F].coordinates = Complex.Add(owner.variable[PGC.VAR_A].coordinates, owner.variable[PGC.VAR_B].coordinates); }
			else if (op == PGC.OP_SUB)    { owner.variable[PGC.VAR_F].coordinates = Complex.Sub(owner.variable[PGC.VAR_A].coordinates, owner.variable[PGC.VAR_B].coordinates); }
			else if (op == PGC.OP_MUL)    { owner.variable[PGC.VAR_F].coordinates = Complex.Mul(owner.variable[PGC.VAR_A].coordinates, owner.variable[PGC.VAR_B].coordinates); }
			else if (op == PGC.OP_DIV)    { owner.variable[PGC.VAR_F].coordinates = Complex.Div(owner.variable[PGC.VAR_A].coordinates, owner.variable[PGC.VAR_B].coordinates); }
			else if (op == PGC.OP_POW)    { owner.variable[PGC.VAR_F].coordinates = Complex.Pow(owner.variable[PGC.VAR_A].coordinates, owner.variable[PGC.VAR_B].coordinates); }
			else if (op == PGC.OP_DOT)    { owner.variable[PGC.VAR_F].coordinates = Complex.Number(Complex.Dot(owner.variable[PGC.VAR_A].coordinates, owner.variable[PGC.VAR_B].coordinates)); }
			else if (op == PGC.OP_CROSS)  { owner.variable[PGC.VAR_F].coordinates = Complex.Number(Complex.Cross(owner.variable[PGC.VAR_A].coordinates, owner.variable[PGC.VAR_B].coordinates)); }
			else if (op == PGC.OP_MAT)    { owner.variable[PGC.VAR_F].coordinates = Complex.Number(Complex.Dot(owner.variable[PGC.VAR_A].coordinates, owner.variable[PGC.VAR_C].coordinates), Complex.Dot(owner.variable[PGC.VAR_B].coordinates, owner.variable[PGC.VAR_C].coordinates)); }
			else if (op == PGC.OP_SQRT)   { owner.variable[PGC.VAR_F].coordinates = Complex.Sqrt(owner.variable[PGC.VAR_A].coordinates); }
			else if (op == PGC.OP_CUBERT) { owner.variable[PGC.VAR_F].coordinates = Complex.Pow(owner.variable[PGC.VAR_A].coordinates, Complex.Number(1.0/3.0)); }
			else if (op == PGC.OP_EXP)    { owner.variable[PGC.VAR_F].coordinates = Complex.Exp(owner.variable[PGC.VAR_A].coordinates); }
			UserDefined();
			}

		//-----------------------------
		// Update coordinate display
		//-----------------------------
			{
			Complex a = null;
			Complex b = null;
			Complex c = null;
			Complex d = null;
			Complex f = null;
			Complex g = null;
			Complex h = null;
			Complex k = null;
			Complex m = null;
			Complex n = null;
			if (owner.variable[PGC.VAR_A].is_displayed) a = owner.variable[PGC.VAR_A].coordinates;
			if (owner.variable[PGC.VAR_B].is_displayed) b = owner.variable[PGC.VAR_B].coordinates;
			if (owner.variable[PGC.VAR_C].is_displayed) c = owner.variable[PGC.VAR_C].coordinates;
			if (owner.variable[PGC.VAR_D].is_displayed) d = owner.variable[PGC.VAR_D].coordinates;
			if (owner.variable[PGC.VAR_F].is_displayed) f = owner.variable[PGC.VAR_F].coordinates;
			if (owner.variable[PGC.VAR_G].is_displayed) g = owner.variable[PGC.VAR_G].coordinates;
			if (owner.variable[PGC.VAR_H].is_displayed) h = owner.variable[PGC.VAR_H].coordinates;
			if (owner.variable[PGC.VAR_K].is_displayed) k = owner.variable[PGC.VAR_K].coordinates;
			if (owner.variable[PGC.VAR_M].is_displayed) m = owner.variable[PGC.VAR_M].coordinates;
			if (owner.variable[PGC.VAR_N].is_displayed) n = owner.variable[PGC.VAR_N].coordinates;
			owner.TextFieldsOutput(a, b, c, d, f, g, h, k, m, n);
			if (plotting)
				{
				if (f != null) PlotAdd(PGC.VAR_F, f);
				if (g != null) PlotAdd(PGC.VAR_G, g);
				if (h != null) PlotAdd(PGC.VAR_H, h);
				if (k != null) PlotAdd(PGC.VAR_K, k);
				if (m != null) PlotAdd(PGC.VAR_M, m);
				if (n != null) PlotAdd(PGC.VAR_N, n);
				}
			}

		//----------------------------------------------
		// Draw background
		//----------------------------------------------
			{
			//-------------------------------------
			// Background color
			//-------------------------------------

			offscreen_graphics.setColor(PGC.color_background);
			offscreen_graphics.fillRect(0,0,graphsize.x,graphsize.y);

			//-------------------------------------
			// Background image
			//-------------------------------------

			if (background_img != null)
				{
				int x = GetXOrigin() - (background_img.getWidth(this) / 2);
				int y = GetYOrigin() - (background_img.getHeight(this) / 2);
				DrawImage(background_img, x, y, this);
				}

			//----------------------
			// Grid and axis lines
			//----------------------

			int grid_scale;
			Complex z_begin = GetComplex(new Point(0,0)); // upper left
			Complex z_end   = GetComplex(graphsize);      // lower right

			if (zoom >= 8)
				{
				grid_scale = zoom;
				z_begin.re = Math.floor(z_begin.re);
				z_begin.im = Math.floor(z_begin.im);
				z_end.re = Math.ceil(z_end.re);
				z_end.im = Math.ceil(z_end.im);
				}
			else
				{
				grid_scale = zoom * 10;
				z_begin.re = Math.floor(z_begin.re / 10.0) * 10.0;
				z_begin.im = Math.floor(z_begin.im / 10.0) * 10.0;
				z_end.re = Math.ceil(z_end.re / 10.0) * 10.0;
				z_end.im = Math.ceil(z_end.im / 10.0) * 10.0;
				}

			Point begin = GetPoint(z_begin);
			Point end   = GetPoint(z_end);

			offscreen_graphics.setColor(PGC.color_grid);

			for (int x = begin.x; x <= end.x; x += grid_scale)
				DrawLine(x, 0, x, graphsize.y);

			for (int y = begin.y; y <= end.y; y += grid_scale)
				DrawLine(0, y, graphsize.x, y);

			offscreen_graphics.setColor(PGC.color_axis);
			DrawLine(0,GetYOrigin(),graphsize.x,GetYOrigin());
			DrawLine(GetXOrigin(),0,GetXOrigin(),graphsize.y);

			//----------------------
			// Autoplot
			//----------------------

			PGC_Eval eval = owner.GetEval();

			for (int index = 0; index < owner.variable.length; index++)
				{
				if (owner.variable[index].is_function && !owner.variable[index].is_displayed)
					{
					String expr = owner.variable[index].expression;

					switch (index)
						{
						case PGC.VAR_F: offscreen_graphics.setColor(PGC.color_func_f_plot); break;
						case PGC.VAR_G: offscreen_graphics.setColor(PGC.color_func_g_plot); break;
						case PGC.VAR_H: offscreen_graphics.setColor(PGC.color_func_h_plot); break;
						case PGC.VAR_K: offscreen_graphics.setColor(PGC.color_func_k_plot); break;
						case PGC.VAR_M: offscreen_graphics.setColor(PGC.color_func_m_plot); break;
						case PGC.VAR_N: offscreen_graphics.setColor(PGC.color_func_n_plot); break;
						}

					eval.Evaluate(expr);
					if (eval.uses_autoplot_t)
						{
						//-----------------
						// Autoplot "t"
						//-----------------
						Point p = null;
						Complex auto_complex = Complex.Number(0.0);

						double t_minimum = owner.t_minimum;
						double t_maximum = owner.t_maximum;
						int    t_count   = owner.t_count;
						if (t_minimum > t_maximum)
							{
							t_minimum = owner.t_maximum;
							t_maximum = owner.t_minimum;
							}
						if (t_count < 1)
							t_count = 1;
						double t_increment = (t_maximum - t_minimum) / (double)t_count;
						double t_limit = t_maximum + t_increment;

						for (double t = t_minimum; t < t_limit; t += t_increment)
							{
							auto_complex.re = t;
							auto_complex.im = 0.0;
							owner.SetAutoplotVariable(auto_complex);
							eval.ReEvaluate();
							Complex plot = eval.value;
							if (plot == null)
								{
								// This is a break in the plot.
								p = null;
								continue;
								}
							if (!Complex.IsDefined(plot))
								{
								// NaN or infinity is a discontinuity in the plot
								p = null;
								continue;
								}
							Point q = GetPoint(plot);
							FillRect(q.x, q.y, 1, 1);
							if (p != null)
								{
								if (PointVisible(p) || PointVisible(q))
									DrawLine(p.x, p.y, q.x, q.y);
								}
							p = q;
							}
						}
					else if (eval.uses_autoplot_x)
						{
						//-----------------
						// Autoplot "x"
						//-----------------
						Point p = null;
						Point auto_point = new Point(0,0);
						Complex auto_complex = Complex.Number(0.0);

						for (int x = 0; x < graphsize.x; x++)
							{
							auto_point.x = x;
							auto_point.y = 0;
							PointToComplex(auto_point, auto_complex);
							auto_complex.im = 0.0;
							owner.SetAutoplotVariable(auto_complex);
							eval.ReEvaluate();
							Complex plot = eval.value;
							if (plot == null)
								{
								// This is a break in the plot.
								p = null;
								continue;
								}
							if (!owner.LooksReal(plot))
								{
								// "x" can only plot real-looking results
								p = null;
								continue;
								}
							if (!Complex.IsDefined(plot))
								{
								// NaN or infinity is a discontinuity in the plot
								p = null;
								continue;
								}
							plot.im = plot.re;
							plot.re = auto_complex.re;
							Point q = GetPoint(plot);
							FillRect(q.x, q.y, 1, 1);
							if (p != null)
								{
								if (PointVisible(p) || PointVisible(q))
									DrawLine(p.x, p.y, q.x, q.y);
								}
							p = q;
							}
						}

					owner.SetAutoplotVariable(null);
					}
				}
			}

		//----------------------
		// User plotting
		//----------------------
			{
			for (int index = 0; index < owner.variable.length; index++)
				{
				if (owner.variable[index].plotting_vec != null)
					{
					Point p = null;

					switch (index)
						{
						case PGC.VAR_F: offscreen_graphics.setColor(PGC.color_func_f_plot); break;
						case PGC.VAR_G: offscreen_graphics.setColor(PGC.color_func_g_plot); break;
						case PGC.VAR_H: offscreen_graphics.setColor(PGC.color_func_h_plot); break;
						case PGC.VAR_K: offscreen_graphics.setColor(PGC.color_func_k_plot); break;
						case PGC.VAR_M: offscreen_graphics.setColor(PGC.color_func_m_plot); break;
						case PGC.VAR_N: offscreen_graphics.setColor(PGC.color_func_n_plot); break;
						}

					for (int k = 0; k < owner.variable[index].plotting_vec.size(); k++)
						{
						Complex plot = (Complex)owner.variable[index].plotting_vec.elementAt(k);
						if (plot == null)
							{
							// This is a break in the plot.
							p = null;
							continue;
							}
						if (!Complex.IsDefined(plot))
							{
							// NaN or infinity is a discontinuity in the plot
							p = null;
							continue;
							}
						Point q = GetPoint(plot);
						FillRect(q.x, q.y, 1, 1);
						if (p != null)
							{
							if (PointVisible(p) || PointVisible(q))
								DrawLine(p.x, p.y, q.x, q.y);
							}
						p = q;
						}
					}
				}
			}

		//----------------------
		// Draw points
		//----------------------
			{
			Point p;
			Point q;

			for (int index = 0; index < owner.variable.length; index++)
				{
				if ((owner.variable[index].is_input) && (owner.variable[index].is_displayed))
					{
					//-----------------------------------------
					// Input points' line segments and labels
					//-----------------------------------------
					if (!Complex.IsDefined(owner.variable[index].coordinates))
						continue;
					p = GetPoint(owner.variable[index].coordinates);
					if (origin_connector)
						{
						offscreen_graphics.setColor(PGC.color_origin_connector_line);
						DrawLine(GetXOrigin(), GetYOrigin(), p.x, p.y);
						}
					DrawPointLabel(owner.variable[index], PGC.color_input_label);
					}
				}

			//--------------------
			// Func labels
			//--------------------
			DrawPointLabel(owner.variable[PGC.VAR_F], PGC.color_func_f_label);
			DrawPointLabel(owner.variable[PGC.VAR_G], PGC.color_func_g_label);
			DrawPointLabel(owner.variable[PGC.VAR_H], PGC.color_func_h_label);
			DrawPointLabel(owner.variable[PGC.VAR_K], PGC.color_func_k_label);
			DrawPointLabel(owner.variable[PGC.VAR_M], PGC.color_func_m_label);
			DrawPointLabel(owner.variable[PGC.VAR_N], PGC.color_func_n_label);

			//------------------------
			// Input points
			//------------------------
			for (int index = 0; index < owner.variable.length; index++)
				{
				if ((owner.variable[index].is_input) && (owner.variable[index].is_displayed))
					{
					p = GetPoint(owner.variable[index].coordinates);
					Rectangle r = new Rectangle(p);
					r.grow(2,2);
					offscreen_graphics.setColor(PGC.color_input_point);
					DrawRect(r.x, r.y, r.width, r.height);
					if (tracking_var == index)
						{
						// Optional: show the tracking arg
						// in a special way.
						}
					}
				}
			//-------------------
			// Func points
			//-------------------
			for (int index = 0; index < owner.variable.length; index++)
				{
				if (owner.variable[index].is_function && owner.variable[index].is_displayed)
					{
					Complex coordinates = owner.variable[index].coordinates;

					if ((coordinates != null) && Complex.IsDefined(coordinates))
						{
						for (int count = 0; true; count++)
							{
							p = GetPoint(coordinates);

							if (origin_connector)
								{
								offscreen_graphics.setColor(PGC.color_origin_connector_line);
								DrawLine(GetXOrigin(), GetYOrigin(), p.x, p.y);
								}
							Rectangle r = new Rectangle(p);
							r.grow(2,2);
							switch (index)
								{
								case PGC.VAR_F: offscreen_graphics.setColor(PGC.color_func_f_point); break;
								case PGC.VAR_G: offscreen_graphics.setColor(PGC.color_func_g_point); break;
								case PGC.VAR_H: offscreen_graphics.setColor(PGC.color_func_h_point); break;
								case PGC.VAR_K: offscreen_graphics.setColor(PGC.color_func_k_point); break;
								case PGC.VAR_M: offscreen_graphics.setColor(PGC.color_func_m_point); break;
								case PGC.VAR_N: offscreen_graphics.setColor(PGC.color_func_n_point); break;
								}

							DrawRect(r.x, r.y, r.width, r.height);

							if ((index == PGC.VAR_F) && (op == PGC.OP_SQRT) && (count < 1))
								{
								coordinates = Complex.Neg(coordinates); // multiply by -1 to get the other square root
								continue;
								}
							else if ((index == PGC.VAR_F) && (op == PGC.OP_CUBERT) && (count < 2))
								{
								coordinates = Complex.Mul(coordinates, Complex.Number(-0.5, Math.sqrt(3.0)/2.0)); // multiply by (i^(4/3)) to get the next cube root
								continue;
								}
							break;
							}
						}
					}
				}

			//-------------------
			// Projection vectors
			//-------------------
			if (((op == PGC.OP_DOT) || (op == PGC.OP_CROSS)) && ((tracking_var != -1) || (animation_thread != null)))
				{
				//------------------------------------------
				// For the dot-product and cross-product,
				// also show the ray and projection lines.
				//------------------------------------------
				Complex projection;
				if (current_var == PGC.VAR_A)
					{
					projection = Complex.Projection(owner.variable[PGC.VAR_A].coordinates,owner.variable[PGC.VAR_B].coordinates);
					p = GetPoint(owner.variable[PGC.VAR_A].coordinates);
					q = GetPoint(owner.variable[PGC.VAR_B].coordinates);
					}
				else
					{
					projection = Complex.Projection(owner.variable[PGC.VAR_B].coordinates,owner.variable[PGC.VAR_A].coordinates);
					p = GetPoint(owner.variable[PGC.VAR_B].coordinates);
					q = GetPoint(owner.variable[PGC.VAR_A].coordinates);
					}
				Point s = GetPoint(projection);

				// Show input connectors on p and q
				offscreen_graphics.setColor(PGC.color_origin_connector_line);
				DrawLine(GetXOrigin(), GetYOrigin(), p.x, p.y);
				DrawLine(GetXOrigin(), GetYOrigin(), q.x, q.y);

				// Show the light ray normal to q.
				offscreen_graphics.setColor(PGC.color_lightray);
				DrawLine(p.x, p.y, s.x, s.y);

				// Show the projection: the shadow that p casts on q.
				offscreen_graphics.setColor(PGC.color_projection);
				Rectangle r = new Rectangle(s);
				r.grow(2,2);
				DrawRect(r.x, r.y, r.width, r.height);
				DrawLine(GetXOrigin(), GetYOrigin(), s.x, s.y);
				}
			if ((op == PGC.OP_MAT) && ((tracking_var == PGC.VAR_C) || ((animation_thread != null) && (current_var == PGC.VAR_C))))
				{
				//------------------------------------------
				// For matrix multiplication, show the
				// the ray and projection lines on both
				// vectors in the matrix.
				//------------------------------------------
				for (int i = PGC.VAR_A; i <= PGC.VAR_B; i++)
					{
					Complex projection = Complex.Projection(owner.variable[PGC.VAR_C].coordinates, owner.variable[i].coordinates);
					p = GetPoint(owner.variable[PGC.VAR_C].coordinates);
					q = GetPoint(owner.variable[i].coordinates);

					Point s = GetPoint(projection);

					// Show input connectors on p and q
					offscreen_graphics.setColor(PGC.color_origin_connector_line);
					DrawLine(GetXOrigin(), GetYOrigin(), p.x, p.y);
					DrawLine(GetXOrigin(), GetYOrigin(), q.x, q.y);

					// Show the light ray normal to q.
					offscreen_graphics.setColor(PGC.color_lightray);
					DrawLine(p.x, p.y, s.x, s.y);

					// Show the projection: the shadow that p casts on q.
					offscreen_graphics.setColor(PGC.color_projection);
					Rectangle r = new Rectangle(s);
					r.grow(2,2);
					DrawRect(r.x, r.y, r.width, r.height);
					DrawLine(GetXOrigin(), GetYOrigin(), s.x, s.y);

					// Label the projection vectors: "x" and "y"
					int j = -1;
					if (i == PGC.VAR_A)
						j = PGC.VAR_PROJECTION_X;
					else if (i == PGC.VAR_B)
						j = PGC.VAR_PROJECTION_Y;
					if (j != -1)
						{
						owner.variable[j].coordinates = projection;
						owner.variable[j].is_displayed = true;
						DrawPointLabel(owner.variable[j], PGC.color_projection_label);
						owner.variable[j].is_displayed = false;
						}
					}
				}
			}

		//-------------------------------------
		// Draw 3d lines around the panel
		//-------------------------------------
		if (owner.option_graph_border)
			{
			offscreen_graphics.setColor(PGC.color_3d_dark);
			offscreen_graphics.drawLine(0,0,graphsize.x-2,0);
			offscreen_graphics.drawLine(0,0,0,graphsize.y-2);
			offscreen_graphics.setColor(Color.black);
			offscreen_graphics.drawLine(1,1,graphsize.x-1,1);
			offscreen_graphics.drawLine(1,1,1,graphsize.y-1);
			offscreen_graphics.setColor(Color.lightGray);
			offscreen_graphics.drawLine(graphsize.x-2,1,graphsize.x-2,graphsize.y-2);
			offscreen_graphics.drawLine(1,graphsize.y-2,graphsize.x-2,graphsize.y-2);
			offscreen_graphics.setColor(PGC.color_3d_light);
			offscreen_graphics.drawLine(graphsize.x-1,0,graphsize.x-1,graphsize.y-1);
			offscreen_graphics.drawLine(0,graphsize.y-1,graphsize.x-1,graphsize.y-1);
			}

		graphics.drawImage(offscreen_img, 0, 0, this);

		painting = false;
		}
	}



//========================================================================
//========================================================================
// PGC details frame
//========================================================================
//========================================================================

class PGC_DetailsFrame extends Frame
	{
	private PGC                  owner = null;
	private Panel                pnl_root;

	public  Panel                pnl_coordinates;
	public  Panel                pnl_expressions;
	public  Panel                pnl_parametric;
	public  Panel                pnl_error_message;
	public  PGC_LabelFixed       lbl_a;
	public  PGC_LabelFixed       lbl_a_x;
	public  PGC_LabelFixed       lbl_a_y;
	public  PGC_LabelFixed       lbl_b;
	public  PGC_LabelFixed       lbl_b_x;
	public  PGC_LabelFixed       lbl_b_y;
	public  PGC_LabelFixed       lbl_c;
	public  PGC_LabelFixed       lbl_c_x;
	public  PGC_LabelFixed       lbl_c_y;
	public  PGC_LabelFixed       lbl_d;
	public  PGC_LabelFixed       lbl_d_x;
	public  PGC_LabelFixed       lbl_d_y;
	public  PGC_LabelFixed       lbl_f;
	public  PGC_LabelFixed       lbl_f_x;
	public  PGC_LabelFixed       lbl_f_y;
	public  PGC_LabelFixed       lbl_g;
	public  PGC_LabelFixed       lbl_g_x;
	public  PGC_LabelFixed       lbl_g_y;
	public  PGC_LabelFixed       lbl_h;
	public  PGC_LabelFixed       lbl_h_x;
	public  PGC_LabelFixed       lbl_h_y;
	public  PGC_LabelFixed       lbl_h_def;
	public  PGC_LabelFixed       lbl_h_expr;
	public  PGC_LabelFixed       lbl_k;
	public  PGC_LabelFixed       lbl_k_x;
	public  PGC_LabelFixed       lbl_k_y;
	public  PGC_LabelFixed       lbl_k_def;
	public  PGC_LabelFixed       lbl_k_expr;
	public  PGC_LabelFixed       lbl_m;
	public  PGC_LabelFixed       lbl_m_x;
	public  PGC_LabelFixed       lbl_m_y;
	public  PGC_LabelFixed       lbl_m_def;
	public  PGC_LabelFixed       lbl_m_expr;
	public  PGC_LabelFixed       lbl_n;
	public  PGC_LabelFixed       lbl_n_x;
	public  PGC_LabelFixed       lbl_n_y;
	public  PGC_LabelFixed       lbl_n_def;
	public  PGC_LabelFixed       lbl_n_expr;
	public  PGC_LabelFixed       lbl_t_min;
	public  PGC_LabelFixed       lbl_t_max;
	public  PGC_LabelFixed       lbl_t_rep;
	public  PGC_LabelFixed       lbl_error_message;
	public  PGC_TextFieldFixed   txt_a_x;
	public  PGC_TextFieldFixed   txt_a_y;
	public  PGC_TextFieldFixed   txt_b_x;
	public  PGC_TextFieldFixed   txt_b_y;
	public  PGC_TextFieldFixed   txt_c_x;
	public  PGC_TextFieldFixed   txt_c_y;
	public  PGC_TextFieldFixed   txt_d_x;
	public  PGC_TextFieldFixed   txt_d_y;
	public  PGC_TextFieldFixed   txt_f_x;
	public  PGC_TextFieldFixed   txt_f_y;
	public  PGC_TextFieldFixed   txt_g_x;
	public  PGC_TextFieldFixed   txt_g_y;
	public  PGC_TextFieldFixed   txt_h_x;
	public  PGC_TextFieldFixed   txt_h_y;
	public  PGC_TextFieldFixed   txt_k_x;
	public  PGC_TextFieldFixed   txt_k_y;
	public  PGC_TextFieldFixed   txt_m_x;
	public  PGC_TextFieldFixed   txt_m_y;
	public  PGC_TextFieldFixed   txt_n_x;
	public  PGC_TextFieldFixed   txt_n_y;
	public  PGC_TextFieldFixed   txt_h_expr;
	public  PGC_TextFieldFixed   txt_k_expr;
	public  PGC_TextFieldFixed   txt_m_expr;
	public  PGC_TextFieldFixed   txt_n_expr;
	public  PGC_TextFieldFixed   txt_t_min;
	public  PGC_TextFieldFixed   txt_t_max;
	public  PGC_TextFieldFixed   txt_t_rep;

	public PGC_DetailsFrame(PGC own, String strName)
		{
		super(strName);
		owner = own;
		}

	private void LayoutWindow()
		{
		Panel panel;
		Component component;
		GridBagConstraints constraints;

		//----------------------------------------------
		// pnl_root (the root panel)
		//----------------------------------------------

		component = pnl_root = PGC.StdPanel();

		//----------------------------------------------
		// pnl_coordinates
		//----------------------------------------------

		pnl_root.add(component = pnl_coordinates = PGC.StdPanel());
		((GridBagLayout)(pnl_root.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine(new Insets(5,5,5,5)));

		pnl_coordinates.add(panel = PGC.StdPanel());
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(panel, PGC.StdConstraintsNewLine(new Insets(0,0,5,0)));
		panel.add(component = new PGC_LabelFixed("Point coordinates", PGC.component_full_text_size));
		((GridBagLayout)(panel.getLayout())).setConstraints(component, PGC.StdConstraints());

		pnl_coordinates.add(component = PGC.StdPanel());
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine());

		pnl_coordinates.add(component = lbl_a = new PGC_LabelFixed("a", new Dimension(20, -1)));
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, PGC.StdConstraints());
		pnl_coordinates.add(component = lbl_a_x = new PGC_LabelFixed(" x= ", new Dimension(22, -1)));
		pnl_coordinates.add(component = txt_a_x = new PGC_TextFieldFixed(owner, PGC.component_number_text_size));
		pnl_coordinates.add(component = lbl_a_y = new PGC_LabelFixed(" y= ", new Dimension(22, -1)));
		pnl_coordinates.add(component = txt_a_y = new PGC_TextFieldFixed(owner, PGC.component_number_text_size));

		pnl_coordinates.add(component = PGC.StdPanel());
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine());

		pnl_coordinates.add(component = lbl_b = new PGC_LabelFixed("b", new Dimension(20, -1)));
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, PGC.StdConstraints());
		pnl_coordinates.add(component = lbl_b_x = new PGC_LabelFixed(" x= ", new Dimension(22, -1)));
		pnl_coordinates.add(component = txt_b_x = new PGC_TextFieldFixed(owner, PGC.component_number_text_size));
		pnl_coordinates.add(component = lbl_b_y = new PGC_LabelFixed(" y= ", new Dimension(22, -1)));
		pnl_coordinates.add(component = txt_b_y = new PGC_TextFieldFixed(owner, PGC.component_number_text_size));

		pnl_coordinates.add(component = PGC.StdPanel());
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine());

		pnl_coordinates.add(component = lbl_c = new PGC_LabelFixed("c", new Dimension(20, -1)));
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, PGC.StdConstraints());
		pnl_coordinates.add(component = lbl_c_x = new PGC_LabelFixed(" x= ", new Dimension(22, -1)));
		pnl_coordinates.add(component = txt_c_x = new PGC_TextFieldFixed(owner, PGC.component_number_text_size));
		pnl_coordinates.add(component = lbl_c_y = new PGC_LabelFixed(" y= ", new Dimension(22, -1)));
		pnl_coordinates.add(component = txt_c_y = new PGC_TextFieldFixed(owner, PGC.component_number_text_size));

		pnl_coordinates.add(component = PGC.StdPanel());
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine());

		pnl_coordinates.add(component = lbl_d = new PGC_LabelFixed("d", new Dimension(20, -1)));
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, PGC.StdConstraints());
		pnl_coordinates.add(component = lbl_d_x = new PGC_LabelFixed(" x= ", new Dimension(22, -1)));
		pnl_coordinates.add(component = txt_d_x = new PGC_TextFieldFixed(owner, PGC.component_number_text_size));
		pnl_coordinates.add(component = lbl_d_y = new PGC_LabelFixed(" y= ", new Dimension(22, -1)));
		pnl_coordinates.add(component = txt_d_y = new PGC_TextFieldFixed(owner, PGC.component_number_text_size));

		pnl_coordinates.add(component = PGC.StdPanel());
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine());

		pnl_coordinates.add(component = lbl_f = new PGC_LabelFixed("f", new Dimension(20, -1)));
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, PGC.StdConstraints());
		pnl_coordinates.add(component = lbl_f_x = new PGC_LabelFixed(" x= ", new Dimension(22, -1)));
		pnl_coordinates.add(component = txt_f_x = new PGC_TextFieldFixed(owner, PGC.component_number_text_size));
		pnl_coordinates.add(component = lbl_f_y = new PGC_LabelFixed(" y= ", new Dimension(22, -1)));
		pnl_coordinates.add(component = txt_f_y = new PGC_TextFieldFixed(owner, PGC.component_number_text_size));

		pnl_coordinates.add(component = PGC.StdPanel());
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine());

		pnl_coordinates.add(component = lbl_g = new PGC_LabelFixed("g", new Dimension(20, -1)));
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, PGC.StdConstraints());
		pnl_coordinates.add(component = lbl_g_x = new PGC_LabelFixed(" x= ", new Dimension(22, -1)));
		pnl_coordinates.add(component = txt_g_x = new PGC_TextFieldFixed(owner, PGC.component_number_text_size));
		pnl_coordinates.add(component = lbl_g_y = new PGC_LabelFixed(" y= ", new Dimension(22, -1)));
		pnl_coordinates.add(component = txt_g_y = new PGC_TextFieldFixed(owner, PGC.component_number_text_size));

		pnl_coordinates.add(component = PGC.StdPanel());
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine());

		pnl_coordinates.add(component = lbl_h = new PGC_LabelFixed("h", new Dimension(20, -1)));
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, PGC.StdConstraints());
		pnl_coordinates.add(component = lbl_h_x = new PGC_LabelFixed(" x= ", new Dimension(22, -1)));
		pnl_coordinates.add(component = txt_h_x = new PGC_TextFieldFixed(owner, PGC.component_number_text_size));
		pnl_coordinates.add(component = lbl_h_y = new PGC_LabelFixed(" y= ", new Dimension(22, -1)));
		pnl_coordinates.add(component = txt_h_y = new PGC_TextFieldFixed(owner, PGC.component_number_text_size));

		pnl_coordinates.add(component = PGC.StdPanel());
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine());

		pnl_coordinates.add(component = lbl_k = new PGC_LabelFixed("k", new Dimension(20, -1)));
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, PGC.StdConstraints());
		pnl_coordinates.add(component = lbl_k_x = new PGC_LabelFixed(" x= ", new Dimension(22, -1)));
		pnl_coordinates.add(component = txt_k_x = new PGC_TextFieldFixed(owner, PGC.component_number_text_size));
		pnl_coordinates.add(component = lbl_k_y = new PGC_LabelFixed(" y= ", new Dimension(22, -1)));
		pnl_coordinates.add(component = txt_k_y = new PGC_TextFieldFixed(owner, PGC.component_number_text_size));

		pnl_coordinates.add(component = PGC.StdPanel());
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine());

		pnl_coordinates.add(component = lbl_m = new PGC_LabelFixed("m", new Dimension(20, -1)));
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, PGC.StdConstraints());
		pnl_coordinates.add(component = lbl_m_x = new PGC_LabelFixed(" x= ", new Dimension(22, -1)));
		pnl_coordinates.add(component = txt_m_x = new PGC_TextFieldFixed(owner, PGC.component_number_text_size));
		pnl_coordinates.add(component = lbl_m_y = new PGC_LabelFixed(" y= ", new Dimension(22, -1)));
		pnl_coordinates.add(component = txt_m_y = new PGC_TextFieldFixed(owner, PGC.component_number_text_size));

		pnl_coordinates.add(component = PGC.StdPanel());
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine());

		pnl_coordinates.add(component = lbl_n = new PGC_LabelFixed("n", new Dimension(20, -1)));
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, PGC.StdConstraints());
		pnl_coordinates.add(component = lbl_n_x = new PGC_LabelFixed(" x= ", new Dimension(22, -1)));
		pnl_coordinates.add(component = txt_n_x = new PGC_TextFieldFixed(owner, PGC.component_number_text_size));
		pnl_coordinates.add(component = lbl_n_y = new PGC_LabelFixed(" y= ", new Dimension(22, -1)));
		pnl_coordinates.add(component = txt_n_y = new PGC_TextFieldFixed(owner, PGC.component_number_text_size));

		pnl_coordinates.add(component = PGC.StdPanel());
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine());

		//----------------------------------------------
		// divider
		//----------------------------------------------

		Panel div1;
		pnl_coordinates.add(component = div1 = PGC.StdPanel());
		constraints = PGC.StdConstraintsNewLine(new Insets(12,0,0,0));
		constraints.fill = GridBagConstraints.HORIZONTAL;
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, constraints);
		component.setBackground(PGC.color_3d_dark);
		div1.add(component = new PGC_LabelFixed(new Dimension(1,1)));

		pnl_coordinates.add(component = div1 = PGC.StdPanel());
		constraints = PGC.StdConstraintsNewLine();
		constraints.fill = GridBagConstraints.HORIZONTAL;
		((GridBagLayout)(pnl_coordinates.getLayout())).setConstraints(component, constraints);
		component.setBackground(PGC.color_3d_light);
		div1.add(component = new PGC_LabelFixed(new Dimension(1,1)));

		//----------------------------------------------
		// user-defined functions
		//----------------------------------------------

		pnl_root.add(component = pnl_expressions = PGC.StdPanel());
		((GridBagLayout)(pnl_root.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine(new Insets(5,5,5,5)));

		pnl_expressions.add(panel = PGC.StdPanel());
		((GridBagLayout)(pnl_expressions.getLayout())).setConstraints(panel, PGC.StdConstraintsNewLine(new Insets(0,0,5,0)));
		panel.add(component = new PGC_LabelFixed("Additional user-defined functions", PGC.component_full_text_size));
		((GridBagLayout)(panel.getLayout())).setConstraints(component, PGC.StdConstraints());

		pnl_expressions.add(component = PGC.StdPanel());
		((GridBagLayout)(pnl_expressions.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine());

		pnl_expressions.add(component = lbl_h_expr = new PGC_LabelFixed("h", new Dimension(20, -1)));
		((GridBagLayout)(pnl_expressions.getLayout())).setConstraints(component, PGC.StdConstraints());
		pnl_expressions.add(component = lbl_h_def  = new PGC_LabelFixed("= ", new Dimension(12, -1))); lbl_h_def.setAlignment(Label.RIGHT);
		pnl_expressions.add(component = txt_h_expr = new PGC_TextFieldFixed(owner, PGC.component_wide_text_size));

		pnl_expressions.add(component = PGC.StdPanel());
		((GridBagLayout)(pnl_expressions.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine());

		pnl_expressions.add(component = lbl_k_expr = new PGC_LabelFixed("k", new Dimension(20, -1)));
		((GridBagLayout)(pnl_expressions.getLayout())).setConstraints(component, PGC.StdConstraints());
		pnl_expressions.add(component = lbl_k_def  = new PGC_LabelFixed("= ", new Dimension(12, -1))); lbl_k_def.setAlignment(Label.RIGHT);
		pnl_expressions.add(component = txt_k_expr = new PGC_TextFieldFixed(owner, PGC.component_wide_text_size));

		pnl_expressions.add(component = PGC.StdPanel());
		((GridBagLayout)(pnl_expressions.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine());

		pnl_expressions.add(component = lbl_m_expr = new PGC_LabelFixed("m", new Dimension(20, -1)));
		((GridBagLayout)(pnl_expressions.getLayout())).setConstraints(component, PGC.StdConstraints());
		pnl_expressions.add(component = lbl_m_def  = new PGC_LabelFixed("= ", new Dimension(12, -1))); lbl_m_def.setAlignment(Label.RIGHT);
		pnl_expressions.add(component = txt_m_expr = new PGC_TextFieldFixed(owner, PGC.component_wide_text_size));

		pnl_expressions.add(component = PGC.StdPanel());
		((GridBagLayout)(pnl_expressions.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine());

		pnl_expressions.add(component = lbl_n_expr = new PGC_LabelFixed("n", new Dimension(20, -1)));
		((GridBagLayout)(pnl_expressions.getLayout())).setConstraints(component, PGC.StdConstraints());
		pnl_expressions.add(component = lbl_n_def  = new PGC_LabelFixed("= ", new Dimension(12, -1))); lbl_n_def.setAlignment(Label.RIGHT);
		pnl_expressions.add(component = txt_n_expr = new PGC_TextFieldFixed(owner, PGC.component_wide_text_size));

		pnl_expressions.add(component = PGC.StdPanel());
		((GridBagLayout)(pnl_expressions.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine());

		//----------------------------------------------
		// divider
		//----------------------------------------------

		Panel div0;
		pnl_expressions.add(component = div0 = PGC.StdPanel());
		constraints = PGC.StdConstraintsNewLine(new Insets(12,0,0,0));
		constraints.fill = GridBagConstraints.HORIZONTAL;
		((GridBagLayout)(pnl_expressions.getLayout())).setConstraints(component, constraints);
		component.setBackground(PGC.color_3d_dark);
		div0.add(component = new PGC_LabelFixed(new Dimension(1,1)));

		pnl_expressions.add(component = div0 = PGC.StdPanel());
		constraints = PGC.StdConstraintsNewLine();
		constraints.fill = GridBagConstraints.HORIZONTAL;
		((GridBagLayout)(pnl_expressions.getLayout())).setConstraints(component, constraints);
		component.setBackground(PGC.color_3d_light);
		div0.add(component = new PGC_LabelFixed(new Dimension(1,1)));

		//----------------------------------------------
		// Parametric variable
		//----------------------------------------------

		pnl_root.add(component = pnl_parametric = PGC.StdPanel());
		((GridBagLayout)(pnl_root.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine(new Insets(5,5,5,5)));

		pnl_parametric.add(panel = PGC.StdPanel());
		((GridBagLayout)(pnl_parametric.getLayout())).setConstraints(panel, PGC.StdConstraintsNewLine(new Insets(0,0,5,0)));
		panel.add(component = new PGC_LabelFixed("Parametric variable", PGC.component_full_text_size));
		((GridBagLayout)(panel.getLayout())).setConstraints(component, PGC.StdConstraints());

		pnl_parametric.add(component = lbl_t_min = new PGC_LabelFixed("t = ", new Dimension(22, -1)));
		((GridBagLayout)(pnl_parametric.getLayout())).setConstraints(component, PGC.StdConstraints());
		pnl_parametric.add(component = txt_t_min = new PGC_TextFieldFixed(owner, new Dimension(92, -1)));

		pnl_parametric.add(component = lbl_t_max = new PGC_LabelFixed(" to ", new Dimension(18, -1)));
		((GridBagLayout)(pnl_parametric.getLayout())).setConstraints(component, PGC.StdConstraints()); lbl_t_max.setAlignment(Label.CENTER);
		pnl_parametric.add(component = txt_t_max = new PGC_TextFieldFixed(owner, new Dimension(92, -1)));

		pnl_parametric.add(component =             new PGC_LabelFixed("", new Dimension(15, -1)));
		pnl_parametric.add(component = lbl_t_rep = new PGC_LabelFixed("plot ", new Dimension(35, -1))); lbl_t_rep.setAlignment(Label.RIGHT);
		((GridBagLayout)(pnl_parametric.getLayout())).setConstraints(component, PGC.StdConstraints());
		pnl_parametric.add(component = txt_t_rep = new PGC_TextFieldFixed(owner, new Dimension(50, -1)));
		pnl_parametric.add(component =             new PGC_LabelFixed(" points", new Dimension(60, -1)));

		pnl_parametric.add(component = PGC.StdPanel());
		((GridBagLayout)(pnl_parametric.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine());

		//----------------------------------------------
		// Error message
		//----------------------------------------------

		pnl_root.add(component = pnl_error_message = PGC.StdPanel());
		((GridBagLayout)(pnl_root.getLayout())).setConstraints(component, PGC.StdConstraintsNewLine(new Insets(5,5,5,5)));
		pnl_error_message.add(component = lbl_error_message = new PGC_LabelFixed("", PGC.component_full_text_size));

		setResizable(false);
		pack();
		}
	public void init()
		{
		LayoutWindow(); // defines pnl_root
		setLayout(new BorderLayout());
		add("Center", pnl_root);
		pack();
		}
	public boolean keyDown(Event evt, int key)
		{
		boolean result;
		if ((key == evt.ENTER) || (key == evt.ESCAPE))
			{owner.GraphFocus(); result = true;}
		else
			result = super.keyDown(evt, key);
		return (result);
		}
	public boolean handleEvent(Event evt)
		{
		if (evt.id == evt.WINDOW_DESTROY)
			{
			if (owner != null)
				{
				owner.TextFieldsInput(null);
				owner.FrameDestroy(this);
				}
			return (true);
			}
		return (false);
		}
	}



//========================================================================
//========================================================================
// PGC user-interface support classes
//========================================================================
//========================================================================

class PGC_Frame extends Frame
	{
	private PGC owner = null;

	public PGC_Frame(PGC own, String strName)
		{
		super(strName);
		owner = own;
		}
	public boolean handleEvent(Event evt)
		{
		if (evt.id == evt.WINDOW_DESTROY)
			{
			if (owner != null)
				owner.MainFrameClose();
			return (true);
			}
		return (super.handleEvent(evt));
		}
	public boolean action(Event evt, Object arg)
		{
		if (owner != null)
			return (owner.action(evt, arg)); // reflect actions to owner
		else
			return (super.action(evt, arg));
		}
	}

class PGC_TextFrame extends Frame
	{
	private PGC       owner = null;
	private Panel     pnl_root;
	private TextArea  textarea;
	private int       height;
	private int       width;

	public PGC_TextFrame(PGC own, String strName, int height_arg, int width_arg)
		{
		super(strName);
		owner  = own;
		height = height_arg;
		width  = width_arg;
		}
	public void init()
		{
		add("Center", pnl_root = new Panel());
		pnl_root.setBackground(PGC.color_panel);
		pnl_root.setLayout(new GridBagLayout());
		pnl_root.add(textarea = new TextArea(height,width));
		textarea.setEditable(false);
		textarea.setBackground(PGC.color_background);
		setResizable(false); // a TextArea can't resize normally
		pack();
		}
	public void Write(Object obj)
		{
		textarea.appendText(obj.toString());
		}
	public void ShowTop()
		{
		textarea.select(0, 0);
		}
	public boolean handleEvent(Event evt)
		{
		if (evt.id == evt.WINDOW_DESTROY)
			{
			if (owner != null)
				owner.FrameDestroy(this);
			return (true);
			}
		else if (evt.id == evt.GOT_FOCUS)
			{
			// Forward all focus requests to the textarea.
			textarea.requestFocus();
			}
		return (false);
		}
	}

class PGC_ChoiceOp extends java.awt.Choice
	{
	private PGC owner = null;
	public PGC_ChoiceOp(PGC own) { super(); owner = own; }
	public boolean action(Event evt, Object arg)
		{
		if ((evt.id == evt.ACTION_EVENT) && (arg != null))
			{
			int op = PGC.OpEvalChoiceLabel(arg.toString());
			if (owner != null)
				owner.SetOp(op);
			return (true);
			}
		return (super.action(evt, arg));
		}
	}

class PGC_ButtonFixed extends java.awt.Button
	{
	//-----------------------------
	// Fixed-size button.
	// (Use -1 for default size.)
	//-----------------------------
	private Dimension dimension = null;
	public PGC_ButtonFixed()                             { super(); }
	public PGC_ButtonFixed(String label)                 { super(label); }
	public PGC_ButtonFixed(Dimension dim)                { super(); setMinimumSize(dim); }
	public PGC_ButtonFixed(String label, Dimension dim)  { super(label); setMinimumSize(dim); }
	public void      setMinimumSize(Dimension dim)       { dimension = dim; }
	public Dimension getMinimumSize()                    { return (minimumSize()); }
	public Dimension getMaximumSize()                    { return (minimumSize()); }
	public Dimension getPreferredSize()                  { return (minimumSize()); }
	public Dimension preferredSize()                     { return (minimumSize()); }
	public Dimension minimumSize()
		{
		Dimension result = super.minimumSize();
		if (dimension != null)
			{
			if (dimension.width  != -1) result.width  = dimension.width;
			if (dimension.height != -1) result.height = dimension.height;
			}
		return (result);
		}
	}

class PGC_LabelFixed extends java.awt.Label
	{
	//-----------------------------
	// Fixed-size label.
	// (Use -1 for default size.)
	//-----------------------------
	private Dimension dimension = null;
	public PGC_LabelFixed()                             { super(); }
	public PGC_LabelFixed(String label)                 { super(label); }
	public PGC_LabelFixed(Dimension dim)                { super(); setMinimumSize(dim); }
	public PGC_LabelFixed(String label, Dimension dim)  { super(label); setMinimumSize(dim); }
	public void      setMinimumSize(Dimension dim)      { dimension = dim; }
	public Dimension getMinimumSize()                   { return (minimumSize()); }
	public Dimension getMaximumSize()                   { return (minimumSize()); }
	public Dimension getPreferredSize()                 { return (minimumSize()); }
	public Dimension preferredSize()                    { return (minimumSize()); }
	public Dimension minimumSize()
		{
		Dimension result = super.minimumSize();
		if (dimension != null)
			{
			if (dimension.width  != -1) result.width  = dimension.width;
			if (dimension.height != -1) result.height = dimension.height;
			}
		return (result);
		}
	}

class PGC_TextFieldFixed extends java.awt.TextField
	{
	//-----------------------------
	// Fixed-size text field.
	// (Use -1 for default size.)
	//-----------------------------
	private PGC owner = null;
	private Dimension dimension = null;
	public PGC_TextFieldFixed(PGC own)                               { super(); owner = own; }
	public PGC_TextFieldFixed(PGC own, String label)                 { super(label); owner = own; }
	public PGC_TextFieldFixed(PGC own, Dimension dim)                { super(); setMinimumSize(dim); owner = own; }
	public PGC_TextFieldFixed(PGC own, String label, Dimension dim)  { super(label); setMinimumSize(dim); owner = own; }
	public void      setMinimumSize(Dimension dim)                   { dimension = dim; }
	public Dimension getMinimumSize()                                { return (minimumSize()); }
	public Dimension getMaximumSize()                                { return (minimumSize()); }
	public Dimension getPreferredSize()                              { return (minimumSize()); }
	public Dimension preferredSize()                                 { return (minimumSize()); }
	public Dimension minimumSize()
		{
		Dimension result = super.minimumSize();
		if (dimension != null)
			{
			if (dimension.width  != -1) result.width  = dimension.width;
			if (dimension.height != -1) result.height = dimension.height;
			}
		return (result);
		}
	public void setText(String s)
		{
		//--------------------------------------------------
		// Minimize setting the text, because it's
		// annoying while the field is still being edited.
		//--------------------------------------------------
		String text = super.getText();
		if (!s.equals(text)) super.setText(s);
		}
	public boolean keyDown(Event evt, int key)
		{
		boolean result;
		if (key == evt.ENTER)
			{owner.TextFieldsInput(this); result = true;}
		else if (key == evt.ESCAPE)
			{owner.GraphFocus(); result = true;}
		else
			result = super.keyDown(evt, key);
		return (result);
		}
	public boolean gotFocus(Event evt, Object what)
		{
		// It would be nice to do this (for certain fields, at least),
		// but there are Netscape browser bugs that make this unacceptable.
		// (This works fine for IE5 and JDK 1.1.7):
		//
		// selectAll();
		// repaint();

		return (super.gotFocus(evt, what));
		}
	public boolean lostFocus(Event evt, Object what)
		{
		if (owner != null)
			owner.TextFieldsInput(this);

		// It would be nice to do this (for certain fields, at least),
		// but there are Netscape browser bugs that make this unacceptable.
		// (This works fine for IE5 and JDK 1.1.7):
		//
		// select(0,0);
		// repaint();

		return (super.lostFocus(evt, what));
		}
	}

class PGC_Variable
	{
	public int     index;                      // index of the variable
	public String  name              = null;   // name of the variable
	public boolean is_input          = false;  // coordinates can be set by the user
	public boolean is_function       = false;  // coordinates are computed as a function
	public boolean is_autoplot_var   = false;  // the autoplot variable ("x" or "t")
	public boolean is_autoincr_var   = false;  // the autoincrement variable ("s")
	public boolean is_displayed      = false;  // true if coordinates are displayed
	public Complex coordinates       = null;   // current x,y coordinate values
	public String  expression        = null;   // a string expression (for functions)
	public Node    node              = null;   // the parsed version of expression (for functions)
	public Vector  plotting_vec      = null;   // a list of points (for plotting functions)
	}


class PGC_Utility
	{
	//--------------------------------------------------
	// Convert string to number: return null on failure
	//--------------------------------------------------
	public static Double DoubleValue(String s)
		{
		Double result = null;
		if (s != null) try { result = Double.valueOf(s); } catch (NumberFormatException e) { }
		return (result);
		}
	public static Integer IntegerValue(String s)
		{
		Integer result = null;
		if (s != null) try { result = Integer.valueOf(s); } catch (NumberFormatException e) { }
		return (result);
		}
	public static Integer IntegerValue(String s, int radix)
		{
		Integer result = null;
		if (s != null) try { result = Integer.valueOf(s, radix); } catch (NumberFormatException e) { }
		return (result);
		}

	//--------------------------------------------------
	// Convert string to number: return 0 on failure
	//--------------------------------------------------
	public static double DValue(String s)
		{
		double result;
		Double d = DoubleValue(s);
		if (d == null)
			result = 0.0;
		else
			result = d.doubleValue();
		return (result);
		}
	public static int IValue(String s)
		{
		int result;
		Integer i = IntegerValue(s);
		if (i == null)
			result = 0;
		else
			result = i.intValue();
		return (result);
		}
	public static int IValue(String s, int radix)
		{
		int result;
		Integer i = IntegerValue(s, radix);
		if (i == null)
			result = 0;
		else
			result = i.intValue();
		return (result);
		}
	}



//========================================================================
//========================================================================
// Node
//========================================================================
//========================================================================

class Node
	{
	public int     kind;            // KIND
	public int     id    = 0;       // TOKEN_TYPE_ID
	public Complex value = null;
	public Node    expr0 = null;
	public Node    expr1 = null;

	public static final int KIND_UNDEFINED   =   0;
	public static final int KIND_NUMBER      =   1;
	public static final int KIND_VAR         =   2;
	public static final int KIND_FUNC        =   3;
	public static final int KIND_NEG         =   4;
	public static final int KIND_REC         =   5;
	public static final int KIND_COMMA       =   6;
	public static final int KIND_OR          =   7;
	public static final int KIND_AND         =   8;
	public static final int KIND_EQ          =   9;
	public static final int KIND_NE          =  10;
	public static final int KIND_LT          =  11;
	public static final int KIND_LE          =  12;
	public static final int KIND_GT          =  13;
	public static final int KIND_GE          =  14;
	public static final int KIND_DOT         =  15;
	public static final int KIND_CROSS       =  16;
	public static final int KIND_ADD         =  17;
	public static final int KIND_SUB         =  18;
	public static final int KIND_MUL         =  19;
	public static final int KIND_DIV         =  20;
	public static final int KIND_REM         =  21;
	public static final int KIND_MOD         =  22;
	public static final int KIND_POW         =  23;

	public static final int ID_I             = PGC_Eval.TOKEN_TYPE_ID_I;
	public static final int ID_J             = PGC_Eval.TOKEN_TYPE_ID_J;
	public static final int ID_PI            = PGC_Eval.TOKEN_TYPE_ID_PI;
	public static final int ID_E             = PGC_Eval.TOKEN_TYPE_ID_E;
	public static final int ID_INFINITY      = PGC_Eval.TOKEN_TYPE_ID_INFINITY;
	public static final int ID_NAN           = PGC_Eval.TOKEN_TYPE_ID_NAN;
	public static final int ID_X             = PGC_Eval.TOKEN_TYPE_ID_X;
	public static final int ID_T             = PGC_Eval.TOKEN_TYPE_ID_T;
	public static final int ID_ABS           = PGC_Eval.TOKEN_TYPE_ID_ABS;
	public static final int ID_ARG           = PGC_Eval.TOKEN_TYPE_ID_ARG;
	public static final int ID_CONJ          = PGC_Eval.TOKEN_TYPE_ID_CONJ;
	public static final int ID_SIGN          = PGC_Eval.TOKEN_TYPE_ID_SIGN;
	public static final int ID_NOT           = PGC_Eval.TOKEN_TYPE_ID_NOT;
	public static final int ID_INT           = PGC_Eval.TOKEN_TYPE_ID_INT;
	public static final int ID_CEIL          = PGC_Eval.TOKEN_TYPE_ID_CEIL;
	public static final int ID_FLOOR         = PGC_Eval.TOKEN_TYPE_ID_FLOOR;
	public static final int ID_SQRT          = PGC_Eval.TOKEN_TYPE_ID_SQRT;
	public static final int ID_EXP           = PGC_Eval.TOKEN_TYPE_ID_EXP;
	public static final int ID_LOG           = PGC_Eval.TOKEN_TYPE_ID_LOG;
	public static final int ID_SIN           = PGC_Eval.TOKEN_TYPE_ID_SIN;
	public static final int ID_COS           = PGC_Eval.TOKEN_TYPE_ID_COS;
	public static final int ID_TAN           = PGC_Eval.TOKEN_TYPE_ID_TAN;
	public static final int ID_COT           = PGC_Eval.TOKEN_TYPE_ID_COT;
	public static final int ID_SEC           = PGC_Eval.TOKEN_TYPE_ID_SEC;
	public static final int ID_CSC           = PGC_Eval.TOKEN_TYPE_ID_CSC;
	public static final int ID_SINH          = PGC_Eval.TOKEN_TYPE_ID_SINH;
	public static final int ID_COSH          = PGC_Eval.TOKEN_TYPE_ID_COSH;
	public static final int ID_TANH          = PGC_Eval.TOKEN_TYPE_ID_TANH;
	public static final int ID_COTH          = PGC_Eval.TOKEN_TYPE_ID_COTH;
	public static final int ID_SECH          = PGC_Eval.TOKEN_TYPE_ID_SECH;
	public static final int ID_CSCH          = PGC_Eval.TOKEN_TYPE_ID_CSCH;
	public static final int ID_ASIN          = PGC_Eval.TOKEN_TYPE_ID_ASIN;
	public static final int ID_ACOS          = PGC_Eval.TOKEN_TYPE_ID_ACOS;
	public static final int ID_ATAN          = PGC_Eval.TOKEN_TYPE_ID_ATAN;
	public static final int ID_ACOT          = PGC_Eval.TOKEN_TYPE_ID_ACOT;
	public static final int ID_ASEC          = PGC_Eval.TOKEN_TYPE_ID_ASEC;
	public static final int ID_ACSC          = PGC_Eval.TOKEN_TYPE_ID_ACSC;
	public static final int ID_ASINH         = PGC_Eval.TOKEN_TYPE_ID_ASINH;
	public static final int ID_ACOSH         = PGC_Eval.TOKEN_TYPE_ID_ACOSH;
	public static final int ID_ATANH         = PGC_Eval.TOKEN_TYPE_ID_ATANH;
	public static final int ID_ACOTH         = PGC_Eval.TOKEN_TYPE_ID_ACOTH;
	public static final int ID_ASECH         = PGC_Eval.TOKEN_TYPE_ID_ASECH;
	public static final int ID_ACSCH         = PGC_Eval.TOKEN_TYPE_ID_ACSCH;
	public static final int ID_RE            = PGC_Eval.TOKEN_TYPE_ID_RE;
	public static final int ID_IM            = PGC_Eval.TOKEN_TYPE_ID_IM;
	public static final int ID_REALONLY      = PGC_Eval.TOKEN_TYPE_ID_REALONLY;

	public static Node NewUndefined()                  { Node n = new Node(); n.kind = KIND_UNDEFINED;                               return (n); }
	public static Node NewNumber(Complex value)        { Node n = new Node(); n.kind = KIND_NUMBER;    n.value = value;              return (n); }
	public static Node NewNumber(double re)            { Node n = new Node(); n.kind = KIND_NUMBER;    n.value = Complex.Number(re); return (n); }
	public static Node NewVar(int id)                  { Node n = new Node(); n.kind = KIND_VAR;       n.id = id;                    return (n); }
	public static Node NewFunc(int id, Node e0)        { Node n = new Node(); n.kind = KIND_FUNC;      n.id = id; n.expr0 = e0;      return (n); }
	public static Node NewNeg(Node e0)                 { Node n = new Node(); n.kind = KIND_NEG;       n.expr0 = e0;                 return (n); }
	public static Node NewRec(Node e0)                 { Node n = new Node(); n.kind = KIND_REC;       n.expr0 = e0;                 return (n); }
	public static Node NewComma(Node e0, Node e1)      { Node n = new Node(); n.kind = KIND_COMMA;     n.expr0 = e0; n.expr1 = e1;   return (n); }
	public static Node NewOr(Node e0, Node e1)         { Node n = new Node(); n.kind = KIND_OR;        n.expr0 = e0; n.expr1 = e1;   return (n); }
	public static Node NewAnd(Node e0, Node e1)        { Node n = new Node(); n.kind = KIND_AND;       n.expr0 = e0; n.expr1 = e1;   return (n); }
	public static Node NewEQ(Node e0, Node e1)         { Node n = new Node(); n.kind = KIND_EQ;        n.expr0 = e0; n.expr1 = e1;   return (n); }
	public static Node NewNE(Node e0, Node e1)         { Node n = new Node(); n.kind = KIND_NE;        n.expr0 = e0; n.expr1 = e1;   return (n); }
	public static Node NewLT(Node e0, Node e1)         { Node n = new Node(); n.kind = KIND_LT;        n.expr0 = e0; n.expr1 = e1;   return (n); }
	public static Node NewLE(Node e0, Node e1)         { Node n = new Node(); n.kind = KIND_LE;        n.expr0 = e0; n.expr1 = e1;   return (n); }
	public static Node NewGT(Node e0, Node e1)         { Node n = new Node(); n.kind = KIND_GT;        n.expr0 = e0; n.expr1 = e1;   return (n); }
	public static Node NewGE(Node e0, Node e1)         { Node n = new Node(); n.kind = KIND_GE;        n.expr0 = e0; n.expr1 = e1;   return (n); }
	public static Node NewDot(Node e0, Node e1)        { Node n = new Node(); n.kind = KIND_DOT;       n.expr0 = e0; n.expr1 = e1;   return (n); }
	public static Node NewCross(Node e0, Node e1)      { Node n = new Node(); n.kind = KIND_CROSS;     n.expr0 = e0; n.expr1 = e1;   return (n); }
	public static Node NewAdd(Node e0, Node e1)        { Node n = new Node(); n.kind = KIND_ADD;       n.expr0 = e0; n.expr1 = e1;   return (n); }
	public static Node NewSub(Node e0, Node e1)        { Node n = new Node(); n.kind = KIND_SUB;       n.expr0 = e0; n.expr1 = e1;   return (n); }
	public static Node NewMul(Node e0, Node e1)        { Node n = new Node(); n.kind = KIND_MUL;       n.expr0 = e0; n.expr1 = e1;   return (n); }
	public static Node NewDiv(Node e0, Node e1)        { Node n = new Node(); n.kind = KIND_DIV;       n.expr0 = e0; n.expr1 = e1;   return (n); }
	public static Node NewRem(Node e0, Node e1)        { Node n = new Node(); n.kind = KIND_REM;       n.expr0 = e0; n.expr1 = e1;   return (n); }
	public static Node NewMod(Node e0, Node e1)        { Node n = new Node(); n.kind = KIND_MOD;       n.expr0 = e0; n.expr1 = e1;   return (n); }
	public static Node NewPow(Node e0, Node e1)        { Node n = new Node(); n.kind = KIND_POW;       n.expr0 = e0; n.expr1 = e1;   return (n); }

	public String Image()
		{
		return Image(0, false);
		}
	public String Image(int parent_level, boolean is_right_child)
		{
		String result;
		int level;
		switch (kind)
			{
			default             : { level = 9; result = " INTERNAL-ERROR "; break; }
			case KIND_UNDEFINED : { level = 9; result = " NaN "; break; }
			case KIND_VAR       : { level = 9; result = PGC_Eval.LexIdName(id); break; }
			case KIND_FUNC      : { level = 8; result = PGC_Eval.LexIdName(id) + " " + expr0.Image(level, false); break; }
			case KIND_NEG       : { level = 8; result = "-" + expr0.Image(level, false); break; }
			case KIND_REC       : { level = 6; result = "1"                       + "/"       + expr1.Image(level, true); break; }
			case KIND_COMMA     : { level = 0; result = expr0.Image(level, false) + ","       + expr1.Image(level, true); break; }
			case KIND_OR        : { level = 1; result = expr0.Image(level, false) + " or "    + expr1.Image(level, true); break; }
			case KIND_AND       : { level = 2; result = expr0.Image(level, false) + " and "   + expr1.Image(level, true); break; }
			case KIND_EQ        : { level = 3; result = expr0.Image(level, false) + " == "    + expr1.Image(level, true); break; }
			case KIND_NE        : { level = 3; result = expr0.Image(level, false) + " != "    + expr1.Image(level, true); break; }
			case KIND_LT        : { level = 3; result = expr0.Image(level, false) + " < "     + expr1.Image(level, true); break; }
			case KIND_LE        : { level = 3; result = expr0.Image(level, false) + " <= "    + expr1.Image(level, true); break; }
			case KIND_GT        : { level = 3; result = expr0.Image(level, false) + " > "     + expr1.Image(level, true); break; }
			case KIND_GE        : { level = 3; result = expr0.Image(level, false) + " >= "    + expr1.Image(level, true); break; }
			case KIND_DOT       : { level = 4; result = expr0.Image(level, false) + " dot "   + expr1.Image(level, true); break; }
			case KIND_CROSS     : { level = 4; result = expr0.Image(level, false) + " cross " + expr1.Image(level, true); break; }
			case KIND_ADD       : { level = 5; result = expr0.Image(level, false) + "+"       + expr1.Image(level, true); break; }
			case KIND_SUB       : { level = 5; result = expr0.Image(level, false) + "-"       + expr1.Image(level, true); break; }
			case KIND_MUL       : { level = 6; result = expr0.Image(level, false) + "*"       + expr1.Image(level, true); break; }
			case KIND_DIV       : { level = 6; result = expr0.Image(level, false) + "/"       + expr1.Image(level, true); break; }
			case KIND_REM       : { level = 6; result = expr0.Image(level, false) + " rem "   + expr1.Image(level, true); break; }
			case KIND_MOD       : { level = 6; result = expr0.Image(level, false) + " mod "   + expr1.Image(level, true); break; }
			case KIND_POW       : { level = 7; result = expr0.Image(level, false) + "^"       + expr1.Image(level, true); break; }
			case KIND_NUMBER    :
				{
				if (value.im == 0.0)
					{
					level = 9;
					result = "" + ImageDouble(value.re);
					}
				else
					{
					level = 0;
					result = ImageDouble(value.re) + "," + ImageDouble(value.im);
					}
				break;
				}
			}
		if (level < parent_level)
			{
			result = "(" + result + ")";
			}
		else if ((level == parent_level) && (is_right_child))
			{
			result = "(" + result + ")";
			}
		return (result);
		}
	public String ImageDouble(double d)
		{
		String result;
		if (d == (double)(int)d)
			result = Integer.toString((int)d);
		else
			result = Double.toString(d);
		return (result);
		}
	public static Node Copy(Node n)
		{
		Node result = null;
		if (n != null)
			{
			result       = new Node();
			result.kind  = n.kind;
			result.id    = n.id;
			result.value = n.value;
			result.expr0 = Copy(n.expr0);
			result.expr1 = Copy(n.expr1);
			}
		return (result);
		}
	public Node Copy()
		{
		return (Copy(this));
		}
	public static boolean IsCopy(Node n0, Node n1)
		{
		boolean result;
		if ((n0 == null) && (n1 == null))
			result = true;
		else if ((n0 == null) || (n1 == null))
			result = false;
		else
			{
			boolean value_is_copy = false;

			if ((n0.value == null) && (n1.value == null))
				value_is_copy = true;
			else if ((n0.value == null) || (n1.value == null))
				value_is_copy = false;
			else
				value_is_copy = (n0.value.re == n1.value.re) && (n0.value.im == n1.value.im);

			result =
				(n0.kind == n1.kind) &&
				(n0.id == n1.id) &&
				(value_is_copy) &&
				IsCopy(n0.expr0, n1.expr0) &&
				IsCopy(n0.expr1, n1.expr1);
			}
		return (result);
		}
	public Node Derivative()
		{
		//-------------------------------------------------------------
		// Return the derivative expression (with respect to x or t).
		//-------------------------------------------------------------
		Node result = null;
		switch (kind)
			{
			case KIND_NUMBER:
				{
				result = NewNumber(0);
				break;
				}
			case KIND_VAR:
				{
				if ((id == ID_X) || (id == ID_T))
					result = NewNumber(1);
				else
					result = NewNumber(0);
				break;
				}
			case KIND_FUNC:
				{
				boolean chain = true;
				switch (id)
					{
					case ID_ARG:
							{
							Node slope = NewDiv(NewFunc(ID_IM, expr0.Copy()), NewFunc(ID_RE, expr0.Copy()));
							Node d_slope = slope.Derivative();
							result = NewDiv(d_slope, NewAdd(NewNumber(1), NewPow(slope, NewNumber(2))));
							chain = false;
							break;
							}
					case ID_ABS:      { result = NewDiv(this.Copy(), expr0.Copy()); break; }
					case ID_CONJ:     { result = NewSub(NewFunc(ID_RE, expr0.Derivative()), NewMul(NewVar(ID_I), NewFunc(ID_IM, expr0.Derivative()))); chain = false; break; }
					case ID_SIGN:     { result = NewNumber(0); break; }
					case ID_NOT:      { result = NewNumber(0); break; }
					case ID_INT:      { result = NewNumber(0); break; }
					case ID_CEIL:     { result = NewNumber(0); break; }
					case ID_FLOOR:    { result = NewNumber(0); break; }
					case ID_SQRT:     { result = NewDiv(NewNumber(0.5), this.Copy()); break; }
					case ID_EXP:      { result = this.Copy(); break; }
					case ID_LOG:      { result = NewDiv(NewNumber(1), expr0.Copy()); break; }
					case ID_SIN:      { result = NewFunc(ID_COS, expr0.Copy()); break; }
					case ID_COS:      { result = NewNeg(NewFunc(ID_SIN, expr0.Copy())); break; }
					case ID_TAN:      { result = NewPow(NewFunc(ID_SEC, expr0.Copy()), NewNumber(2)); break; }
					case ID_COT:      { result = NewNeg(NewPow(NewFunc(ID_CSC, expr0.Copy()), NewNumber(2))); break; }
					case ID_SEC:      { result = NewMul(NewFunc(ID_SEC, expr0.Copy()), NewFunc(ID_TAN, expr0.Copy())); break; }
					case ID_CSC:      { result = NewMul(NewNeg(NewFunc(ID_CSC, expr0.Copy())), NewFunc(ID_COT, expr0.Copy())); break; }
					case ID_SINH:     { result = NewFunc(ID_COSH, expr0.Copy()); break; }
					case ID_COSH:     { result = NewFunc(ID_SINH, expr0.Copy()); break; }
					case ID_TANH:     { result = NewPow(NewFunc(ID_SECH, expr0.Copy()), NewNumber(2)); break; }
					case ID_COTH:     { result = NewNeg(NewPow(NewFunc(ID_CSCH, expr0.Copy()), NewNumber(2))); break; }
					case ID_SECH:     { result = NewMul(NewNeg(NewFunc(ID_SECH, expr0.Copy())), NewFunc(ID_TANH, expr0.Copy())); break; }
					case ID_CSCH:     { result = NewMul(NewNeg(NewFunc(ID_CSCH, expr0.Copy())), NewFunc(ID_COTH, expr0.Copy())); break; }
					case ID_ASIN:     { result = NewPow(NewFunc(ID_SQRT, NewSub(NewNumber(1), NewPow(expr0.Copy(), NewNumber(2)))), NewNumber(-1)); break; }
					case ID_ACOS:     { result = NewNeg(NewPow(NewFunc(ID_SQRT, NewSub(NewNumber(1), NewPow(expr0.Copy(), NewNumber(2)))), NewNumber(-1))); break; }
					case ID_ATAN:     { result = NewPow(NewAdd(NewNumber(1), NewPow(expr0.Copy(), NewNumber(2))), NewNumber(-1)); break; }
					case ID_ACOT:     { result = NewNeg(NewPow(NewAdd(NewNumber(1), NewPow(expr0.Copy(), NewNumber(2))), NewNumber(-1))); break; }
					case ID_ASEC:     { result = NewPow(NewMul(expr0.Copy(), NewFunc(ID_SQRT, NewSub(NewPow(expr0.Copy(), NewNumber(2)), NewNumber(1)))), NewNumber(-1)); break; }
					case ID_ACSC:     { result = NewNeg(NewPow(NewMul(expr0.Copy(), NewFunc(ID_SQRT, NewSub(NewPow(expr0.Copy(), NewNumber(2)), NewNumber(1)))), NewNumber(-1))); break; }
					case ID_ASINH:    { result = NewPow(NewFunc(ID_SQRT, NewAdd(NewPow(expr0.Copy(), NewNumber(2)), NewNumber(1))), NewNumber(-1)); break; }
					case ID_ACOSH:    { result = NewPow(NewFunc(ID_SQRT, NewSub(NewPow(expr0.Copy(), NewNumber(2)), NewNumber(1))), NewNumber(-1)); break; }
					case ID_ATANH:    { result = NewPow(NewSub(NewNumber(1), NewPow(expr0.Copy(), NewNumber(2))), NewNumber(-1)); break; }
					case ID_ACOTH:    { result = NewPow(NewSub(NewNumber(1), NewPow(expr0.Copy(), NewNumber(2))), NewNumber(-1)); break; }
					case ID_ASECH:    { result = NewPow(NewFunc(ID_SQRT, NewSub(NewPow(expr0.Copy(), NewNumber(2)), NewPow(expr0.Copy(), NewNumber(4)))), NewNumber(-1)); break; }
					case ID_ACSCH:    { result = NewPow(NewFunc(ID_SQRT, NewAdd(NewPow(expr0.Copy(), NewNumber(2)), NewPow(expr0.Copy(), NewNumber(4)))), NewNumber(-1)); break; }
					case ID_RE:       { result = NewFunc(ID_RE, expr0.Derivative()); chain = false; break; }
					case ID_IM:       { result = NewFunc(ID_IM, expr0.Derivative()); chain = false; break; }
					case ID_REALONLY: { result = NewFunc(ID_REALONLY, expr0.Derivative()); chain = false; break; }
					}
				if (chain && (result != null))
					{
					// Apply chain rule
					result = NewMul(result, expr0.Derivative());
					}
				break;
				}
			case KIND_NEG:
				{
				Node d0 = expr0.Derivative();
				result = NewNeg(d0);
				break;
				}
			case KIND_REC:
				{
				Node d0 = expr0.Derivative();
				result = NewDiv(NewNeg(d0), NewPow(expr0.Copy(), NewNumber(2)));
				break;
				}
			case KIND_COMMA:
				{
				Node d0 = NewFunc(ID_RE, expr0.Derivative());
				Node d1 = NewFunc(ID_RE, expr1.Derivative());
				result = NewAdd(d0, NewMul(NewVar(ID_I), d1));
				break;
				}
			case KIND_AND:
			case KIND_OR:
			case KIND_EQ:
			case KIND_NE:
			case KIND_LT:
			case KIND_LE:
			case KIND_GT:
			case KIND_GE:
			case KIND_DOT:
			case KIND_CROSS:
				{
				result = NewNumber(0); // ???
				break;
				}
			case KIND_ADD:
				{
				Node d0 = expr0.Derivative();
				Node d1 = expr1.Derivative();
				result = NewAdd(d0, d1);
				break;
				}
			case KIND_SUB:
				{
				Node d0 = expr0.Derivative();
				Node d1 = expr1.Derivative();
				result = NewSub(d0, d1);
				break;
				}
			case KIND_MUL:
				{
				Node d0 = expr0.Derivative();
				Node d1 = expr1.Derivative();
				result = NewAdd(NewMul(expr0.Copy(), d1), NewMul(d0, expr1.Copy()));
				break;
				}
			case KIND_DIV:
				{
				Node d0 = expr0.Derivative();
				Node d1 = expr1.Derivative();
				Node diff = NewSub(NewMul(expr1.Copy(), d0), NewMul(d1, expr0.Copy()));
				result = NewDiv(diff, NewPow(expr1.Copy(), NewNumber(2)));
				break;
				}
			case KIND_REM:
			case KIND_MOD:
				{
				break;
				}
			case KIND_POW:
				{
				Node d0 = expr0.Derivative();
				Node d1 = expr1.Derivative();
				Node part0 = NewMul(NewMul(expr1.Copy(), NewPow(expr0.Copy(), NewSub(expr1.Copy(), NewNumber(1)))), d0);
				Node part1 = NewMul(NewMul(this.Copy(), NewFunc(ID_LOG, expr0.Copy())), d1);
				result = NewAdd(part0, part1);
				break;
				}
			default:
				{
				break;
				}
			}
		if (result == null)
			result = NewUndefined();
		return (result);
		}
	}



//========================================================================
//========================================================================
// PGC_Eval
//========================================================================
//========================================================================

class PGC_Eval
	{
	private PGC         owner;
	public  String      input;                                   // the input expression
	public  Node        node;                                    // the expression tree, or null
	public  Complex     value;                                   // the calculated value of the input expression, or null
	public  String      message;                                 // an error message, or the empty string if no error
	public  boolean     empty;                                   // true if the input expression is the empty string (or all blanks)
	public  boolean     uses_autoplot;                           // true if the input expression uses either autoplot variable
	public  boolean     uses_autoplot_x;                         // true if the input expression uses autoplot variable "x"
	public  boolean     uses_autoplot_t;                         // true if the input expression uses autoplot variable "t"
	public  boolean[]   uses_input = new boolean[PGC.VAR_COUNT]; // true if the input expression uses the input variable
	public  boolean[]   allowed = new boolean[PGC.VAR_COUNT];    // true if evaluation of the user-defined function is allowed
	public  String      recursive;                               // name of a recursive function (or null if none)

	public PGC_Eval(PGC own)
		{
		owner = own;

		for (int i = 0; i < allowed.length; i++)
			allowed[i] = true;
		}
	public void Evaluate(String in)
		{
		//---------------------------------------------------
		// Evaluate the input string as an expression.
		// After evaluation, this.value contains the result.
		//---------------------------------------------------
		if (in == null) in = "";

		input            = in;
		message          = "";
		uses_autoplot    = false;
		uses_autoplot_x  = false;
		uses_autoplot_t  = false;

		for (int i = 0; i < uses_input.length; i++)
			uses_input[i] = false;

		LexInitialize();
		empty = (token_type == TOKEN_TYPE_EOS);
		recursive = null;
		node = ParseExpression();
		value = EvalNode(node);

		if (recursive != null)
			ParseError("Recursive definition with " + recursive);
		}
	public void Evaluate(Node n)
		{
		//---------------------------------------------------
		// Evaluate the node.
		// After evaluation, this.value contains the result.
		//---------------------------------------------------
		message          = "";
		uses_autoplot    = false;
		uses_autoplot_x  = false;
		uses_autoplot_t  = false;

		for (int i = 0; i < uses_input.length; i++)
			uses_input[i] = false;

		empty = false;
		recursive = null;
		node = n;

		value = EvalNode(n);

		if (recursive != null)
			ParseError("Recursive definition with " + recursive);
		}
	public void ReEvaluate()
		{
		//---------------------------------------------------
		// Re-evaluate the current node.
		// This is a special "faster" version that
		// calculates only this.value, and does not
		// set any other memebers.
		//---------------------------------------------------

		value = EvalNode(node);
		}

	//----------------------------------------------------------------
	// PGC_Eval: Lexical
	//----------------------------------------------------------------

	private int              pos;
	private int              ch;
	private String           token;
	private int              token_pos;
	private int              token_type;
	private int              CHAR_EOS = -1;

	public static final int TOKEN_TYPE_UNDEFINED          = 0;
	public static final int TOKEN_TYPE_EOS                = 1;
	public static final int TOKEN_TYPE_NUMBER             = 2;
	public static final int TOKEN_TYPE_SINGLE             = 3;
	public static final int TOKEN_TYPE_SYMBOL             = 0x0800;
	public static final int TOKEN_TYPE_SYMBOL_COMMA       = TOKEN_TYPE_SYMBOL |  1;
	public static final int TOKEN_TYPE_SYMBOL_ID_DOT      = TOKEN_TYPE_SYMBOL |  2;
	public static final int TOKEN_TYPE_SYMBOL_ID_CROSS    = TOKEN_TYPE_SYMBOL |  3;
	public static final int TOKEN_TYPE_SYMBOL_PLUS        = TOKEN_TYPE_SYMBOL |  4;
	public static final int TOKEN_TYPE_SYMBOL_MINUS       = TOKEN_TYPE_SYMBOL |  5;
	public static final int TOKEN_TYPE_SYMBOL_MUL         = TOKEN_TYPE_SYMBOL |  6;
	public static final int TOKEN_TYPE_SYMBOL_DIV         = TOKEN_TYPE_SYMBOL |  7;
	public static final int TOKEN_TYPE_SYMBOL_ID_REM      = TOKEN_TYPE_SYMBOL |  8;
	public static final int TOKEN_TYPE_SYMBOL_ID_MOD      = TOKEN_TYPE_SYMBOL |  9;
	public static final int TOKEN_TYPE_SYMBOL_POW         = TOKEN_TYPE_SYMBOL | 10;
	public static final int TOKEN_TYPE_SYMBOL_PERIOD      = TOKEN_TYPE_SYMBOL | 11;
	public static final int TOKEN_TYPE_SYMBOL_LPAREN      = TOKEN_TYPE_SYMBOL | 12;
	public static final int TOKEN_TYPE_SYMBOL_RPAREN      = TOKEN_TYPE_SYMBOL | 13;
	public static final int TOKEN_TYPE_SYMBOL_BAR         = TOKEN_TYPE_SYMBOL | 14;
	public static final int TOKEN_TYPE_SYMBOL_EQ          = TOKEN_TYPE_SYMBOL | 15;
	public static final int TOKEN_TYPE_SYMBOL_NE          = TOKEN_TYPE_SYMBOL | 16;
	public static final int TOKEN_TYPE_SYMBOL_LT          = TOKEN_TYPE_SYMBOL | 17;
	public static final int TOKEN_TYPE_SYMBOL_LE          = TOKEN_TYPE_SYMBOL | 18;
	public static final int TOKEN_TYPE_SYMBOL_GT          = TOKEN_TYPE_SYMBOL | 19;
	public static final int TOKEN_TYPE_SYMBOL_GE          = TOKEN_TYPE_SYMBOL | 20;
	public static final int TOKEN_TYPE_SYMBOL_ID_AND      = TOKEN_TYPE_SYMBOL | 21;
	public static final int TOKEN_TYPE_SYMBOL_ID_OR       = TOKEN_TYPE_SYMBOL | 22;
	public static final int TOKEN_TYPE_ID                 = 0x4000;
	public static final int TOKEN_TYPE_ID_CLASS_VAR       = TOKEN_TYPE_ID | 0x1000;
	public static final int TOKEN_TYPE_ID_CLASS_FUNC      = TOKEN_TYPE_ID | 0x2000;
	public static final int TOKEN_TYPE_ID_I               = TOKEN_TYPE_ID_CLASS_VAR  |  1;
	public static final int TOKEN_TYPE_ID_J               = TOKEN_TYPE_ID_CLASS_VAR  |  2;
	public static final int TOKEN_TYPE_ID_PI              = TOKEN_TYPE_ID_CLASS_VAR  |  3;
	public static final int TOKEN_TYPE_ID_E               = TOKEN_TYPE_ID_CLASS_VAR  |  4;
	public static final int TOKEN_TYPE_ID_INFINITY        = TOKEN_TYPE_ID_CLASS_VAR  |  5;
	public static final int TOKEN_TYPE_ID_NAN             = TOKEN_TYPE_ID_CLASS_VAR  |  6;
	public static final int TOKEN_TYPE_ID_A               = TOKEN_TYPE_ID_CLASS_VAR  |  7;
	public static final int TOKEN_TYPE_ID_AX              = TOKEN_TYPE_ID_CLASS_VAR  |  8;
	public static final int TOKEN_TYPE_ID_AY              = TOKEN_TYPE_ID_CLASS_VAR  |  9;
	public static final int TOKEN_TYPE_ID_B               = TOKEN_TYPE_ID_CLASS_VAR  | 10;
	public static final int TOKEN_TYPE_ID_BX              = TOKEN_TYPE_ID_CLASS_VAR  | 11;
	public static final int TOKEN_TYPE_ID_BY              = TOKEN_TYPE_ID_CLASS_VAR  | 12;
	public static final int TOKEN_TYPE_ID_C               = TOKEN_TYPE_ID_CLASS_VAR  | 13;
	public static final int TOKEN_TYPE_ID_CX              = TOKEN_TYPE_ID_CLASS_VAR  | 14;
	public static final int TOKEN_TYPE_ID_CY              = TOKEN_TYPE_ID_CLASS_VAR  | 15;
	public static final int TOKEN_TYPE_ID_D               = TOKEN_TYPE_ID_CLASS_VAR  | 16;
	public static final int TOKEN_TYPE_ID_DX              = TOKEN_TYPE_ID_CLASS_VAR  | 17;
	public static final int TOKEN_TYPE_ID_DY              = TOKEN_TYPE_ID_CLASS_VAR  | 18;
	public static final int TOKEN_TYPE_ID_F               = TOKEN_TYPE_ID_CLASS_VAR  | 19;
	public static final int TOKEN_TYPE_ID_FX              = TOKEN_TYPE_ID_CLASS_VAR  | 20;
	public static final int TOKEN_TYPE_ID_FY              = TOKEN_TYPE_ID_CLASS_VAR  | 21;
	public static final int TOKEN_TYPE_ID_G               = TOKEN_TYPE_ID_CLASS_VAR  | 22;
	public static final int TOKEN_TYPE_ID_GX              = TOKEN_TYPE_ID_CLASS_VAR  | 23;
	public static final int TOKEN_TYPE_ID_GY              = TOKEN_TYPE_ID_CLASS_VAR  | 24;
	public static final int TOKEN_TYPE_ID_H               = TOKEN_TYPE_ID_CLASS_VAR  | 25;
	public static final int TOKEN_TYPE_ID_HX              = TOKEN_TYPE_ID_CLASS_VAR  | 26;
	public static final int TOKEN_TYPE_ID_HY              = TOKEN_TYPE_ID_CLASS_VAR  | 27;
	public static final int TOKEN_TYPE_ID_K               = TOKEN_TYPE_ID_CLASS_VAR  | 28;
	public static final int TOKEN_TYPE_ID_KX              = TOKEN_TYPE_ID_CLASS_VAR  | 29;
	public static final int TOKEN_TYPE_ID_KY              = TOKEN_TYPE_ID_CLASS_VAR  | 30;
	public static final int TOKEN_TYPE_ID_M               = TOKEN_TYPE_ID_CLASS_VAR  | 31;
	public static final int TOKEN_TYPE_ID_MX              = TOKEN_TYPE_ID_CLASS_VAR  | 32;
	public static final int TOKEN_TYPE_ID_MY              = TOKEN_TYPE_ID_CLASS_VAR  | 33;
	public static final int TOKEN_TYPE_ID_N               = TOKEN_TYPE_ID_CLASS_VAR  | 34;
	public static final int TOKEN_TYPE_ID_NX              = TOKEN_TYPE_ID_CLASS_VAR  | 35;
	public static final int TOKEN_TYPE_ID_NY              = TOKEN_TYPE_ID_CLASS_VAR  | 36;
	public static final int TOKEN_TYPE_ID_X               = TOKEN_TYPE_ID_CLASS_VAR  | 37;
	public static final int TOKEN_TYPE_ID_T               = TOKEN_TYPE_ID_CLASS_VAR  | 38;
	public static final int TOKEN_TYPE_ID_S               = TOKEN_TYPE_ID_CLASS_VAR  | 39;
	public static final int TOKEN_TYPE_ID_Z               = TOKEN_TYPE_ID_CLASS_VAR  | 40;
	public static final int TOKEN_TYPE_ID_ABS             = TOKEN_TYPE_ID_CLASS_FUNC |  1;
	public static final int TOKEN_TYPE_ID_ARG             = TOKEN_TYPE_ID_CLASS_FUNC |  2;
	public static final int TOKEN_TYPE_ID_CONJ            = TOKEN_TYPE_ID_CLASS_FUNC |  3;
	public static final int TOKEN_TYPE_ID_SIGN            = TOKEN_TYPE_ID_CLASS_FUNC |  4;
	public static final int TOKEN_TYPE_ID_NOT             = TOKEN_TYPE_ID_CLASS_FUNC |  5;
	public static final int TOKEN_TYPE_ID_INT             = TOKEN_TYPE_ID_CLASS_FUNC |  6;
	public static final int TOKEN_TYPE_ID_CEIL            = TOKEN_TYPE_ID_CLASS_FUNC |  7;
	public static final int TOKEN_TYPE_ID_FLOOR           = TOKEN_TYPE_ID_CLASS_FUNC |  8;
	public static final int TOKEN_TYPE_ID_SQRT            = TOKEN_TYPE_ID_CLASS_FUNC |  9;
	public static final int TOKEN_TYPE_ID_EXP             = TOKEN_TYPE_ID_CLASS_FUNC | 10;
	public static final int TOKEN_TYPE_ID_LOG             = TOKEN_TYPE_ID_CLASS_FUNC | 11;
	public static final int TOKEN_TYPE_ID_SIN             = TOKEN_TYPE_ID_CLASS_FUNC | 12;
	public static final int TOKEN_TYPE_ID_COS             = TOKEN_TYPE_ID_CLASS_FUNC | 13;
	public static final int TOKEN_TYPE_ID_TAN             = TOKEN_TYPE_ID_CLASS_FUNC | 14;
	public static final int TOKEN_TYPE_ID_COT             = TOKEN_TYPE_ID_CLASS_FUNC | 15;
	public static final int TOKEN_TYPE_ID_SEC             = TOKEN_TYPE_ID_CLASS_FUNC | 16;
	public static final int TOKEN_TYPE_ID_CSC             = TOKEN_TYPE_ID_CLASS_FUNC | 17;
	public static final int TOKEN_TYPE_ID_SINH            = TOKEN_TYPE_ID_CLASS_FUNC | 18;
	public static final int TOKEN_TYPE_ID_COSH            = TOKEN_TYPE_ID_CLASS_FUNC | 19;
	public static final int TOKEN_TYPE_ID_TANH            = TOKEN_TYPE_ID_CLASS_FUNC | 20;
	public static final int TOKEN_TYPE_ID_COTH            = TOKEN_TYPE_ID_CLASS_FUNC | 21;
	public static final int TOKEN_TYPE_ID_SECH            = TOKEN_TYPE_ID_CLASS_FUNC | 22;
	public static final int TOKEN_TYPE_ID_CSCH            = TOKEN_TYPE_ID_CLASS_FUNC | 23;
	public static final int TOKEN_TYPE_ID_ASIN            = TOKEN_TYPE_ID_CLASS_FUNC | 24;
	public static final int TOKEN_TYPE_ID_ACOS            = TOKEN_TYPE_ID_CLASS_FUNC | 25;
	public static final int TOKEN_TYPE_ID_ATAN            = TOKEN_TYPE_ID_CLASS_FUNC | 26;
	public static final int TOKEN_TYPE_ID_ACOT            = TOKEN_TYPE_ID_CLASS_FUNC | 27;
	public static final int TOKEN_TYPE_ID_ASEC            = TOKEN_TYPE_ID_CLASS_FUNC | 28;
	public static final int TOKEN_TYPE_ID_ACSC            = TOKEN_TYPE_ID_CLASS_FUNC | 29;
	public static final int TOKEN_TYPE_ID_ASINH           = TOKEN_TYPE_ID_CLASS_FUNC | 30;
	public static final int TOKEN_TYPE_ID_ACOSH           = TOKEN_TYPE_ID_CLASS_FUNC | 31;
	public static final int TOKEN_TYPE_ID_ATANH           = TOKEN_TYPE_ID_CLASS_FUNC | 32;
	public static final int TOKEN_TYPE_ID_ACOTH           = TOKEN_TYPE_ID_CLASS_FUNC | 33;
	public static final int TOKEN_TYPE_ID_ASECH           = TOKEN_TYPE_ID_CLASS_FUNC | 34;
	public static final int TOKEN_TYPE_ID_ACSCH           = TOKEN_TYPE_ID_CLASS_FUNC | 35;
	public static final int TOKEN_TYPE_ID_RE              = TOKEN_TYPE_ID_CLASS_FUNC | 36;
	public static final int TOKEN_TYPE_ID_IM              = TOKEN_TYPE_ID_CLASS_FUNC | 37;
	public static final int TOKEN_TYPE_ID_REALONLY        = TOKEN_TYPE_ID_CLASS_FUNC | 38;

	private void LexInitialize()
		{
		pos = 0;
		LexGetChar();
		LexGetToken();
		}
	private boolean LexIsEos()
		{
		return (ch == CHAR_EOS);
		}
	private boolean LexIsSpace()
		{
		return ((ch >= 0) && (ch <= (int)' '));
		}
	private boolean LexIsIdentifierChar()
		{
		return
			((ch >= (int)'a') && (ch <= (int)'z')) ||
			((ch >= (int)'A') && (ch <= (int)'Z')) ||
			((ch == (int)'_'));
		}
	private boolean LexIsDigit()
		{
		return ((ch >= (int)'0') && (ch <= (int)'9'));
		}
	private void LexGetChar()
		{
		if (pos >= input.length())
			ch = CHAR_EOS;
		else
			{
			ch = (int)input.charAt(pos);
			pos++;
			}
		}
	private void LexAdvance()
		{
		token += (char)ch;
		LexGetChar();
		}
	private void LexWhiteSpace()
		{
		while (true)
			{
			while (LexIsSpace()) LexGetChar();
			if (ch == (char)'{') // }
				{
				int depth = 1;
				LexGetChar();
				while (true)
					{
					if (LexIsEos()) break;
					if (ch == (char)'{') depth++;
					if (ch == (char)'}') depth--;
					LexGetChar();
					if (depth == 0) break;
					}
				continue;
				}
			break;
			}
		}
	private void LexGetToken()
		{
		token = "";
		token_type = TOKEN_TYPE_UNDEFINED;
		LexWhiteSpace();
		token_pos = pos;
		if (LexIsEos())
			{
			token_type = TOKEN_TYPE_EOS;
			}
		else
			{
			if (LexIsIdentifierChar())
				{
				while (LexIsIdentifierChar()) LexAdvance();
				token_type = LexTokenSymbol(token);
				if (token_type == TOKEN_TYPE_UNDEFINED)
					token_type = LexTokenId(token);
				}
			else if (LexIsDigit() || (ch == (char)'.'))
				{
				token_type = TOKEN_TYPE_NUMBER;
				boolean dot = (ch == (char)'.');
				while (LexIsDigit()) LexAdvance();
				if (ch == (char)'.')
					{
					LexAdvance();
					if (dot && !LexIsDigit())
						token_type = TOKEN_TYPE_SYMBOL_PERIOD;
					else
						while (LexIsDigit()) LexAdvance();
					}
				if ((token_type == TOKEN_TYPE_NUMBER) && (ch == (char)'e') || (ch == (char)'E'))
					{
					LexAdvance();
					if ((ch == (char)'+') || (ch == (char)'-'))
						LexAdvance();
					while (LexIsDigit()) LexAdvance();
					}
				}
			else if (ch == (char)'=')
				{
				LexAdvance();
				if (ch == (char)'=')
					{ LexAdvance(); token_type = TOKEN_TYPE_SYMBOL_EQ; }
				else
					{ token_type = TOKEN_TYPE_SYMBOL_EQ; }
				}
			else if (ch == (char)'!')
				{
				LexAdvance();
				if (ch == (char)'=')
					{ LexAdvance(); token_type = TOKEN_TYPE_SYMBOL_NE; }
				else
					{ token_type = TOKEN_TYPE_SINGLE; }
				}
			else if (ch == (char)'<')
				{
				LexAdvance();
				if (ch == (char)'=')
					{ LexAdvance(); token_type = TOKEN_TYPE_SYMBOL_LE; }
				else if (ch == (char)'>')
					{ LexAdvance(); token_type = TOKEN_TYPE_SYMBOL_NE; }
				else
					{ token_type = TOKEN_TYPE_SYMBOL_LT; }
				}
			else if (ch == (char)'>')
				{
				LexAdvance();
				if (ch == (char)'=')
					{ LexAdvance(); token_type = TOKEN_TYPE_SYMBOL_GE; }
				else
					{ token_type = TOKEN_TYPE_SYMBOL_GT; }
				}
			else
				{
				LexAdvance();
				token_type = LexTokenSymbol(token);
				if (token_type == TOKEN_TYPE_UNDEFINED)
					token_type = TOKEN_TYPE_SINGLE;
				}
			}
		}
	private int LexTokenSymbol(String id)
		{
		int result = TOKEN_TYPE_UNDEFINED;
		     if (id.equalsIgnoreCase(","))          result = TOKEN_TYPE_SYMBOL_COMMA;
		else if (id.equalsIgnoreCase("dot"))        result = TOKEN_TYPE_SYMBOL_ID_DOT;
		else if (id.equalsIgnoreCase("cross"))      result = TOKEN_TYPE_SYMBOL_ID_CROSS;
		else if (id.equalsIgnoreCase("+"))          result = TOKEN_TYPE_SYMBOL_PLUS;
		else if (id.equalsIgnoreCase("-"))          result = TOKEN_TYPE_SYMBOL_MINUS;
		else if (id.equalsIgnoreCase("*"))          result = TOKEN_TYPE_SYMBOL_MUL;
		else if (id.equalsIgnoreCase("/"))          result = TOKEN_TYPE_SYMBOL_DIV;
		else if (id.equalsIgnoreCase("rem"))        result = TOKEN_TYPE_SYMBOL_ID_REM;
		else if (id.equalsIgnoreCase("mod"))        result = TOKEN_TYPE_SYMBOL_ID_MOD;
		else if (id.equalsIgnoreCase("^"))          result = TOKEN_TYPE_SYMBOL_POW;
		else if (id.equalsIgnoreCase("."))          result = TOKEN_TYPE_SYMBOL_PERIOD;
		else if (id.equalsIgnoreCase("("))          result = TOKEN_TYPE_SYMBOL_LPAREN;
		else if (id.equalsIgnoreCase(")"))          result = TOKEN_TYPE_SYMBOL_RPAREN;
		else if (id.equalsIgnoreCase("|"))          result = TOKEN_TYPE_SYMBOL_BAR;
		else if (id.equalsIgnoreCase("and"))        result = TOKEN_TYPE_SYMBOL_ID_AND;
		else if (id.equalsIgnoreCase("or"))         result = TOKEN_TYPE_SYMBOL_ID_OR;
		return (result);
		}
	private int LexTokenId(String id)
		{
		int result = TOKEN_TYPE_ID;
		     if (id.equalsIgnoreCase("i"))         result = TOKEN_TYPE_ID_I;
		else if (id.equalsIgnoreCase("j"))         result = TOKEN_TYPE_ID_J;
		else if (id.equalsIgnoreCase("pi"))        result = TOKEN_TYPE_ID_PI;
		else if (id.equalsIgnoreCase("e"))         result = TOKEN_TYPE_ID_E;
		else if (id.equalsIgnoreCase("infinity"))  result = TOKEN_TYPE_ID_INFINITY;
		else if (id.equalsIgnoreCase("nan"))       result = TOKEN_TYPE_ID_NAN;
		else if (id.equalsIgnoreCase("a"))         result = TOKEN_TYPE_ID_A;
		else if (id.equalsIgnoreCase("ax"))        result = TOKEN_TYPE_ID_AX;
		else if (id.equalsIgnoreCase("ay"))        result = TOKEN_TYPE_ID_AY;
		else if (id.equalsIgnoreCase("b"))         result = TOKEN_TYPE_ID_B;
		else if (id.equalsIgnoreCase("bx"))        result = TOKEN_TYPE_ID_BX;
		else if (id.equalsIgnoreCase("by"))        result = TOKEN_TYPE_ID_BY;
		else if (id.equalsIgnoreCase("c"))         result = TOKEN_TYPE_ID_C;
		else if (id.equalsIgnoreCase("cx"))        result = TOKEN_TYPE_ID_CX;
		else if (id.equalsIgnoreCase("cy"))        result = TOKEN_TYPE_ID_CY;
		else if (id.equalsIgnoreCase("d"))         result = TOKEN_TYPE_ID_D;
		else if (id.equalsIgnoreCase("dx"))        result = TOKEN_TYPE_ID_DX;
		else if (id.equalsIgnoreCase("dy"))        result = TOKEN_TYPE_ID_DY;
		else if (id.equalsIgnoreCase("f"))         result = TOKEN_TYPE_ID_F;
		else if (id.equalsIgnoreCase("fx"))        result = TOKEN_TYPE_ID_FX;
		else if (id.equalsIgnoreCase("fy"))        result = TOKEN_TYPE_ID_FY;
		else if (id.equalsIgnoreCase("g"))         result = TOKEN_TYPE_ID_G;
		else if (id.equalsIgnoreCase("gx"))        result = TOKEN_TYPE_ID_GX;
		else if (id.equalsIgnoreCase("gy"))        result = TOKEN_TYPE_ID_GY;
		else if (id.equalsIgnoreCase("h"))         result = TOKEN_TYPE_ID_H;
		else if (id.equalsIgnoreCase("hx"))        result = TOKEN_TYPE_ID_HX;
		else if (id.equalsIgnoreCase("hy"))        result = TOKEN_TYPE_ID_HY;
		else if (id.equalsIgnoreCase("k"))         result = TOKEN_TYPE_ID_K;
		else if (id.equalsIgnoreCase("kx"))        result = TOKEN_TYPE_ID_KX;
		else if (id.equalsIgnoreCase("ky"))        result = TOKEN_TYPE_ID_KY;
		else if (id.equalsIgnoreCase("m"))         result = TOKEN_TYPE_ID_M;
		else if (id.equalsIgnoreCase("mx"))        result = TOKEN_TYPE_ID_MX;
		else if (id.equalsIgnoreCase("my"))        result = TOKEN_TYPE_ID_MY;
		else if (id.equalsIgnoreCase("n"))         result = TOKEN_TYPE_ID_N;
		else if (id.equalsIgnoreCase("nx"))        result = TOKEN_TYPE_ID_NX;
		else if (id.equalsIgnoreCase("ny"))        result = TOKEN_TYPE_ID_NY;
		else if (id.equalsIgnoreCase("x"))         result = TOKEN_TYPE_ID_X;
		else if (id.equalsIgnoreCase("t"))         result = TOKEN_TYPE_ID_T;
		else if (id.equalsIgnoreCase("s"))         result = TOKEN_TYPE_ID_S;
		else if (id.equalsIgnoreCase("z"))         result = TOKEN_TYPE_ID_Z;
		else if (id.equalsIgnoreCase("abs"))       result = TOKEN_TYPE_ID_ABS;
		else if (id.equalsIgnoreCase("arg"))       result = TOKEN_TYPE_ID_ARG;
		else if (id.equalsIgnoreCase("conj"))      result = TOKEN_TYPE_ID_CONJ;
		else if (id.equalsIgnoreCase("sign"))      result = TOKEN_TYPE_ID_SIGN;
		else if (id.equalsIgnoreCase("not"))       result = TOKEN_TYPE_ID_NOT;
		else if (id.equalsIgnoreCase("int"))       result = TOKEN_TYPE_ID_INT;
		else if (id.equalsIgnoreCase("ceil"))      result = TOKEN_TYPE_ID_CEIL;
		else if (id.equalsIgnoreCase("floor"))     result = TOKEN_TYPE_ID_FLOOR;
		else if (id.equalsIgnoreCase("sqrt"))      result = TOKEN_TYPE_ID_SQRT;
		else if (id.equalsIgnoreCase("exp"))       result = TOKEN_TYPE_ID_EXP;
		else if (id.equalsIgnoreCase("log"))       result = TOKEN_TYPE_ID_LOG;
		else if (id.equalsIgnoreCase("ln"))        result = TOKEN_TYPE_ID_LOG;
		else if (id.equalsIgnoreCase("sin"))       result = TOKEN_TYPE_ID_SIN;
		else if (id.equalsIgnoreCase("cos"))       result = TOKEN_TYPE_ID_COS;
		else if (id.equalsIgnoreCase("tan"))       result = TOKEN_TYPE_ID_TAN;
		else if (id.equalsIgnoreCase("cot"))       result = TOKEN_TYPE_ID_COT;
		else if (id.equalsIgnoreCase("sec"))       result = TOKEN_TYPE_ID_SEC;
		else if (id.equalsIgnoreCase("csc"))       result = TOKEN_TYPE_ID_CSC;
		else if (id.equalsIgnoreCase("sinh"))      result = TOKEN_TYPE_ID_SINH;
		else if (id.equalsIgnoreCase("cosh"))      result = TOKEN_TYPE_ID_COSH;
		else if (id.equalsIgnoreCase("tanh"))      result = TOKEN_TYPE_ID_TANH;
		else if (id.equalsIgnoreCase("coth"))      result = TOKEN_TYPE_ID_COTH;
		else if (id.equalsIgnoreCase("sech"))      result = TOKEN_TYPE_ID_SECH;
		else if (id.equalsIgnoreCase("csch"))      result = TOKEN_TYPE_ID_CSCH;
		else if (id.equalsIgnoreCase("asin"))      result = TOKEN_TYPE_ID_ASIN;
		else if (id.equalsIgnoreCase("acos"))      result = TOKEN_TYPE_ID_ACOS;
		else if (id.equalsIgnoreCase("atan"))      result = TOKEN_TYPE_ID_ATAN;
		else if (id.equalsIgnoreCase("acot"))      result = TOKEN_TYPE_ID_ACOT;
		else if (id.equalsIgnoreCase("asec"))      result = TOKEN_TYPE_ID_ASEC;
		else if (id.equalsIgnoreCase("acsc"))      result = TOKEN_TYPE_ID_ACSC;
		else if (id.equalsIgnoreCase("asinh"))     result = TOKEN_TYPE_ID_ASINH;
		else if (id.equalsIgnoreCase("acosh"))     result = TOKEN_TYPE_ID_ACOSH;
		else if (id.equalsIgnoreCase("atanh"))     result = TOKEN_TYPE_ID_ATANH;
		else if (id.equalsIgnoreCase("acoth"))     result = TOKEN_TYPE_ID_ACOTH;
		else if (id.equalsIgnoreCase("asech"))     result = TOKEN_TYPE_ID_ASECH;
		else if (id.equalsIgnoreCase("acsch"))     result = TOKEN_TYPE_ID_ACSCH;
		else if (id.equalsIgnoreCase("arcsin"))    result = TOKEN_TYPE_ID_ASIN;
		else if (id.equalsIgnoreCase("arccos"))    result = TOKEN_TYPE_ID_ACOS;
		else if (id.equalsIgnoreCase("arctan"))    result = TOKEN_TYPE_ID_ATAN;
		else if (id.equalsIgnoreCase("arccot"))    result = TOKEN_TYPE_ID_ACOT;
		else if (id.equalsIgnoreCase("arcsec"))    result = TOKEN_TYPE_ID_ASEC;
		else if (id.equalsIgnoreCase("arccsc"))    result = TOKEN_TYPE_ID_ACSC;
		else if (id.equalsIgnoreCase("arcsinh"))   result = TOKEN_TYPE_ID_ASINH;
		else if (id.equalsIgnoreCase("arccosh"))   result = TOKEN_TYPE_ID_ACOSH;
		else if (id.equalsIgnoreCase("arctanh"))   result = TOKEN_TYPE_ID_ATANH;
		else if (id.equalsIgnoreCase("arccoth"))   result = TOKEN_TYPE_ID_ACOTH;
		else if (id.equalsIgnoreCase("arcsech"))   result = TOKEN_TYPE_ID_ASECH;
		else if (id.equalsIgnoreCase("arccsch"))   result = TOKEN_TYPE_ID_ACSCH;
		else if (id.equalsIgnoreCase("re"))        result = TOKEN_TYPE_ID_RE;
		else if (id.equalsIgnoreCase("im"))        result = TOKEN_TYPE_ID_IM;
		else if (id.equalsIgnoreCase("realonly"))  result = TOKEN_TYPE_ID_REALONLY;
		return (result);
		}
	public static String LexIdName(int id)
		{
		String result;
		switch (id)
			{
			case TOKEN_TYPE_ID_I        : result = "i";        break;
			case TOKEN_TYPE_ID_J        : result = "j";        break;
			case TOKEN_TYPE_ID_PI       : result = "pi";       break;
			case TOKEN_TYPE_ID_E        : result = "e";        break;
			case TOKEN_TYPE_ID_INFINITY : result = "infinity"; break;
			case TOKEN_TYPE_ID_NAN      : result = "nan";      break;
			case TOKEN_TYPE_ID_A        : result = "a";        break;
			case TOKEN_TYPE_ID_AX       : result = "ax";       break;
			case TOKEN_TYPE_ID_AY       : result = "ay";       break;
			case TOKEN_TYPE_ID_B        : result = "b";        break;
			case TOKEN_TYPE_ID_BX       : result = "bx";       break;
			case TOKEN_TYPE_ID_BY       : result = "by";       break;
			case TOKEN_TYPE_ID_C        : result = "c";        break;
			case TOKEN_TYPE_ID_CX       : result = "cx";       break;
			case TOKEN_TYPE_ID_CY       : result = "cy";       break;
			case TOKEN_TYPE_ID_D        : result = "d";        break;
			case TOKEN_TYPE_ID_DX       : result = "dx";       break;
			case TOKEN_TYPE_ID_DY       : result = "dy";       break;
			case TOKEN_TYPE_ID_F        : result = "f";        break;
			case TOKEN_TYPE_ID_FX       : result = "fx";       break;
			case TOKEN_TYPE_ID_FY       : result = "fy";       break;
			case TOKEN_TYPE_ID_G        : result = "g";        break;
			case TOKEN_TYPE_ID_GX       : result = "gx";       break;
			case TOKEN_TYPE_ID_GY       : result = "gy";       break;
			case TOKEN_TYPE_ID_H        : result = "h";        break;
			case TOKEN_TYPE_ID_HX       : result = "hx";       break;
			case TOKEN_TYPE_ID_HY       : result = "hy";       break;
			case TOKEN_TYPE_ID_K        : result = "k";        break;
			case TOKEN_TYPE_ID_KX       : result = "kx";       break;
			case TOKEN_TYPE_ID_KY       : result = "ky";       break;
			case TOKEN_TYPE_ID_M        : result = "m";        break;
			case TOKEN_TYPE_ID_MX       : result = "mx";       break;
			case TOKEN_TYPE_ID_MY       : result = "my";       break;
			case TOKEN_TYPE_ID_N        : result = "n";        break;
			case TOKEN_TYPE_ID_NX       : result = "nx";       break;
			case TOKEN_TYPE_ID_NY       : result = "ny";       break;
			case TOKEN_TYPE_ID_X        : result = "x";        break;
			case TOKEN_TYPE_ID_T        : result = "t";        break;
			case TOKEN_TYPE_ID_S        : result = "s";        break;
			case TOKEN_TYPE_ID_Z        : result = "z";        break;
			case TOKEN_TYPE_ID_ABS      : result = "abs";      break;
			case TOKEN_TYPE_ID_ARG      : result = "arg";      break;
			case TOKEN_TYPE_ID_CONJ     : result = "conj";     break;
			case TOKEN_TYPE_ID_SIGN     : result = "sign";     break;
			case TOKEN_TYPE_ID_NOT      : result = "not";      break;
			case TOKEN_TYPE_ID_INT      : result = "int";      break;
			case TOKEN_TYPE_ID_CEIL     : result = "ceil";     break;
			case TOKEN_TYPE_ID_FLOOR    : result = "floor";    break;
			case TOKEN_TYPE_ID_SQRT     : result = "sqrt";     break;
			case TOKEN_TYPE_ID_EXP      : result = "exp";      break;
			case TOKEN_TYPE_ID_LOG      : result = "log";      break;
			case TOKEN_TYPE_ID_SIN      : result = "sin";      break;
			case TOKEN_TYPE_ID_COS      : result = "cos";      break;
			case TOKEN_TYPE_ID_TAN      : result = "tan";      break;
			case TOKEN_TYPE_ID_COT      : result = "cot";      break;
			case TOKEN_TYPE_ID_SEC      : result = "sec";      break;
			case TOKEN_TYPE_ID_CSC      : result = "csc";      break;
			case TOKEN_TYPE_ID_SINH     : result = "sinh";     break;
			case TOKEN_TYPE_ID_COSH     : result = "cosh";     break;
			case TOKEN_TYPE_ID_TANH     : result = "tanh";     break;
			case TOKEN_TYPE_ID_COTH     : result = "coth";     break;
			case TOKEN_TYPE_ID_SECH     : result = "sech";     break;
			case TOKEN_TYPE_ID_CSCH     : result = "csch";     break;
			case TOKEN_TYPE_ID_ASIN     : result = "arcsin";   break;
			case TOKEN_TYPE_ID_ACOS     : result = "arccos";   break;
			case TOKEN_TYPE_ID_ATAN     : result = "arctan";   break;
			case TOKEN_TYPE_ID_ACOT     : result = "arccot";   break;
			case TOKEN_TYPE_ID_ASEC     : result = "arcsec";   break;
			case TOKEN_TYPE_ID_ACSC     : result = "arccsc";   break;
			case TOKEN_TYPE_ID_ASINH    : result = "arcsinh";  break;
			case TOKEN_TYPE_ID_ACOSH    : result = "arccosh";  break;
			case TOKEN_TYPE_ID_ATANH    : result = "arctanh";  break;
			case TOKEN_TYPE_ID_ACOTH    : result = "arccoth";  break;
			case TOKEN_TYPE_ID_ASECH    : result = "arcsech";  break;
			case TOKEN_TYPE_ID_ACSCH    : result = "arccsch";  break;
			case TOKEN_TYPE_ID_RE       : result = "re";       break;
			case TOKEN_TYPE_ID_IM       : result = "im";       break;
			case TOKEN_TYPE_ID_REALONLY : result = "realonly"; break;
			default                     : result = "INTERNAL-ERROR";    break;
			}

		return (result);
		}

	//----------------------------------------------------------------
	// PGC_Eval: Parse
	//----------------------------------------------------------------

	private Node ParseExpression()
		{
		Node result;
		if (token_type == TOKEN_TYPE_EOS)
			{
			// An empty top-level expression is syntactically legal.
			result = null;
			}
		else
			{
			result = ParseExpr(0, false);
			if (token_type != TOKEN_TYPE_EOS)
				ParseError("Unexpected text, at " + token_pos);
			}
		return (result);
		}
	private Node ParseExpr(int level, boolean inside_bars)
		{
		Node result = null;
		if (level == 8)
			{
			//--------------------------------------
			// Unary prefix ops
			//--------------------------------------
			if ((token_type == TOKEN_TYPE_SYMBOL_PLUS) || (token_type == TOKEN_TYPE_SYMBOL_MINUS))
				{
				int sym = token_type;
				LexGetToken();
				result = ParseExpr(level, inside_bars);
				if (sym == TOKEN_TYPE_SYMBOL_MINUS)
					result = Node.NewNeg(result);
				}
			else if ((token_type & TOKEN_TYPE_ID_CLASS_FUNC) == TOKEN_TYPE_ID_CLASS_FUNC)
				{
				int id = token_type;
				LexGetToken();
				result = ParseExpr(level, inside_bars);
				result = Node.NewFunc(id, result);
				}
			else
				result = ParseExpr(level+1, inside_bars);
			}
		else if (level == 9)
			{
			//--------------------------------------
			// Atoms and nesting ops
			//--------------------------------------
			if (token_type == TOKEN_TYPE_SYMBOL_LPAREN)
				{
				LexGetToken();
				result = ParseExpr(0, false);
				if (token_type == TOKEN_TYPE_SYMBOL_RPAREN)
					LexGetToken();
				else
					ParseError("Expected a matching ')', at " + token_pos);
				}
			else if ((token_type == TOKEN_TYPE_SYMBOL_BAR) && !inside_bars)
				{
				LexGetToken();
				result = ParseExpr(0, true);
				if (token_type == TOKEN_TYPE_SYMBOL_BAR)
					LexGetToken();
				else
					ParseError("Expected a matching '|', at " + token_pos);
				result = Node.NewFunc(TOKEN_TYPE_ID_ABS, result);
				}
			else if (token_type == TOKEN_TYPE_NUMBER)
				{
				Double double_val = PGC_Utility.DoubleValue(token);
				if (double_val == null)
					ParseError("Bad number format '" + token + "', at " + token_pos);
				else
					result = Node.NewNumber(double_val.doubleValue());
				LexGetToken();
				}
			else if ((token_type & TOKEN_TYPE_ID_CLASS_VAR) == TOKEN_TYPE_ID_CLASS_VAR)
				{
				int id = token_type;
				result = Node.NewVar(id);
				LexGetToken();
				}
			else if ((token_type & TOKEN_TYPE_ID) == TOKEN_TYPE_ID)
				{
				ParseError("Unknown identifier '" + token + "', at " + token_pos);
				LexGetToken();
				}
			else
				{
				ParseError("Expected an expression, at " + token_pos);
				}
			}
		else
			{
			result = ParseExpr(level+1, inside_bars);
			}
		if (result == null)
			result = Node.NewUndefined();
		result = ParseExprTail(level, result, inside_bars);
		return (result);
		}
	private Node ParseExprTail(int level, Node expr0, boolean inside_bars)
		{
		Node result = null;

		if (
		    ((level == 0) && (token_type == TOKEN_TYPE_SYMBOL_COMMA     )) ||
		    ((level == 1) && (token_type == TOKEN_TYPE_SYMBOL_ID_OR     )) ||
		    ((level == 2) && (token_type == TOKEN_TYPE_SYMBOL_ID_AND    )) ||
		    ((level == 3) && (token_type == TOKEN_TYPE_SYMBOL_EQ        )) ||
		    ((level == 3) && (token_type == TOKEN_TYPE_SYMBOL_NE        )) ||
		    ((level == 3) && (token_type == TOKEN_TYPE_SYMBOL_LT        )) ||
		    ((level == 3) && (token_type == TOKEN_TYPE_SYMBOL_LE        )) ||
		    ((level == 3) && (token_type == TOKEN_TYPE_SYMBOL_GT        )) ||
		    ((level == 3) && (token_type == TOKEN_TYPE_SYMBOL_GE        )) ||
		    ((level == 4) && (token_type == TOKEN_TYPE_SYMBOL_ID_DOT    )) ||
		    ((level == 4) && (token_type == TOKEN_TYPE_SYMBOL_ID_CROSS  )) ||
		    ((level == 5) && (token_type == TOKEN_TYPE_SYMBOL_PLUS      )) ||
		    ((level == 5) && (token_type == TOKEN_TYPE_SYMBOL_MINUS     )) ||
		    ((level == 6) && (token_type == TOKEN_TYPE_SYMBOL_MUL       )) ||
		    ((level == 6) && (token_type == TOKEN_TYPE_SYMBOL_DIV       )) ||
		    ((level == 6) && (token_type == TOKEN_TYPE_SYMBOL_ID_REM    )) ||
		    ((level == 6) && (token_type == TOKEN_TYPE_SYMBOL_ID_MOD    )) ||
		    ((level == 7) && (token_type == TOKEN_TYPE_SYMBOL_POW       ))
		   )
			{
			//------------------------------------------
			// Binary op, left-associative
			//------------------------------------------
			int sym = token_type;
			LexGetToken();
			Node expr1 = ParseExpr(level+1, inside_bars);
			switch (sym)
				{
				case TOKEN_TYPE_SYMBOL_COMMA     : result = Node.NewComma(expr0, expr1); break;
				case TOKEN_TYPE_SYMBOL_ID_OR     : result = Node.NewOr(expr0, expr1);    break;
				case TOKEN_TYPE_SYMBOL_ID_AND    : result = Node.NewAnd(expr0, expr1);   break;
				case TOKEN_TYPE_SYMBOL_EQ        : result = Node.NewEQ(expr0, expr1);    break;
				case TOKEN_TYPE_SYMBOL_NE        : result = Node.NewNE(expr0, expr1);    break;
				case TOKEN_TYPE_SYMBOL_LT        : result = Node.NewLT(expr0, expr1);    break;
				case TOKEN_TYPE_SYMBOL_LE        : result = Node.NewLE(expr0, expr1);    break;
				case TOKEN_TYPE_SYMBOL_GT        : result = Node.NewGT(expr0, expr1);    break;
				case TOKEN_TYPE_SYMBOL_GE        : result = Node.NewGE(expr0, expr1);    break;
				case TOKEN_TYPE_SYMBOL_ID_DOT    : result = Node.NewDot(expr0, expr1);   break;
				case TOKEN_TYPE_SYMBOL_ID_CROSS  : result = Node.NewCross(expr0, expr1); break;
				case TOKEN_TYPE_SYMBOL_PLUS      : result = Node.NewAdd(expr0, expr1);   break;
				case TOKEN_TYPE_SYMBOL_MINUS     : result = Node.NewSub(expr0, expr1);   break;
				case TOKEN_TYPE_SYMBOL_MUL       : result = Node.NewMul(expr0, expr1);   break;
				case TOKEN_TYPE_SYMBOL_DIV       : result = Node.NewDiv(expr0, expr1);   break;
				case TOKEN_TYPE_SYMBOL_ID_REM    : result = Node.NewRem(expr0, expr1);   break;
				case TOKEN_TYPE_SYMBOL_ID_MOD    : result = Node.NewMod(expr0, expr1);   break;
				case TOKEN_TYPE_SYMBOL_POW       : result = Node.NewPow(expr0, expr1);   break;
				}
			result = ParseExprTail(level, result, inside_bars);
			}
		//// else if (false)
		////	{
		////	//------------------------------------------
		////	// Binary op, right-associative
		////	//------------------------------------------
		////	int sym = token_type;
		////	LexGetToken();
		////	Node expr1 = ParseExpr(level, inside_bars);
		////	result = Node.NewXXX(expr0, expr1);
		////	}
		else if ((level == 6) &&
		         (((token_type == TOKEN_TYPE_SYMBOL_LPAREN)) ||
		          ((token_type == TOKEN_TYPE_SYMBOL_BAR) && !inside_bars) ||
		          ((token_type == TOKEN_TYPE_NUMBER)) ||
			  ((token_type & TOKEN_TYPE_ID_CLASS_VAR) == TOKEN_TYPE_ID_CLASS_VAR) ||
			  ((token_type & TOKEN_TYPE_ID_CLASS_FUNC) == TOKEN_TYPE_ID_CLASS_FUNC)
			 ))
			{
			//------------------------------------------
			// The empty binary op, left-associative
			//
			// Example: "2x" means "2*x".
			//------------------------------------------
			Node expr1 = ParseExpr(level+1, inside_bars);
			result = Node.NewMul(expr0, expr1);
			result = ParseExprTail(level, result, inside_bars);
			}
		else if ((level == 9) && (token_type == TOKEN_TYPE_SYMBOL_PERIOD))
			{
			//------------------------------------------
			// Unary postfix op, left-associative
			//------------------------------------------
			LexGetToken();
			if (token.equalsIgnoreCase("x"))
				{
				result = Node.NewFunc(TOKEN_TYPE_ID_RE, expr0);
				LexGetToken();
				}
			else if (token.equalsIgnoreCase("y"))
				{
				result = Node.NewFunc(TOKEN_TYPE_ID_IM, expr0);
				LexGetToken();
				}
			else
				ParseError("Expected '.x' or '.y', at " + token_pos);
			result = ParseExprTail(level, result, inside_bars);
			}
		else
			{
			result = expr0;
			}
		return (result);
		}
	private void ParseError(String new_message)
		{
		if (message.length() == 0)
			message = new_message;
		}

	//----------------------------------------------------------------
	// PGC_Eval: Eval
	//----------------------------------------------------------------

	private Complex EvalNode(Node n)
		{
		Complex result = null;
		if (n != null)
			{
			switch (n.kind)
				{
				case Node.KIND_NUMBER   : { result = n.value; break; }
				case Node.KIND_VAR      : { result = EvalVariable(n.id); break; }
				case Node.KIND_FUNC     : { Complex val0 = EvalNode(n.expr0); if (val0 != null) result = EvalFunction(n.id, val0); break; }
				case Node.KIND_NEG      : { Complex val0 = EvalNode(n.expr0); if (val0 != null) result = Complex.Neg(val0); break; }
				case Node.KIND_REC      : { Complex val0 = EvalNode(n.expr0); if (val0 != null) result = Complex.Reciprocal(val0); break; }
				case Node.KIND_COMMA    : { Complex val0 = EvalNode(n.expr0); Complex val1 = EvalNode(n.expr1); if ((val0 != null) && (val1 != null)) result = Complex.Number(val0.re, val1.re);                                            break; }
				case Node.KIND_OR       : { Complex val0 = EvalNode(n.expr0); Complex val1 = EvalNode(n.expr1); if ((val0 != null) && (val1 != null)) result = Complex.Number(Complex.Or(val0, val1));                                      break; }
				case Node.KIND_AND      : { Complex val0 = EvalNode(n.expr0); Complex val1 = EvalNode(n.expr1); if ((val0 != null) && (val1 != null)) result = Complex.Number(Complex.And(val0, val1));                                     break; }
				case Node.KIND_EQ       : { Complex val0 = EvalNode(n.expr0); Complex val1 = EvalNode(n.expr1); if ((val0 != null) && (val1 != null)) result = Complex.Number(owner.LooksEqual(val0, val1));                                break; }
				case Node.KIND_NE       : { Complex val0 = EvalNode(n.expr0); Complex val1 = EvalNode(n.expr1); if ((val0 != null) && (val1 != null)) result = Complex.Number(!owner.LooksEqual(val0, val1));                               break; }
				case Node.KIND_LT       : { Complex val0 = EvalNode(n.expr0); Complex val1 = EvalNode(n.expr1); if ((val0 != null) && (val1 != null)) result = Complex.Number(Complex.Less(val0, val1));                                    break; }
				case Node.KIND_LE       : { Complex val0 = EvalNode(n.expr0); Complex val1 = EvalNode(n.expr1); if ((val0 != null) && (val1 != null)) result = Complex.Number(Complex.Less(val0, val1) || owner.LooksEqual(val0, val1));    break; }
				case Node.KIND_GT       : { Complex val0 = EvalNode(n.expr0); Complex val1 = EvalNode(n.expr1); if ((val0 != null) && (val1 != null)) result = Complex.Number(Complex.Greater(val0, val1));                                 break; }
				case Node.KIND_GE       : { Complex val0 = EvalNode(n.expr0); Complex val1 = EvalNode(n.expr1); if ((val0 != null) && (val1 != null)) result = Complex.Number(Complex.Greater(val0, val1) || owner.LooksEqual(val0, val1)); break; }
				case Node.KIND_DOT      : { Complex val0 = EvalNode(n.expr0); Complex val1 = EvalNode(n.expr1); if ((val0 != null) && (val1 != null)) result = Complex.Number(Complex.Dot(val0, val1));                                     break; }
				case Node.KIND_CROSS    : { Complex val0 = EvalNode(n.expr0); Complex val1 = EvalNode(n.expr1); if ((val0 != null) && (val1 != null)) result = Complex.Number(Complex.Cross(val0, val1));                                   break; }
				case Node.KIND_ADD      : { Complex val0 = EvalNode(n.expr0); Complex val1 = EvalNode(n.expr1); if ((val0 != null) && (val1 != null)) result = Complex.Add(val0, val1); break; }
				case Node.KIND_SUB      : { Complex val0 = EvalNode(n.expr0); Complex val1 = EvalNode(n.expr1); if ((val0 != null) && (val1 != null)) result = Complex.Sub(val0, val1); break; }
				case Node.KIND_MUL      : { Complex val0 = EvalNode(n.expr0); Complex val1 = EvalNode(n.expr1); if ((val0 != null) && (val1 != null)) result = Complex.Mul(val0, val1); break; }
				case Node.KIND_DIV      : { Complex val0 = EvalNode(n.expr0); Complex val1 = EvalNode(n.expr1); if ((val0 != null) && (val1 != null)) result = Complex.Div(val0, val1); break; }
				case Node.KIND_REM      : { Complex val0 = EvalNode(n.expr0); Complex val1 = EvalNode(n.expr1); if ((val0 != null) && (val1 != null)) result = Complex.Rem(val0, val1); break; }
				case Node.KIND_MOD      : { Complex val0 = EvalNode(n.expr0); Complex val1 = EvalNode(n.expr1); if ((val0 != null) && (val1 != null)) result = Complex.Mod(val0, val1); break; }
				case Node.KIND_POW      : { Complex val0 = EvalNode(n.expr0); Complex val1 = EvalNode(n.expr1); if ((val0 != null) && (val1 != null)) result = Complex.Pow(val0, val1); break; }
				default                 : { break; }
				}
			}
		return (result);
		}

	private Complex EvalVariable(int id)
		{
		Complex result = null;
		switch (id)
			{
			case TOKEN_TYPE_ID_I        : { result = Complex.Number(0.0, 1.0);                  break; }
			case TOKEN_TYPE_ID_J        : { result = Complex.Number(0.0, 1.0);                  break; }
			case TOKEN_TYPE_ID_PI       : { result = Complex.Number(3.14159265358979324);       break; }
			case TOKEN_TYPE_ID_E        : { result = Complex.Number(2.71828182845904524);       break; }
			case TOKEN_TYPE_ID_INFINITY : { result = Complex.Number(Double.POSITIVE_INFINITY);  break; }
			case TOKEN_TYPE_ID_NAN      : { result = Complex.Number(Double.NaN);                break; }

			case TOKEN_TYPE_ID_A        : { result = owner.GetVariableCoordinates(PGC.VAR_A); if (result != null) result = Complex.Number(result);    uses_input[PGC.VAR_A] = true; break; }
			case TOKEN_TYPE_ID_AX       : { result = owner.GetVariableCoordinates(PGC.VAR_A); if (result != null) result = Complex.Number(result.re); uses_input[PGC.VAR_A] = true; break; }
			case TOKEN_TYPE_ID_AY       : { result = owner.GetVariableCoordinates(PGC.VAR_A); if (result != null) result = Complex.Number(result.im); uses_input[PGC.VAR_A] = true; break; }
			case TOKEN_TYPE_ID_B        : { result = owner.GetVariableCoordinates(PGC.VAR_B); if (result != null) result = Complex.Number(result);    uses_input[PGC.VAR_B] = true; break; }
			case TOKEN_TYPE_ID_BX       : { result = owner.GetVariableCoordinates(PGC.VAR_B); if (result != null) result = Complex.Number(result.re); uses_input[PGC.VAR_B] = true; break; }
			case TOKEN_TYPE_ID_BY       : { result = owner.GetVariableCoordinates(PGC.VAR_B); if (result != null) result = Complex.Number(result.im); uses_input[PGC.VAR_B] = true; break; }
			case TOKEN_TYPE_ID_C        : { result = owner.GetVariableCoordinates(PGC.VAR_C); if (result != null) result = Complex.Number(result);    uses_input[PGC.VAR_C] = true; break; }
			case TOKEN_TYPE_ID_CX       : { result = owner.GetVariableCoordinates(PGC.VAR_C); if (result != null) result = Complex.Number(result.re); uses_input[PGC.VAR_C] = true; break; }
			case TOKEN_TYPE_ID_CY       : { result = owner.GetVariableCoordinates(PGC.VAR_C); if (result != null) result = Complex.Number(result.im); uses_input[PGC.VAR_C] = true; break; }
			case TOKEN_TYPE_ID_D        : { result = owner.GetVariableCoordinates(PGC.VAR_D); if (result != null) result = Complex.Number(result);    uses_input[PGC.VAR_D] = true; break; }
			case TOKEN_TYPE_ID_DX       : { result = owner.GetVariableCoordinates(PGC.VAR_D); if (result != null) result = Complex.Number(result.re); uses_input[PGC.VAR_D] = true; break; }
			case TOKEN_TYPE_ID_DY       : { result = owner.GetVariableCoordinates(PGC.VAR_D); if (result != null) result = Complex.Number(result.im); uses_input[PGC.VAR_D] = true; break; }

			case TOKEN_TYPE_ID_F        : { result = EvalUserDefinedFunction(PGC.VAR_F, 0);     break; }
			case TOKEN_TYPE_ID_FX       : { result = EvalUserDefinedFunction(PGC.VAR_F, 1);     break; }
			case TOKEN_TYPE_ID_FY       : { result = EvalUserDefinedFunction(PGC.VAR_F, 2);     break; }
			case TOKEN_TYPE_ID_G        : { result = EvalUserDefinedFunction(PGC.VAR_G, 0);     break; }
			case TOKEN_TYPE_ID_GX       : { result = EvalUserDefinedFunction(PGC.VAR_G, 1);     break; }
			case TOKEN_TYPE_ID_GY       : { result = EvalUserDefinedFunction(PGC.VAR_G, 2);     break; }
			case TOKEN_TYPE_ID_H        : { result = EvalUserDefinedFunction(PGC.VAR_H, 0);     break; }
			case TOKEN_TYPE_ID_HX       : { result = EvalUserDefinedFunction(PGC.VAR_H, 1);     break; }
			case TOKEN_TYPE_ID_HY       : { result = EvalUserDefinedFunction(PGC.VAR_H, 2);     break; }
			case TOKEN_TYPE_ID_K        : { result = EvalUserDefinedFunction(PGC.VAR_K, 0);     break; }
			case TOKEN_TYPE_ID_KX       : { result = EvalUserDefinedFunction(PGC.VAR_K, 1);     break; }
			case TOKEN_TYPE_ID_KY       : { result = EvalUserDefinedFunction(PGC.VAR_K, 2);     break; }
			case TOKEN_TYPE_ID_M        : { result = EvalUserDefinedFunction(PGC.VAR_M, 0);     break; }
			case TOKEN_TYPE_ID_MX       : { result = EvalUserDefinedFunction(PGC.VAR_M, 1);     break; }
			case TOKEN_TYPE_ID_MY       : { result = EvalUserDefinedFunction(PGC.VAR_M, 2);     break; }
			case TOKEN_TYPE_ID_N        : { result = EvalUserDefinedFunction(PGC.VAR_N, 0);     break; }
			case TOKEN_TYPE_ID_NX       : { result = EvalUserDefinedFunction(PGC.VAR_N, 1);     break; }
			case TOKEN_TYPE_ID_NY       : { result = EvalUserDefinedFunction(PGC.VAR_N, 2);     break; }

			case TOKEN_TYPE_ID_X        : { result = owner.GetAutoplotVariable(); uses_autoplot = true; uses_autoplot_x = true; break; }
			case TOKEN_TYPE_ID_T        : { result = owner.GetAutoplotVariable(); uses_autoplot = true; uses_autoplot_t = true; break; }
			case TOKEN_TYPE_ID_S        : { result = owner.GetAutoincrementVariable(); break; }
			case TOKEN_TYPE_ID_Z        : { result = Complex.Number((double)owner.GetZoom(), 0.0); break; }
			}
		return (result);
		}
	private Complex EvalFunction(int id, Complex val)
		{
		Complex result = null;
		if (val != null)
			{
			switch (id)
				{
				case TOKEN_TYPE_ID_ABS      : { result = Complex.Number(Complex.Abs(val));               break; }
				case TOKEN_TYPE_ID_ARG      : { result = Complex.Number(Complex.Arg(val));               break; }
				case TOKEN_TYPE_ID_CONJ     : { result = Complex.Conj(val);                              break; }
				case TOKEN_TYPE_ID_SIGN     : { result = Complex.Sign(val);                              break; }
				case TOKEN_TYPE_ID_NOT      : { result = Complex.Number(Complex.Not(val));               break; }
				case TOKEN_TYPE_ID_INT      : { result = Complex.Int(val);                               break; }
				case TOKEN_TYPE_ID_CEIL     : { result = Complex.Ceil(val);                              break; }
				case TOKEN_TYPE_ID_FLOOR    : { result = Complex.Floor(val);                             break; }
				case TOKEN_TYPE_ID_SQRT     : { result = Complex.Sqrt(val);                              break; }
				case TOKEN_TYPE_ID_EXP      : { result = Complex.Exp(val);                               break; }
				case TOKEN_TYPE_ID_LOG      : { result = Complex.Log(val);                               break; }
				case TOKEN_TYPE_ID_SIN      : { result = Complex.Sin(val);                               break; }
				case TOKEN_TYPE_ID_COS      : { result = Complex.Cos(val);                               break; }
				case TOKEN_TYPE_ID_TAN      : { result = Complex.Tan(val);                               break; }
				case TOKEN_TYPE_ID_COT      : { result = Complex.Cot(val);                               break; }
				case TOKEN_TYPE_ID_SEC      : { result = Complex.Sec(val);                               break; }
				case TOKEN_TYPE_ID_CSC      : { result = Complex.Csc(val);                               break; }
				case TOKEN_TYPE_ID_SINH     : { result = Complex.Sinh(val);                              break; }
				case TOKEN_TYPE_ID_COSH     : { result = Complex.Cosh(val);                              break; }
				case TOKEN_TYPE_ID_TANH     : { result = Complex.Tanh(val);                              break; }
				case TOKEN_TYPE_ID_COTH     : { result = Complex.Coth(val);                              break; }
				case TOKEN_TYPE_ID_SECH     : { result = Complex.Sech(val);                              break; }
				case TOKEN_TYPE_ID_CSCH     : { result = Complex.Csch(val);                              break; }
				case TOKEN_TYPE_ID_ASIN     : { result = Complex.Asin(val);                              break; }
				case TOKEN_TYPE_ID_ACOS     : { result = Complex.Acos(val);                              break; }
				case TOKEN_TYPE_ID_ATAN     : { result = Complex.Atan(val);                              break; }
				case TOKEN_TYPE_ID_ACOT     : { result = Complex.Acot(val);                              break; }
				case TOKEN_TYPE_ID_ASEC     : { result = Complex.Asec(val);                              break; }
				case TOKEN_TYPE_ID_ACSC     : { result = Complex.Acsc(val);                              break; }
				case TOKEN_TYPE_ID_ASINH    : { result = Complex.Asinh(val);                             break; }
				case TOKEN_TYPE_ID_ACOSH    : { result = Complex.Acosh(val);                             break; }
				case TOKEN_TYPE_ID_ATANH    : { result = Complex.Atanh(val);                             break; }
				case TOKEN_TYPE_ID_ACOTH    : { result = Complex.Acoth(val);                             break; }
				case TOKEN_TYPE_ID_ASECH    : { result = Complex.Asech(val);                             break; }
				case TOKEN_TYPE_ID_ACSCH    : { result = Complex.Acsch(val);                             break; }
				case TOKEN_TYPE_ID_RE       : { result = Complex.Number(val.re);                         break; }
				case TOKEN_TYPE_ID_IM       : { result = Complex.Number(val.im);                         break; }
				case TOKEN_TYPE_ID_REALONLY : { if (owner.LooksReal(val)) result = Complex.Number(val);  break; }
				}
			}
		return (result);
		}
	private Complex EvalUserDefinedFunction(int function, int component)
		{
		Complex result = null;

		if (allowed[function])
			{
			PGC_Eval eval = new PGC_Eval(owner);

			//------------------------------------------
			// Pass down the "allowed" attributes.
			// Also, remove this function from eval's
			// list of "allowed" functions.
			// This prevents recursive definitions
			// from causing problems.
			//------------------------------------------
			for (int i = 0; i < allowed.length; i++)
				eval.allowed[i] = allowed[i];
			eval.allowed[function] = false;

			//------------------------------------------
			// Evaluate.
			//------------------------------------------
			owner.GetVariableEvaluation(function, eval);
			result = eval.value;

			if (result != null)
				{
				if (component == 1) result = Complex.Number(result.re);
				if (component == 2) result = Complex.Number(result.im);
				}

			//------------------------------------------
			// Inherit the "recursive" and "uses..."
			// attributes.
			//------------------------------------------

			if (eval.recursive != null)  recursive       = eval.recursive;
			if (eval.uses_autoplot)      uses_autoplot   = true;
			if (eval.uses_autoplot_x)    uses_autoplot_x = true;
			if (eval.uses_autoplot_t)    uses_autoplot_t = true;

			for (int i = 0; i < uses_input.length; i++)
				if (eval.uses_input[i]) uses_input[i] = true;
			}
		else
			{
			//------------------------------------------
			// If the function is not on the allowed
			// list, then simply return null.  This
			// case indicates that the user-defined
			// function contains a recursive definition.
			//------------------------------------------
			recursive = owner.variable[function].name;
			}
		return (result);
		}
	}



//========================================================================
//========================================================================
// Complex
//========================================================================
//========================================================================

class Complex
	{
	public double re;
	public double im;

	public Complex()
		{
		re = 0.0;
		im = 0.0;
		}
	public Complex(Complex c)
		{
		re = c.re;
		im = c.im;
		}
	public Complex(double re_arg, double im_arg)
		{
		if (re_arg == -0.0) re_arg = 0.0;
		if (im_arg == -0.0) im_arg = 0.0;
		re = re_arg;
		im = im_arg;
		}
	public static Complex Number(double re, double im)
		{
		return (new Complex(re, im));
		}
	public static Complex Number(double re)
		{
		return (new Complex(re, 0.0));
		}
	public static Complex Number(boolean b)
		{
		if (b)
			return (new Complex(1.0, 0.0));
		else
			return (new Complex(0.0, 0.0));
		}
	public static Complex Number()
		{
		return (new Complex(0.0, 0.0));
		}
	public static Complex Number(Complex c)
		{
		return (new Complex(c.re, c.im));
		}
	public static boolean IsNaN(Complex c)
		{
		return (Double.isNaN(c.re) || Double.isNaN(c.im));
		}
	public static boolean IsInfinite(Complex c)
		{
		return (Double.isInfinite(c.re) || Double.isInfinite(c.im));
		}
	public static boolean IsDefined(Complex c)
		{
		return (!IsNaN(c) && !IsInfinite(c));
		}
	public static Complex Add(Complex c, Complex d)
		{
		double re = c.re + d.re;
		double im = c.im + d.im;
		return (new Complex(re, im));
		}
	public static Complex Sub(Complex c, Complex d)
		{
		double re = c.re - d.re;
		double im = c.im - d.im;
		return (new Complex(re, im));
		}
	public static Complex Neg(Complex c)
		{
		double re = -c.re;
		double im = -c.im;
		return (new Complex(re, im));
		}
	public static Complex Mul(Complex c, Complex d)
		{
		double re;
		double im;
		if ((c.im == 0.0) && (d.im == 0.0))
			{
			re = c.re * d.re;
			im = 0.0;
			}
		else
			{
			re = (c.re * d.re) - (c.im * d.im);
			im = (c.re * d.im) + (c.im * d.re);
			}
		return (new Complex(re, im));
		}
	public static Complex Mul(Complex c, double d)
		{
		double re = c.re * d;
		double im = c.im * d;
		return (new Complex(re, im));
		}
	public static Complex Muli(Complex c, double d)
		{
		double re = -c.im * d;
		double im = c.re * d;
		return (new Complex(re, im));
		}
	public static Complex Div(Complex c, Complex d)
		{
		double re;
		double im;
		if ((c.im == 0.0) && (d.im == 0.0))
			{
			re = c.re / d.re; // allow infinity results
			im = 0.0;
			}
		else
			{
			double d2 = Dot(d, d);
			re = ((c.re * d.re) + (c.im * d.im)) / d2;
			im = ((c.im * d.re) - (c.re * d.im)) / d2;
			}
		return (new Complex(re, im));
		}
	public static Complex Reciprocal(Complex c)
		{
		double re;
		double im;
		if (c.im == 0.0)
			{
			re = 1.0 / c.re; // allow infinity results
			im = 0.0;
			}
		else
			{
			double c2 = Dot(c, c);
			re = c.re / c2;
			im = -c.im / c2;
			}
		return (new Complex(re, im));
		}
	public static double Abs(Complex c) // a.k.a. "modulus"
		{
		double result = Math.sqrt(Dot(c, c));
		return (result);
		}
	public static double Arg(Complex c) // a.k.a. "argument"
		{
		double result = Math.atan2(c.im, c.re);
		return (result);
		}
	public static Complex Conj(Complex c)
		{
		double re = c.re;
		double im = -c.im;
		return (new Complex(re, im));
		}
	public static Complex Sign(Complex c)
		{
		double re;
		double im;
		double abs = Abs(c);
		if (abs == 0.0)
			{
			re = 0.0;
			im = 0.0;
			}
		else
			{
			re = c.re / Abs(c);
			im = c.im / Abs(c);
			}
		return (new Complex(re, im));
		}
	public static Complex Int(Complex c)
		{
		double re;
		double im;
		if (c.re >= 0.0) re = Math.floor(c.re); else re = Math.ceil(c.re);
		if (c.im >= 0.0) im = Math.floor(c.im); else im = Math.ceil(c.im);
		return (new Complex(re, im));
		}
	public static Complex Ceil(Complex c)
		{
		double re = Math.ceil(c.re);
		double im = Math.ceil(c.im);
		return (new Complex(re, im));
		}
	public static Complex Floor(Complex c)
		{
		double re = Math.floor(c.re);
		double im = Math.floor(c.im);
		return (new Complex(re, im));
		}
	public static boolean Less(Complex c, Complex d)
		{
		boolean result = ((c.re < d.re) || ((c.re == d.re) && (c.im < d.im)));
		return (result);
		}
	public static boolean Greater(Complex c, Complex d)
		{
		boolean result = ((c.re > d.re) || ((c.re == d.re) && (c.im > d.im)));
		return (result);
		}
	public static boolean Not(Complex c)
		{
		boolean result = ((c.re == 0.0) && (c.im == 0.0));
		return (result);
		}
	public static boolean And(Complex c, Complex d)
		{
		boolean result = (((c.re != 0.0) || (c.im != 0.0)) && ((d.re != 0.0) || (d.im != 0.0)));
		return (result);
		}
	public static boolean Or(Complex c, Complex d)
		{
		boolean result = (((c.re != 0.0) || (c.im != 0.0)) || ((d.re != 0.0) || (d.im != 0.0)));
		return (result);
		}
	public static Complex Rem(Complex c, Complex d)
		{
		double re = Math.IEEEremainder(c.re, d.re);
		double im = 0.0;
		return (new Complex(re, im));
		}
	public static Complex Mod(Complex c, Complex d)
		{
		double div = c.re / d.re;
		double fraction = div - Math.floor(div);
		double re = d.re * fraction;
		double im = 0.0;
		return (new Complex(re, im));
		}
	public static Complex Sqrt(Complex c)
		{
		double re = Math.sqrt((Abs(c) + c.re) * 0.5);
		double im;
		if (re == 0.0)
			im = Math.sqrt(-c.re);
		else
			im = 0.5 * c.im / re;
		return (new Complex(re, im));
		}
	public static Complex Pow(Complex c, Complex d)
		{
		// z = c^d
		double z_re;
		double z_im;
		if ((c.re == 0.0) && (c.im == 0.0) && (d.im == 0.0) && (d.re > 0.0))
			{
			// Special case: 0^d is defined if d is a positive real
			z_re = 0.0;
			z_im = 0.0;
			}
		else if ((c.im == 0.0) && (c.re > 0.0))
			{
			double c_ln   = Math.log(c.re);
			double re     = d.re * c_ln;
			double im     = d.im * c_ln;
			double exp_re = Math.exp(re);
			z_re = exp_re * Math.cos(im);
			z_im = exp_re * Math.sin(im);
			}
		else if (d.im == 0.0)
			{
			double c_r     = Abs(c);
			double c_theta = Arg(c);
			double exp_r   = Math.exp(d.re * Math.log(c_r));
			z_re = exp_r * Math.cos(d.re * c_theta);
			z_im = exp_r * Math.sin(d.re * c_theta);

			////
			//// // Alternative implementation.
			//// // FUTURE: check to see if this
			//// // version is faster.
			////
			//// Complex c_ln = Log(c);
			//// double re = c_ln.re * d.re;
			//// double im = c_ln.im * d.re;
			//// double exp_re = Math.exp(re);
			//// z_re = exp_re * Math.cos(im);
			//// z_im = exp_re * Math.sin(im);
			}
		else
			{
			double c_r     = Abs(c);
			double c_theta = Arg(c);
			double c_ln_r  = Math.log(c_r);
			double z_r     = Math.exp((d.re * c_ln_r) - (d.im * c_theta));
			double z_theta = (d.im * c_ln_r) + (d.re * c_theta);
			z_re = z_r * Math.cos(z_theta);
			z_im = z_r * Math.sin(z_theta);
			}
		return (new Complex(z_re, z_im));
		}
	public static Complex Exp(Complex c)
		{
		double exp_re = Math.exp(c.re);
		double re = exp_re * Math.cos(c.im);
		double im = exp_re * Math.sin(c.im);
		return (new Complex(re, im));
		}
	public static Complex Log(Complex c)
		{
		double abs_sq = Dot(c, c);
		double re = Math.log(abs_sq) * 0.5;
		double im = Arg(c);
		return (new Complex(re, im));
		}
	public static Complex Sin(Complex c)
		{
		double ex = Math.exp(c.im);
		double ex_pos = ex / 2.0;
		double ex_neg = 0.5 / ex;
		double re = Math.sin(c.re) * (ex_pos + ex_neg);
		double im = Math.cos(c.re) * (ex_pos - ex_neg);
		return (new Complex(re, im));
		}
	public static Complex Cos(Complex c)
		{
		double ex = Math.exp(c.im);
		double ex_pos = ex / 2.0;
		double ex_neg = 0.5 / ex;
		double re = Math.cos(c.re) *   (ex_pos + ex_neg);
		double im = Math.sin(c.re) * - (ex_pos - ex_neg);
		return (new Complex(re, im));
		}
	public static Complex Tan(Complex c)
		{
		Complex sin = Sin(c);
		Complex cos = Cos(c);
		return (Div(sin, cos));
		}
	public static Complex Cot(Complex c)
		{
		Complex sin = Sin(c);
		Complex cos = Cos(c);
		return (Div(cos, sin));
		}
	public static Complex Sec(Complex c)
		{
		Complex cos = Cos(c);
		return (Div(Complex.Number(1.0, 0.0), cos));
		}
	public static Complex Csc(Complex c)
		{
		Complex sin = Sin(c);
		return (Div(Complex.Number(1.0, 0.0), sin));
		}
	public static Complex Sinh(Complex c)
		{
		double ex = Math.exp(c.re);
		double ex_pos = ex / 2.0;
		double ex_neg = 0.5 / ex;
		double re = Math.cos(c.im) * (ex_pos - ex_neg);
		double im = Math.sin(c.im) * (ex_pos + ex_neg);
		return (new Complex(re, im));
		}
	public static Complex Cosh(Complex c)
		{
		double ex = Math.exp(c.re);
		double ex_pos = ex / 2.0;
		double ex_neg = 0.5 / ex;
		double re = Math.cos(c.im) * (ex_pos + ex_neg);
		double im = Math.sin(c.im) * (ex_pos - ex_neg);
		return (new Complex(re, im));
		}
	public static Complex Tanh(Complex c)
		{
		Complex sinh = Sinh(c);
		Complex cosh = Cosh(c);
		return (Div(sinh, cosh));
		}
	public static Complex Coth(Complex c)
		{
		Complex sinh = Sinh(c);
		Complex cosh = Cosh(c);
		return (Div(cosh, sinh));
		}
	public static Complex Sech(Complex c)
		{
		Complex cosh = Cosh(c);
		return (Div(Complex.Number(1.0), cosh));
		}
	public static Complex Csch(Complex c)
		{
		Complex sinh = Sinh(c);
		return (Div(Complex.Number(1.0), sinh));
		}
	public static Complex Asin(Complex c)
		{
		Complex w = Sqrt(Sub(Complex.Number(1.0), Mul(c, c)));
		return (Muli(Log(Add(Muli(c, 1.0), w)), -1.0));
		}
	public static Complex Acos(Complex c)
		{
		Complex w = Sqrt(Sub(Complex.Number(1.0), Mul(c, c)));
		return (Muli(Log(Add(c, Muli(w, 1.0))), -1.0));
		}
	public static Complex Atan(Complex c)
		{
		Complex i = Complex.Number(0.0, 1.0);
		return (Muli(Log(Div(Add(i, c), Sub(i, c))), 0.5));
		}
	public static Complex Asec(Complex c)
		{
		return (Acos(Div(Complex.Number(1.0), c)));
		}
	public static Complex Acsc(Complex c)
		{
		return (Asin(Div(Complex.Number(1.0), c)));
		}
	public static Complex Acot(Complex c)
		{
		Complex i = Complex.Number(0.0, 1.0);
		return (Muli(Log(Div(Add(i, c), Sub(c, i))), -0.5));
		}
	public static Complex Asinh(Complex c)
		{
		Complex w = Sqrt(Add(Mul(c, c), Complex.Number(1.0)));
		return (Log(Add(c, w)));
		}
	public static Complex Acosh(Complex c)
		{
		Complex w = Sqrt(Sub(Mul(c, c), Complex.Number(1.0)));
		return (Log(Add(c, w)));
		}
	public static Complex Atanh(Complex c)
		{
		Complex one = Complex.Number(1.0);
		return (Mul(Log(Div(Add(one, c), Sub(one, c))), 0.5));
		}
	public static Complex Asech(Complex c)
		{
		return (Acosh(Div(Complex.Number(1.0), c)));
		}
	public static Complex Acsch(Complex c)
		{
		return (Asinh(Div(Complex.Number(1.0), c)));
		}
	public static Complex Acoth(Complex c)
		{
		Complex one = Complex.Number(1.0);
		return (Mul(Log(Div(Add(one, c), Sub(c, one))), 0.5));
		}

	//----------------
	// 2-D Vector
	//----------------

	public static double Dot(Complex c, Complex d)
		{
		double result = (c.re * d.re) + (c.im * d.im);
		if (result == -0.0) result = 0.0;
		return (result);
		}
	public static double Cross(Complex c, Complex d)
		{
		double result = (c.re * d.im) - (c.im * d.re);
		if (result == -0.0) result = 0.0;
		return (result);
		}
	public static Complex Projection(Complex c, Complex d)
		{
		//----------------------------------------------
		// This returns the projection vector: the
		// "shadow" that c casts on a line parallel
		// to d, assuming light rays that are
		// perpendicular to d.
		//
		// The projection vector is:
		//
		//      d * (c dot d) / (d dot d)
		// or
		//      d * cos theta * |c|/|d|
		//----------------------------------------------
		double d2 = Dot(d, d);
		double scale = Dot(c, d) / d2;
		return (Mul(d, scale));
		}
	}


